/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.collection;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.jdt.core.dom.IBinding;
import org.intoJ.inferType.model.model.DeclarationElement;

/**
 * This set contains DeclarationElements, like variables or methods. Each
 * element inside is unique (as the name "Set" implies) - according to the key
 * of the binding inside.
 * 
 * @see DeclarationElement.equals()
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class DESet_Key {

	protected Set actualSet = new HashSet();

	/**
	 * Adds a new declaration element to the set, that is, a variable or a
	 * method. If the element is already in, it will not be added again. An
	 * element is already in, if it is equal to another as defined in
	 * DeclarationElement.equals().
	 * 
	 * @param element
	 */
	public void addDeclaration(DeclarationElement element) {

		actualSet.add(element); // Set already checks whether element is already
		// inside using equals
	}

	/**
	 * Add a collection of declaration elements.
	 * 
	 * @param collectionOfDeclarationElements
	 */
	public void addDeclarations(DESet_Key collectionOfDeclarationElements) {

		actualSet.addAll(collectionOfDeclarationElements.actualSet);
	}

	/**
	 * Checks according to key.
	 * 
	 * @param element
	 * @return boolean
	 */
	public boolean containsDeclaration(DeclarationElement element) {

		return (actualSet.contains(element));
	}

	/**
	 * Checks if this set contains the given set.
	 * 
	 * @param elements
	 * @return boolean
	 */
	public boolean containsDeclarations(DESet_Key elements) {

		return (actualSet.containsAll(elements.actualSet));
	}

	/**
	 * Removes declarations from this set
	 * 
	 * @param elements
	 */
	public void removeDeclarations(DESet_Key elements) {

		actualSet.removeAll(elements.actualSet);
	}
	public void removeDeclaration(DeclarationElement element) {

		actualSet.remove(element);
	}
	/**
	 * Returns an iterator.
	 * 
	 * @return Iterator
	 */
	public Iterator deIterator() {

		return actualSet.iterator();
	}

	/**
	 * 
	 * The number of declaration elements in this set. Note that composite DEs
	 * contribute only once, no matter how many DEs are contained within.
	 * 
	 * @return int
	 */
	public int size() {

		return actualSet.size();
	}

	/**
	 * True if empty.
	 * 
	 * @return boolean
	 */
	public boolean isEmpty() {

		return actualSet.isEmpty();
	}

	/**
	 * Returns this set as a collection
	 * 
	 * @return Collection
	 */
	public Collection toCollection() {

		return actualSet;
	}

	/**
	 * Method equals.
	 * @param o Object
	 * @return boolean
	 */
	public boolean equals(Object o) {

		if (o instanceof DESet_Key) {
			DESet_Key set = (DESet_Key) o;
			return ((set.size() == this.size()) && (set.containsDeclarations(this)));
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {

		if (actualSet.size() == 0)
			return "No elements (empty set)";

		StringBuffer result = new StringBuffer();
		for (Iterator i = actualSet.iterator(); i.hasNext();) {
			DeclarationElement d = (DeclarationElement) i.next();
			result.append(d + ",");
		}

		if (result.length() > 0)
			result.deleteCharAt(result.length() - 1);
		return result.toString();
	}

	/**
	 * Returns a representative string of the members of the DESet.
	 * Different from toString() the members are seperated by a line break.
	 * 
	 * @return representative string
	 */
	public String toStringLineWrapped() {

		if (actualSet.size() == 0)
			return "No elements (empty set)";

		StringBuffer result = new StringBuffer();

		// first element
		Iterator i = actualSet.iterator();
		result.append(i.next());

		// other elements
		while (i.hasNext()) {
			DeclarationElement d = (DeclarationElement) i.next();
			result.append("\n " + d);
		}
		return result.toString();
	}

	/**
	 * Checks whether the declaration with the given simple name 
	 * is part of this collection.
	 * 
	 * @param name
	 * @return boolean
	 */
	public boolean containsDeclaration(String name) {

		return (getDeclarationWithSimpleName(name) != null);
	}

	/**
	 * Checks whether the declaration with the given simple name 
	 * is part of this collection (exactly once).
	 * 
	 * @param name
	 * @return boolean
	 */
	public boolean containsDeclarationOnce(String name) {

		int count = 0;
		for (Iterator i = actualSet.iterator(); i.hasNext();) {
			DeclarationElement e = (DeclarationElement) i.next();
			if (e.containsDeclarationOnce(name))
				count++;
		}
		if (count == 1)
			return true;
		return false;
	}

	/**
	 * Returns all the declaration elements with the given simple name.
	 * 
	 * @param name
	 * @return DeclarationElement
	 */
	public DeclarationElement getDeclarationWithSimpleName(String name) {

		for (Iterator i = actualSet.iterator(); i.hasNext();) {
			DeclarationElement e = (DeclarationElement) i.next();
			if (e.containsDeclarationOnce(name))
				return e;
		}
		return null;

	}

	/**
	 * Return the declaration element which contains this binding. Note if no
	 * declaration element matching to this binding is found null is returned.
	 * 
	 * @param binding
	 * @return DeclarationElement
	 */
	public DeclarationElement getDeclarationElementOfBinding(IBinding binding) {

		for (Iterator i = actualSet.iterator(); i.hasNext();) {
			DeclarationElement currentDE = (DeclarationElement) i.next();
			if (currentDE.hasKey(binding.getKey()))
				return currentDE;
		}
		return null;
	}
}
