/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.collection;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.intoJ.inferType.model.model.TypeElement;

/**
 * This set contains TypeElements (types). Each element inside is unique (as the
 * name "Set" implies).
 * 
 * @see org.intoJ.model.TypeElement.equals()
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class TypeSet {

	private Set internal = new HashSet();

	/**
	 * Adds a type.
	 * 
	 * @param element TypeElement
	 */
	public void addType(TypeElement element) {

		internal.add(element);
	}

	/**
	 * Adds several types.
	 * 
	 * @param typeSet TypeSet
	 */
	public void addTypes(TypeSet typeSet) {

		internal.addAll(typeSet.toCollection());
	}

	/**
	 * Returns an iterator.
	 * 
	 * @return Iterator
	 */
	public Iterator typeIterator() {

		return internal.iterator();
	}

	/**
	 * @return int
	 */
	public int size() {

		return internal.size();
	}

	/**
	 * Returns this set as a collection.
	 * 
	 * @return Collection
	 */
	public Collection toCollection() {

		return internal;
	}

	/**
	 * Returns this set as a list.
	 * 
	 * @return List
	 */
	public List getAsList() {

		return new ArrayList(internal);
	}

	/**
	 * Removes a type.
	 * 
	 * @param type
	 * @return boolean
	 */
	public boolean removeType(TypeElement type) {

		return internal.remove(type);

	}

	/**
	 * Returns the type element for a binding (if it exists).
	 * 
	 * @param binding ITypeBinding
	 * @return TypeElement
	 */
	public TypeElement findTypeElementForBinding(ITypeBinding binding) {

		for (Iterator i = typeIterator(); i.hasNext();) {
			TypeElement te = (TypeElement) i.next();
			if (te.getBinding() != null && te.getBinding().getKey().equals(binding.getKey()))
				return te;
		}
		return null;
	}

	/**
	 * Checks whether a TypeElement with this binding exists in this set.
	 * 
	 * @param binding ITypeBinding
	 * @return boolean
	 */
	public boolean contains(ITypeBinding binding) {

		return (findTypeElementForBinding(binding) != null);
	}

	/**
	 * True if empty.
	 * 
	 * @return boolean
	 */
	public boolean isEmpty() {

		return internal.isEmpty();
	}

	/**
	 * Method toString.
	 * 
	 * @return String
	 */
	public String toString() {

		StringBuffer result = new StringBuffer();

		for (Iterator i = internal.iterator(); i.hasNext();) {
			TypeElement d = (TypeElement) i.next();

			result.append(d + ",");
		}

		if (result.length() > 0)
			result.deleteCharAt(result.length() - 1);
		return result.toString();
	}

	/**
	 * Returns true if the type set contains the given type once.
	 * 
	 * @param string the fqName of the type
	 * @return true if the type set contains the type once
	 */
	public boolean containsTypeOnce(String string) {

		int count = 0;
		for (Iterator i = internal.iterator(); i.hasNext();) {
			TypeElement e = (TypeElement) i.next();
			if (e.getFullyQualifiedName().equals(string))
				count++;
		}
		if (count == 1)
			return true;
		return false;
	}

	/**
	 * Return the type element inside the type set with the given
	 * full qualified name. If such a type element is not available null is returned.   
	 * 
	 * @param string the fqName of the type
	 * @return the type element with the given fqName (possibly null)
	 */
	public TypeElement getTypeElement(String string) {

		for (Iterator i = internal.iterator(); i.hasNext();) {
			TypeElement e = (TypeElement) i.next();
			if (e.getFullyQualifiedName().equals(string))
				return e;
		}
		return null;
	}

	/**
	 * Checks whether the element is part of this collection.
	 * 
	 * @param el
	 * @return boolean
	 */
	public boolean contains(TypeElement el) {

		return internal.contains(el);
	}
}