/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.model;

import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;

/**
 * 
 * A SingleDeclarationElement is one Declaration Element/Invocation Element.
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Andreas Mei&szlig;ner
 *  
 */
public class SingleDeclarationElement extends DeclarationElement {

	private IBinding binding;
	private ITypeBinding declaringClass;
	private int startPosition;
	private TypeElement newDeclaredType = null;
	/**
	 * Infer Type needs the declaredType as well
	 * for methods: return type; for variables: declared type.
	 */
	private ITypeBinding declaredType;
	/**
	 * We need the index because overridden parameters must be
	 * compared via their index in their parent (binding.getParent()).
	 */
	private int parameterIndex;
	private boolean isParameter;
	private IMethodBinding parameterMethod = null;
	private String[] params = null;

	/**
	 * Root constructor (called by all other constructors)
	 * 
	 * @param binding The binding (must NOT be null)
	 * @param startPosition Starting position inside parent (may be 0 if unknown)
	 */
	public SingleDeclarationElement(IBinding binding, int startPosition) {

		this.binding = binding;
		this.startPosition = startPosition;

		// Basic stuff
		this.declaringClass = null;
		this.declaredType = null;
		this.isParameter = false;
		this.parameterIndex = 0;

		if (binding.getKind() == IBinding.METHOD) {
			this.declaringClass = ((IMethodBinding) binding).getDeclaringClass();
			this.declaredType = ((IMethodBinding) binding).getReturnType();
		}
		if (binding.getKind() == IBinding.VARIABLE) {
			this.declaringClass = ((IVariableBinding) binding).getDeclaringClass();
			this.declaredType = ((IVariableBinding) binding).getType();
		}
		//ARRAY: The "array creation dummy de"
		if (binding.getKind() == IBinding.TYPE) {
			this.declaredType = (ITypeBinding) binding;
		}
		// Note that local variables (and parameters) don't have a declaring
		// class.
	}

	/**
	 * This contructor is used for local variables.
	 * 
	 * @param binding
	 * @param startPosition
	 * @param declaringClass
	 */
	public SingleDeclarationElement(IBinding binding, int startPosition, ITypeBinding declaringClass) {

		this(binding, startPosition);
		this.declaringClass = declaringClass;
	}

	/**
	 * This constructor is used for parameters
	 * 
	 * @param binding
	 * @param startPosition
	 * @param declaringClass
	 * @param index Index of this parameter inside the method parameter list.
	 * @param parent IMethodBinding
	 */
	public SingleDeclarationElement(IBinding binding, int startPosition, ITypeBinding declaringClass, int index, IMethodBinding parent) {

		this(binding, startPosition);
		this.declaringClass = declaringClass;
		this.parameterIndex = index;
		this.isParameter = true;
		this.parameterMethod = parent;
	}

	/**
	 * For methods and fields of types.
	 * 
	 * @param binding2
	 */
	public SingleDeclarationElement(IBinding binding2) {

		this(binding2, 0);
	}

	/**
	 * Returns the binding.
	 * 
	 * @return IBinding
	 */
	public IBinding getBinding() {

		return binding;
	}

	/**
	 * Returns the declaring class (only one).
	 * 
	 * @return ITypeBinding
	 */
	public ITypeBinding getDeclaringClass() {

		return declaringClass;
	}

	/**
	 * Method toString.
	 * 
	 * @return String
	 */
	public String toString() {

		// return getCompleteInformation();
		return getFullyQualifiedNameOfType() + " " + getBeautifiedName();
	}

	/**
	 * Returns the fully qualified name of this element.
	 * TODO shouldn't we use Util_TypeNames.getQualifiedNameForAllTypesOfTypes here?
	 * 
	 * @return String
	 */
	public String getFullyQualifiedNameOfType() {

		if (binding instanceof IVariableBinding)
			return ((IVariableBinding) binding).getType().getQualifiedName();
		if (binding instanceof IMethodBinding)
			return ((IMethodBinding) binding).getReturnType().getQualifiedName();
		return "";
	}

	/**
	 * Returns a beaufified name of this element.
	 * 
	 * @return String
	 */
	public String getBeautifiedName() {

		String name = "Unknown";

		if (binding instanceof IVariableBinding)
			name = binding.getName();
		if (binding instanceof IMethodBinding)
			name = getRepresentativeStringOfMethod((IMethodBinding) binding);
		if (binding instanceof ITypeBinding)
			name = "Array creation of type " + binding.getName();

		if (!name.endsWith("[]"))
			for (int i = 0; i < getArrayDimensionsOfType(); i++)
				name += "[]";

		return name;
	}

	/**
	 * Compares Declaration Elements according to their signature. This differs
	 * from the equals method: Overridden methods, fields, and parameters are
	 * equal according to equalsSignature, but not according to equals.
	 * 
	 * Note that overridden fields and methods can be compared according to
	 * their simple name. Parameters, however, can not: Their names can differ,
	 * they must be compared according to their position.
	 * 
	 * @param d
	 * @return boolean
	 */
	public boolean equalsSignature(DeclarationElement d) {

		if (d instanceof SingleDeclarationElement) {
			SingleDeclarationElement e = (SingleDeclarationElement) d;

			if (e.getBinding() instanceof IVariableBinding) {
				return variableEquals(this, e);
			} else if (e.getBinding() instanceof IMethodBinding) {
				return methodEquals(this.getBinding(), e.getBinding());
			}
		}
		return false;

	}

	/**
	 * This method is crucial to the success of the mission. It checks for
	 * variable and/or method equality.
	 * 
	 * DeclarationElements represent a lot of things: 
	 * -> Invocations on variables and methods (Methods and Fields) 
	 * -> Uses (where something is invoked on) (Methods, Fields, Local variables, Parameters) 
	 * -> Declarations (a method or field declaration)
	 * 
	 * Each and every one of them shall be equal, for example, the invocation of
	 * a method and the declaration of it should be the same, because, for
	 * example, we need to check whether a type contains all of the methods
	 * invoked on a certain variable.
	 * 
	 * This is what the equals method is about.
	 * @param o Object
	 * @return boolean
	 */
	public boolean equals(Object o) {

		if (o instanceof SingleDeclarationElement) {
			SingleDeclarationElement other = (SingleDeclarationElement) o;
			return (other.getBinding().getKey().equals(this.getBinding().getKey()));
		}
		return false;
	}

	/**
	 * Method hashCode.
	 * @return int
	 */
	public int hashCode() {

		// Use the complete information for hashing. Very important as this element
		// will be part of many hashed lists, maps, sets, you name it.
		return getCompleteInformation().hashCode();
	}

	/**
	 * Returns the kind of this element (one of the constants from DeclarationElement):
	 * DeclarationElement.PACKAGE
	 * DeclarationElement.INTERFACE
	 * DeclarationElement.CLASS
	 * DeclarationElement.METHOD
	 * DeclarationElement.FIELD
	 * DeclarationElement.PARAMETER
	 * DeclarationElement.LOCALVARIABLE
	 * DeclarationElement.UNKOWN
	 * 
	 * @return int
	 */
	public int getKind() {

		switch (binding.getKind()) {
			case IBinding.PACKAGE:
				return PACKAGE;
			case IBinding.TYPE: {
				ITypeBinding b = (ITypeBinding) binding;
				if (b.isInterface())
					return INTERFACE;
				return CLASS;
			}
			case IBinding.METHOD:
				return METHOD;
			case IBinding.VARIABLE: {
				IVariableBinding b = (IVariableBinding) binding;
				if (b.isField())
					return FIELD;
				else if (isParameter)
					return PARAMETER;
				else
					return LOCALVARIABLE;
			}
			default:
				return UNKNOWN;
		}
	}

	/**
	 * Returns the simple name of this element.
	 * 
	 * @return String
	 */
	public String getSimpleName() {

		return binding.getName();
	}

	/**
	 * Returns the declared type of this element.
	 * 
	 * @return ITypeBinding
	 */
	public ITypeBinding getDeclaredType() {

		return declaredType;
	}

	/**
	 * Returns the starting position of this element inside the source file.
	 * 
	 * @return int
	 */
	public int getStartPosition() {

		return this.startPosition;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#getModifiers()
	 */
	public int getModifiers() {

		return binding.getModifiers();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#hasKey(java.lang.String)
	 */
	public boolean hasKey(String key) {

		return key.equals(binding.getKey());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#containsDeclarationOnce(java.lang.String)
	 */
	public boolean containsDeclarationOnce(String string) {

		return this.getSimpleName().equals(string);
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getDeclaringClassesToString()
	 */
	public String getDeclaringClassesToString() {
		//FIXME: Bugfix. Not done yet. Happens to List.length access for arrays.
		if(getDeclaringClass() == null)  return "not resolvable";
		else return getDeclaringClass().getQualifiedName();
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getMethodBinding()
	 */
	public IMethodBinding getMethodBinding() {

		return parameterMethod;
	}

	public void setMethodParameterTypes(String[] params) {

		this.params = params;
	}

	public String[] getMethodParameterTypes() {

		if (params != null)
			return params;

		if (getBinding() instanceof IMethodBinding) {
			IMethodBinding b = (IMethodBinding) getBinding();
			ITypeBinding[] parameters = b.getParameterTypes();
			String[] returner = new String[parameters.length];
			for (int i = 0; i < returner.length; i++)
				returner[i] = parameters[i].getName();
			return returner;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#setNewDeclaredType(org.intoJ.inferType2.ast.model.TypeElement)
	 */
	public void setNewDeclaredType(TypeElement newDeclaredType) {

		this.newDeclaredType = newDeclaredType;
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getNewDeclaredType()
	 */
	public TypeElement getNewDeclaredType() {

		return this.newDeclaredType;
	}

	private String getRepresentativeStringOfMethod(IMethodBinding binding) {

		StringBuffer returning = new StringBuffer(binding.getName() + "(");
		ITypeBinding[] parameters = binding.getParameterTypes();
		for (int i = 0; i < parameters.length; i++) {
			returning.append(parameters[i].getName() + ",");
		}
		if (parameters.length > 0)
			returning.deleteCharAt(returning.length() - 1);
		returning.append(")");
		return returning.toString();
	}

	private boolean methodEquals(IBinding us, IBinding o) {

		if (us instanceof IMethodBinding) {
			// Test for method equality:
			IMethodBinding self = (IMethodBinding) us;
			IMethodBinding other = (IMethodBinding) o;

			if (!self.getName().equals(other.getName()))
				return false;
			if (!self.getReturnType().getQualifiedName().equals(other.getReturnType().getQualifiedName()))
				return false;
			ITypeBinding[] b1 = self.getParameterTypes();
			ITypeBinding[] b2 = other.getParameterTypes();
			if (!(b1.length == b2.length))
				return false;
			for (int i = 0; i < b1.length; i++) {
				if (!b1[i].getQualifiedName().equals(b2[i].getQualifiedName()))
					return false;
			}
			return true;
		}

		return false;
	}

	private boolean variableEquals(SingleDeclarationElement us, SingleDeclarationElement o) {

		// Test for variable equality:
		if (us.getBinding() instanceof IVariableBinding) {
			IVariableBinding self = (IVariableBinding) us.getBinding();
			IVariableBinding other = (IVariableBinding) o.getBinding();

			// Local variables, fields, and parameters in catch clauses and for
			// looks can be compared
			// according to their name and type; method parameters must be
			// compared according to their
			// method's signature and their index inside the method header.

			if (isParameter) {
				if (!o.isParameter)
					return false;
				return ((methodEquals(us.parameterMethod, o.parameterMethod)) && this.parameterIndex == o.parameterIndex);

			}

			// Fields and locals

			if (!self.getName().equals(other.getName()))
				return false;
			if (!self.getType().getQualifiedName().equals(other.getType().getQualifiedName()))
				return false;

			if ((us.getKind() != o.getKind()))
				return false; // otherweise, local variables and fields of
			// the same name are equal.

			// if it is a local, the other one can only be equivalent if it
			// is inside the same
			// class file - and in the same block. But this is the same as
			// being on the same position.

			if (getKind() == LOCALVARIABLE) {
				if (us.getStartPosition() != o.getStartPosition())
					return false;
			}

			return true;
		}
		return false;
	}

	public String getCompleteInformation() {

		StringBuffer b = new StringBuffer("SINGLE ");
		b.append(getBeautifiedName());
		b.append(" (" + getKindAsString() + ") ");
		b.append("in " + getDeclaringClassesToString());
		if (getMethodBinding() != null)
			b.append(" (Method: " + getMethodBinding().getName() + ")");
		return b.toString();
	}
}