/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.multirefac;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaModelMarker;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.search.IJavaSearchScope;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.RefactoringCore;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.ui.refactoring.RefactoringUI;
import org.eclipse.swt.widgets.Display;
import org.intoJ.inferType.calculation.calculation.typeNames.DefaultTypeNameGenerator;
import org.intoJ.inferType.exception.BadDEException;
import org.intoJ.inferType.exception.InternalCalculationException;
import org.intoJ.inferType.exception.UserAbortedException;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.model.model.DeclarationElement;
import org.intoJ.inferType.model.model.SingleDeclarationElement;
import org.intoJ.inferType3.ExternalInferTypeRunnable;
import org.intoJ.inferType3.multirefac.gathering.DeclarationElementFinder;

/**
 * <p>
 * This class performs InferType2 Multi Refactoring on a given IJavaElement
 * (project, package, type), which means that all DEs of all the types inside
 * the scope are refactored according to InferType2, until no more changes are
 * possible.
 * </p>
 * 
 * 
 * <hr>
 * <h3>The algorithm:</h3>
 * 
 * <pre>
 *  currentCompilationUnit = first
 *  changedAll=false
 *  done=false
 *  do
 *   changed=false
 *   types = getTypesForCompilationUnit()
 *   for each type
 *     des = getDEsForType()
 *     for each de
 *       if (isExecutable())
 *         execute()
 *         changed=true
 *         changedAll=true
 *         goto continue
 *       endif
 *     endfor
 *   endfor
 *   
 *   :continue
 *   if (isLastCompilationUnit())
 *     if (!changedAll) done=true
 *     currentCompilationUnit = first
 *     changedAll=false
 *   else
 *     if (!changed)
 *       currentCompilationUnit = next compilation unit
 *   endif    
 *  
 *  while not done.
 * </pre>
 * 
 * <hr>
 * 
 * <p>
 * For more information about the ongoing process, please enable one or more of
 * the tracing options available in this plugin and the inferType plugin. If you
 * are only concerned with the ongoing process, enable only the multi refac
 * trace and, optionally, grep "STATUS" from the console output.
 * </p>
 * 
 * @see DefaultTypeNameGenerator
 * 
 * $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class MultiRefactorer {

    protected IJavaProject javaProject;

    private String errorMessageIfNot;

    private IJavaSearchScope projectScope;

    private String errorMessage;

    private boolean error;

    int standardLen = 1000;

    int currentWorkedICUs = 0; // see below

    int currentWorkedLenAmount = 0; // see below

    private final boolean changeTypeArgs;
    private final boolean noAbstractClasses;

    /**
     * @param changeTypeArgs
     * @param noAbstractClasses
     */
    public MultiRefactorer(boolean changeTypeArgs, boolean noAbstractClasses) {
        this.changeTypeArgs = changeTypeArgs;
        this.noAbstractClasses = noAbstractClasses;
    }

    /**
     * Main method for Infer Type Multi Refac.
     * 
     * Valid javaElements are IJavaProjects, IPackageFragments and
     * ICompilationUnits.
     * 
     * @param theJavaElement
     *            selected java element
     * @param monitor
     *            the monitor
     * @throws InterruptedException
     */
    public void run(IJavaElement javaElement, IProgressMonitor monitor) throws InterruptedException {

        javaProject = javaElement.getJavaProject();

        try {
            monitor.beginTask("Starting up...", standardLen);

            projectScope = SearchEngine.createJavaSearchScope(new IJavaProject[] { javaProject });

            long iteration = 1; // the whole iteration.
            boolean thereWasAChange = false; // change during the whole
                                                // iteration?

            long changesMade = 0;
            long occurrencesUpdated = 0;

            error = false;
            boolean done = false; // not done.

            int currentICUIndex = 0; // Start at ICU 0

            String resetToName = null;

            long startTime = System.currentTimeMillis();

            boolean clearCache = true;
            do {

                // get SORTED list of compilation units.
                List allICUs = grabCompilationUnits(javaElement);

                ICompilationUnit currentICU = null;
                if (resetToName == null) {
                    // get to current index (this, of course, assumes that new
                    // types are always added after the current position).
                    currentICU = (ICompilationUnit) allICUs.get(currentICUIndex);
                } else {
                    // get the compilation unit to reset to.
                    for (Iterator i = allICUs.iterator(); i.hasNext();) {
                        ICompilationUnit u = (ICompilationUnit) i.next();
                        if (u.getElementName().equals(resetToName)) {
                            currentICU = u;
                            break;
                        }
                    }
                }

                ILog.log(1, "---------------------------------------------------------------------");
                ILog.log(1, "IT Multirefac STATUS UPDATE - Iteration " + attachZeroes(iteration, 2) + ", Changes made so far: " + attachZeroes(changesMade, 4) + ", Occurrences updated: " + attachZeroes(occurrencesUpdated, 4) + ", Compilation Unit: "
                        + attachZeroes(currentICUIndex + 1, 3) + "/" + attachZeroes(allICUs.size(), 3) + " (" + currentICU.getElementName() + ")");
                ILog.log(1, "---------------------------------------------------------------------");

                org.intoJ.inferType.util.ILog.log(org.intoJ.inferType.util.ILog.LOG_EVAL, 0, "New MultiRefac iteration: " + iteration);

                monitor.setTaskName("Infer Type Multi Iteration " + attachZeroes(iteration, 2) + ": Changes: " + attachZeroes(changesMade, 4) + "; Files: " + attachZeroes(currentICUIndex + 1, 3)
                        + "/" + attachZeroes(allICUs.size(), 3));

                if (currentICUIndex != currentWorkedICUs) {
                    int absolute = (standardLen * (currentICUIndex + 1)) / allICUs.size();
                    int relative = absolute - currentWorkedLenAmount;
                    if (relative > 0) { // if <0, wait (effect induced by adding
                                        // comp. units)
                        currentWorkedLenAmount = absolute;
                        currentWorkedICUs = currentICUIndex;
                        monitor.worked(relative);
                    }
                }

                // get a list of ITypes in a DEFINED ORDER
                IType[] types = currentICU.getAllTypes();

                int numberOfChangesInthisICU = 0;
                // (2-2) ...each IType
                String currentDeInfo = "";
                
                for (int i = 0; i < types.length; i++) {

                    IType currentType = types[i];
                    final int currentIno = i + 1;
                    ILog.log(2, "IType " + currentIno + " of " + types.length + ": " + currentType.getElementName());

                    // Do not attempt to find declaration elements of Exceptions, Enums, Annotations
                    // or Throwables
                    if (isUnsupportedType(currentType))
                        continue; // go to next type inside the type for loop

                    /**
                     * We get the declaration elements for this type. The
                     * resulting list contains single and composite declaration
                     * elements(!).
                     */
                    DESet_Key des_unsorted = DeclarationElementFinder.findDeclarationSet(currentType, projectScope, new SubProgressMonitor(monitor, 0), 500);

                    // Sort the DEs according to key.
                    List des = sortDeclarationElements(des_unsorted);

                    ILog.log(2, "Found " + des.size() + " declaration elements declared with this type.");

                    // (2-3) ...each DeclarationElement
                    int j = 1;

                    // for (Iterator deIter = des.iterator(); deIter.hasNext();)
                    // {
                    while (!des.isEmpty()) {
                        // DeclarationElement currentDE = (DeclarationElement)
                        // deIter.next();
                        DeclarationElement currentDE = (DeclarationElement) des.get(0);
                        try {

                            currentDeInfo = currentDE.getCompleteInformation();
                            ILog.log(3, "DE (" + j + "/" + des.size() + ") : " + currentDeInfo);
                            numberOfChangesInthisICU = doInferType3Runnable(new SubProgressMonitorWoTaskName(monitor, 0), currentDE, currentType, clearCache);
                            clearCache = false;

                            des.remove(0);
                            des_unsorted.removeDeclaration(currentDE);

                            j++;

                        } catch (BadDEException e) {
                            ILog.log(4, "Could not parse this DE. Reason: " + e.getMessage());
                            // continue
                        } catch (UserAbortedException e) {
                            throw new UserAbortedException("Infer Type", "Infer Type was canceled.");
                        } catch (Exception e) {
                            ILog.logError("Exception while running Infer Type.", e);
                            e.printStackTrace();
                            error = true;
                            return; // PANIC
                        }

                        if (numberOfChangesInthisICU > 0)
                            break; // out of this inner 'de' for loop
                    }

                    if (numberOfChangesInthisICU > 0)
                        break; // out of the 'type' for loop
                }
                // }
                if (numberOfChangesInthisICU > 0) {

                    // There was a change; clean and rebuild.
                    changesMade++;
                    occurrencesUpdated += numberOfChangesInthisICU;

                    if (!cleanAndRebuild()) {
                        Activator.getDefault().getLog().log(
                                new Status(IStatus.ERROR, 
                                        Activator.getDefault().getBundle().getSymbolicName(),
                                        IStatus.OK,
                                        "Error while refactoring " + currentDeInfo, 
                                        null));                        
                        error = true;
                        break;
                    }
                    projectScope = SearchEngine.createJavaSearchScope(new IJavaProject[] { javaProject });
                    thereWasAChange = true;

                    /**
                     * the loop will restart now. However, it cannot restart at
                     * the current index because the ICU at this index MAY have
                     * changed. This can happen if the current ICU is an
                     * IT-induced one, say, "ClassA_for_parameter1", and in this
                     * loop a new ICU was introduced, say "ClassA_for_aField".
                     * The latter will be inserted in the list before the old
                     * one.
                     * 
                     * It is also possible that several new types will be
                     * introduced, some of them for out assignments, which may
                     * have completely arbitrary names.
                     * 
                     * @see DefaultTypeNameGenerator.
                     * 
                     * So, just reset to the type which currently was under
                     * discussion
                     * 
                     */
                    resetToName = currentICU.getElementName();

                } else { // no change in this CU

                    currentICUIndex++;
                    resetToName = null; // do not reset to a name.

                    if (currentICUIndex == allICUs.size()) {
                        if (!thereWasAChange)
                            done = true; // that's it.
                        currentICUIndex = 0; // restart.
                        monitor.beginTask("New iteration...", standardLen); // HACK
                        thereWasAChange = false;
                        iteration++;
                    }
                }

            } while (!done);

            long timeNow = System.currentTimeMillis();

            ILog.log(1, "---------------------------------------------------------------------");
            if (error) {
                errorMessage = "Multirefac has come to a halt due to an error. Sorry.";

            } else {
                errorMessage = "Multirefac has come to its final end. Thank you for using IT technology.";
            }
            ILog.log(1, errorMessage);
            ILog.log(1, "Done. Iteration count: " + iteration + " Changes made: " + changesMade + ", Occurrences updated: " + occurrencesUpdated + ".");
            ILog.log(1, "Time taken: " + readableTime(timeNow - startTime));
            ILog.log(1, "---------------------------------------------------------------------");

            monitor.done();

        } catch (JavaModelException e) {
            ILog.logError("There was a problem accessing the Java Model.", e);
            throw new InterruptedException("There was a problem accessing the Java Model.");
        } catch (CoreException e) {
            ILog.logError(("Core Exception in MultiRefac"), e);
            throw new InternalCalculationException("Core Exception", "Problem in MultiRefac: " + e.getMessage());
        }
    }

    private boolean cleanAndRebuild() {

        // CLEAN AND REBUILD:

        try {
            javaProject.getProject().build(IncrementalProjectBuilder.CLEAN_BUILD, new NullProgressMonitor());
            javaProject.getProject().build(IncrementalProjectBuilder.FULL_BUILD, new NullProgressMonitor());
            javaProject.close();
            javaProject.open(new NullProgressMonitor());
        } catch (CoreException e) {
            ILog.log(1, "Sorry. CoreException while rebuilding.");
            return false;
        }

        if (hasRealErrors(javaProject.getProject())) {
            ILog.log(1, "Sorry. Aborted due to compilation errors in project. ");
            return false;
        }

        return true;
    }

    /**
     * @param currentType
     * @return
     * @throws JavaModelException
     */
    private boolean isUnsupportedType(IType currentType) throws JavaModelException {
        if (currentType.isAnnotation() || currentType.isEnum()) {
            return true;
        }

        boolean exceptionType = false;
        ITypeHierarchy h = (currentType.newSupertypeHierarchy(new NullProgressMonitor()));
        IType[] supers = h.getAllSuperclasses(currentType);
        for (int sl = 0; sl < supers.length; sl++) {
            if ((supers[sl].getFullyQualifiedName().equals("java.lang.Exception")) || (supers[sl].getFullyQualifiedName().equals("java.lang.Throwable"))) {
                ILog.log(4, "Exception or Throwable; continuing with next compilation unit...");
                exceptionType = true;
                break; // out of this inner for loop
            }
        }
        return exceptionType;
    }


    private IBinding getBinding(DeclarationElement currentDE, boolean clearCache) {
        SingleDeclarationElement singleDE = (SingleDeclarationElement) currentDE;

        if (singleDE.getBinding() == null || singleDE.getBinding().getJavaElement() == null
                || singleDE.getDeclaringClass() == null) 
        {
        	return null;
        }
        
        return singleDE.getBinding();
    }


    /**
     * 
     * Returns true if there was a change.
     * 
     * @param monitor
     * @param des
     * @param currentDE
     * @param currentType 
     * @param clearCache 
     * @return
     * @throws InterruptedException
     */
    private int doInferType3Runnable(final IProgressMonitor monitor, DeclarationElement currentDE, IType currentType, boolean clearCache) throws InterruptedException {

        ILog.log(3, "=============> GOTO InferTypeProcessor <===============");
        int numberOfChanges = 0;
        try {

            IBinding binding = getBinding(currentDE, clearCache);
            if (binding != null) {
                SingleDeclarationElement singleDE = (SingleDeclarationElement) currentDE;
        
    
                ITypeBinding typeBinding = stripBinding(singleDE.getDeclaredType());
                if (typeBinding != null) {
                    if (typeBinding.getErasure().getJavaElement().equals(currentType)) {
                        numberOfChanges = doSingleRunnable(monitor, clearCache, singleDE, null);
                        if (numberOfChanges > 0) {
                            return numberOfChanges;
                        }
                    }
                    numberOfChanges = handleParametrizedType(monitor, clearCache, typeBinding, singleDE, currentType, new ArrayList<String>());
                }
            }
        } finally {
            ILog.log(3, "=============> COMEFROM InferTypeProcessor <===============");
            ILog.log(3, "Done multi-refaccing DE " + currentDE.getCompleteInformation());
        }
            
        return numberOfChanges;
    }

    private ITypeBinding stripBinding(ITypeBinding typeBinding) {
        if (typeBinding.isWildcardType()) {
            typeBinding = typeBinding.getBound();
            if (typeBinding == null) {
                return null;
            }
        }
        if (typeBinding.isArray()) {
            typeBinding = typeBinding.getElementType();
        }
        return typeBinding;
    }

    private int handleParametrizedType(IProgressMonitor monitor, boolean clearCache, ITypeBinding typeBinding, SingleDeclarationElement singleDE, IType currentType, List<String> previousParamKeys) throws UserAbortedException, InterruptedException {
        int numberOfChanges = 0;
        if (typeBinding.isParameterizedType()) {
            ITypeBinding[] typeArguments = typeBinding.getTypeArguments();
            ITypeBinding[] typeParameters = typeBinding.getTypeDeclaration().getTypeParameters();
            for (int i = 0; i < typeParameters.length; i++) {
                List<String> currentList = new ArrayList<String>(previousParamKeys);
                currentList.add(typeParameters[i].getKey());

                ITypeBinding argument = stripBinding(typeArguments[i]);
                if (argument != null) {
                    if (argument.getErasure().getJavaElement().equals(currentType)) {
                        numberOfChanges = doSingleRunnable(monitor, clearCache, singleDE, currentList);
                        if (numberOfChanges > 0) {
                            return numberOfChanges;
                        }
                    }
                    
                    numberOfChanges = handleParametrizedType(monitor, clearCache, argument, singleDE, currentType, currentList);
                    if (numberOfChanges > 0) {
                        return numberOfChanges;
                    }
                }
                
                currentList.remove(currentList.size() - 1);
            }
        }
        return numberOfChanges;
    }

    private int doSingleRunnable(final IProgressMonitor monitor, boolean clearCache, SingleDeclarationElement singleDE, List<String> typeVariableKeys)
            throws InterruptedException, UserAbortedException {
        
        int numberOfChanges = 0;
        
        ExternalInferTypeRunnable itRunnable = new ExternalInferTypeRunnable(singleDE.getBinding(),
                singleDE.getDeclaringClass(), singleDE.getStartPosition(),
                typeVariableKeys,
                clearCache, changeTypeArgs, noAbstractClasses, false);
        
   
        try {
            itRunnable.run(monitor);
            numberOfChanges = itRunnable.getNumberOfChanges();
        } catch (OperationCanceledException e1) {
            throw new UserAbortedException("cancled", "");
        } catch (InvocationTargetException e) {
            Throwable target = e.getTargetException();
            if (target instanceof OperationCanceledException) {
                throw new UserAbortedException("cancled", "");                    
            }
            target.printStackTrace();
            ILog.logError("Error refactoring " + singleDE.getCompleteInformation(), target);
        }
        return numberOfChanges;
    }

    public boolean hasRealErrors(IResource resource) {

        try {
            if (resource != null && resource.exists()) {
                IMarker[] marker = resource.findMarkers(IJavaModelMarker.JAVA_MODEL_PROBLEM_MARKER, false, IResource.DEPTH_INFINITE);
                for (int i = 0; i < marker.length; i++) {
                    if (((Integer) marker[i].getAttribute(IMarker.SEVERITY)).intValue() == IMarker.SEVERITY_ERROR)
                        return true;
                }
            }
        } catch (CoreException e) {
            // assume there are no problems
        }
        return false;
    }

    /**
     * Returns all ICompilationUnits of the given IJavaElement.
     * 
     * @param javaElement
     *            java element
     * @return List of ICompilationUnit
     * @throws InterruptedException
     */
    private List grabCompilationUnits(IJavaElement javaElement) throws JavaModelException {

        List scopeElements = new ArrayList();
        ICompilationUnit[] units = null;

        if (javaElement instanceof ICompilationUnit)
            scopeElements.add(javaElement);

        else if (javaElement instanceof IPackageFragment) {
            units = ((IPackageFragment) javaElement).getCompilationUnits();
            for (int j = 0; j < units.length; j++)
                scopeElements.add(units[j]);
        }

        else if (javaElement instanceof IJavaProject) {
            IPackageFragment[] packages = ((IJavaProject) javaElement).getPackageFragments();
            for (int i = 0; i < packages.length; i++) {
                if (packages[i].getKind() != IPackageFragmentRoot.K_BINARY) {
                    units = packages[i].getCompilationUnits();
                    for (int j = 0; j < units.length; j++)
                        scopeElements.add(units[j]);
                }
            }
        }

        Collections.sort(scopeElements, new CompilationUnitComparator());

        return scopeElements;
    }

    /**
     * This comparator compares ICompilationUnits.
     * 
     * @author Andreas Mei&szlig;ner
     */
    private class CompilationUnitComparator implements Comparator {

        /*
         * (non-Javadoc)
         * 
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object arg0, Object arg1) {

            if ((arg0 == null) || (arg1 == null))
                return 0;
            if (!(arg0 instanceof ICompilationUnit))
                return 0;
            if (!(arg1 instanceof ICompilationUnit))
                return 0;

            ICompilationUnit icu0 = (ICompilationUnit) arg0;
            ICompilationUnit icu1 = (ICompilationUnit) arg1;

            return icu0.getElementName().compareTo(icu1.getElementName());
        }

    }

    private List sortDeclarationElements(DESet_Key set) {

        List l = new ArrayList(set.toCollection());
        Collections.sort(l, new DeclarationElementComparator());
        return l;
    }

    /**
     * This comparator compares DeclarationElements.
     * 
     * @author Philip Mayer
     */
    private class DeclarationElementComparator implements Comparator {

        /*
         * (non-Javadoc)
         * 
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object arg0, Object arg1) {

            if ((arg0 == null) || (arg1 == null))
                return 0;
            if (!(arg0 instanceof DeclarationElement))
                return 0;
            if (!(arg1 instanceof DeclarationElement))
                return 0;

            DeclarationElement de0 = (DeclarationElement) arg0;
            DeclarationElement de1 = (DeclarationElement) arg1;

            return de0.getSimpleName().compareTo(de1.getSimpleName()); // simple
                                                                        // name
                                                                        // does
                                                                        // not
                                                                        // change
                                                                        // in a
                                                                        // refactoring!
        }

    }

    public static String readableTime(long time) {

        long allSeconds = time / 1000;

        long allMinutes = allSeconds / 60;
        long allHours = allMinutes / 60;

        long minutes = allMinutes % 60;
        long seconds = allSeconds % 60;

        return attachZero(allHours) + ":" + attachZero(minutes) + ":" + attachZero(seconds);
    }

    /**
     * @param number
     * @return
     */
    private static String attachZero(long number) {

        return (number < 10) ? "0" + number : String.valueOf(number);
    }

    private static String attachZeroes(long number, int digits) {

        int add = digits - String.valueOf(number).length();
        if (add <= 0)
            return String.valueOf(number);
        StringBuffer b = new StringBuffer();
        for (int i = 0; i < add; i++)
            b.append("0");
        b.append(number);
        return b.toString();
    }

    public boolean isError() {

        return error;
    }

    public String getErrorMessage() {

        return errorMessage;
    }
    
    private static class SubProgressMonitorWoTaskName extends SubProgressMonitor {
        public SubProgressMonitorWoTaskName(IProgressMonitor monitor, int ticks) {
            super(monitor, ticks, SUPPRESS_SUBTASK_LABEL);
        }

        public void setTaskName(String name) {
            //ignore
        }
    }
}
