/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.actions;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ltk.ui.refactoring.RefactoringWizardOpenOperation;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.ui.InferTypeWizard;
import org.intoJ.inferType3.internal.util.RefactoringSaveHelper;

/**
 * Abstract superclass of all actions.
 *
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */

public class AbstractInferTypeAction {
    
    protected ISelection selection;    
	
    /**
     * Starts Infer Type by running a RefactoringWizardOpenOperation.
     * 
     * @param declarationElement the selected element
     * @throws InterruptedException
     */
    protected final void runWithSelectedElement(SelectedElement declarationElement) throws InterruptedException {
        if (checkPrerequesites(declarationElement)) {        
            InferTypeRefactoring refactoring= new InferTypeRefactoring(declarationElement);
            InferTypeWizard refactoringWizard= new InferTypeWizard(refactoring);
            
            RefactoringSaveHelper saveHelper = new RefactoringSaveHelper();
            if (saveHelper.saveEditors(getShell())) {
                RefactoringWizardOpenOperation op= new RefactoringWizardOpenOperation(refactoringWizard);                
                op.run(getShell(), "Infer Type");
            }
        }
    }
    
    /**
     * Sets the selection.
     * 
     * @param action IAction
     * @param selection ISelection
     */
    public void selectionChanged(IAction action, ISelection selection) {
        this.selection = selection;
    }
        

    /**
     * Checks whether it is possible to work with selected declaration element.
     * 
     * @param de the declaration element
     * @return true if it is possible to work with the selected declaration
     *         element
     */
    private boolean checkPrerequesites(SelectedElement de) {

        if (de.getFullyQualifiedNameOfType().equals("void")) {
            showError("No return type", "Infer Type cannot be performed on constructors or methods with a 'void' return type.");
            return false;
        }

        //ARRAY use element type
        ITypeBinding typeToCheck = de.getDeclaredType();
        if (typeToCheck.isArray()) {
            typeToCheck = typeToCheck.getElementType();
        }

        if (!typeToCheck.isFromSource()) {
            showError("Non-project type", "Declared type of selected declaration element is not part of this project."
                    + "\nPlease select a declaration element declared with a project type.");
            return false;
        }
        if (typeToCheck.isLocal()) {
            showError("Local type", "Declared type of selected declaration element is local."
                    + "\nInfer Type cannot be performed on declaration elements declared with local types.");
            return false;
        }
        if (typeToCheck.isArray()) {
            showError("Array type", "Declared type of selected declaration element is an array type."
                    + "\nInfer Type cannot be performed on declaration elements declared with array types.");
            return false;
        }
        if (typeToCheck.isPrimitive()) {
            showError("Primitive type", "Declared type of selected declaration element is primitive."
                    + "\nInfer Type cannot be performed on declaration elements declared with primitive types.");
            return false;
        }
        if (typeToCheck.isAnnotation()) {
            showError("Annotation", "Infer Type is not defined for annotations.");
            return false;
        }
        if (typeToCheck.isTypeVariable()) {
            showError("Type Variable", "Infer Type is not defined for type variables.");
            return false;
        }
        if (typeToCheck.isWildcardType()) {
            showError("Wildcard Type", "Infer Type is not defined for wildcard types.");
            return false;
        }
        if (typeToCheck.isEnum()) {
            showError("Enum", "Infer Type is not defined for enums.");
            return false;
        }
        if (typeToCheck.isAnnotation()) {
            showError("Annotation", "Infer Type is not defined for annotation types.");
            return false;
        }
        if (de.getBinding() instanceof ITypeBinding) {
            if (((ITypeBinding)de.getBinding()).getTypeDeclaration().isGenericType()) {
                showError("Import Declarations", "Refactoring of imports and corresponding declaration elements  is not supported for generic types.");
                return false;                
            }
        }
        return true;
    }

    /**
     * Shows an error message.
     * 
     * @param title the title of the dialog
     * @param msg the error message
     */
    protected final void showError(String title, String msg) {
        MessageDialog.openError(getShell(), title, msg);
    }

    /**
     * Returns the current shell
     * 
     * @return shell
     */
    protected Shell getShell() {

        IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
        if (window == null) {
            return null;
        }
        return window.getShell();
    }   
}
