/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.CompilationUnitRange;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ArrayType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.GenericType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ParameterizedType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeEnvironment;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeVariable;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.WildcardType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraintVariable;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.IndependentTypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ParameterTypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ReturnTypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.SubTypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.TypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.VariableVariable2;
import org.eclipse.jdt.internal.corext.refactoring.util.RefactoringASTParser;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.creation.InferTypeConstraintsCreator;
import org.intoJ.inferType3.internal.constraintModel.util.ConstraintTraversal;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.variables.BoundVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.GenericMethodTypeArgumentVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutableBoundVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutableElementVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutableTypeVariableWithDescr;
import org.intoJ.inferType3.internal.constraintModel.variables.IndependentParameterTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.IndependentReturnTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.StaticVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.SuperVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ThisVariable;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.util.BindingUtil;
import org.intoJ.inferType3.internal.util.CacheMap;
import org.intoJ.inferType3.internal.util.VariableSet;

/**
 * The model that holds all type constraints and 
 * variables for Infer Type.
 * <p>
 * The instances of the variables and constraints will be created here.
 * In addition, element variables and constraints between them are created
 * here automatically (triggered by the {@link InferTypeConstraintsCreator}).
 * </p>
 * @author Hannes Kegel
 *
 */
public class InferTypeConstraintsModel {

    // used as identifiers for data connected to the variables
    private static final String DATA_USAGE = "us";
    private static final String DATA_SIMPLE_NAME = "sn";
    private static final String DATA_METHOD_NAME = "mn";
    private static final String DATA_ACCESS_SET = "as";
    private static final String DATA_NEW_TYPE = "nt";
    private static final String DATA_ELEMENT_VARIABLES = "evs"; 
    private static final String DATA_IMMUTABILITY_ = "imm"; 

    private static final Map<String, ElementVariable> EMPTY_COLLECTION_ELEMENT_VARIABLES_MAP = Collections.emptyMap();

    /**
     * The set of constraint variables (element type:
     * <code>ConstraintVariable2</code>)
     */
    private final VariableSet constraintVariables = new VariableSet();
   
    /**
     * The set of all constraints
     */
    private final Set<ITypeConstraint2> allConstraints = new HashSet<ITypeConstraint2>();

    /** 
     * The type environment to use. ITypeBindings should not be held by the constraints
     * because they are memory intensive. 
     */
    private TypeEnvironment typeEnvironment;
    
    /**
     * The selected element to be refactored.
     */
    private final SelectedElement selectedElement;
    
    /**
     * The cached constraint variable that matches the selected element.
     */
    private ConstraintVariable2 selectedVariable = null;
    
    /**
     * If the scope is not one variable (for the import selection) this contains
     * all variables in the selection.
     */
    private List<ConstraintVariable2> completeTypeSelection = new ArrayList<ConstraintVariable2>();

    /**
     * Cache for TTypes. Keys are binding keys.
     */
    private Map<String, TType> tTypeCache = new CacheMap<String, TType>(128);
    
    /**
     * The current refactoring status. Used to record errors.
     */
    private RefactoringStatus refactoringStatus = new RefactoringStatus();
    
    /**
     * The single constructor.
     * 
     * @param typeEnvironment the type environment used
     * @param selectedElement the element to be refactored
     */
    public InferTypeConstraintsModel(final TypeEnvironment typeEnvironment, SelectedElement selectedElement) {
        this.typeEnvironment = typeEnvironment;
        this.selectedElement = selectedElement;
    }

    /**
     * Returns the constraint variable that matches the SelectedElement given in the constructor.
     * 
     * @return the selected constraint variable
     * @throws JavaModelException
     */
    public ConstraintVariable2 getSelectedConstraintVariable() throws JavaModelException {
        if (selectedVariable == null && selectedElement != null) {
            ICompilationUnit compilationUnit = (ICompilationUnit) selectedElement.getBinding().getJavaElement().getAncestor(IJavaElement.COMPILATION_UNIT);

            int declOffset = selectedElement.getTypeStartPos();

            for (Object element : getConstraintVariables()) {
                ConstraintVariable2 variable = (ConstraintVariable2) element;
                if (variable instanceof ITypeConstraintVariable) {
                    ITypeConstraintVariable typeConstraintVariable = (ITypeConstraintVariable) variable;
                    if (compilationUnit.equals(typeConstraintVariable.getRange().getCompilationUnit()) && declOffset == typeConstraintVariable.getRange().getSourceRange().getOffset()) {
                        if (selectedElement.getTypeVariableKeys() != null) {
                            for (String key : selectedElement.getTypeVariableKeys()) {
                                variable = getElementVariables(variable).get(key);
                            }
                        }
                        selectedVariable = variable;
                        break;
                    }
                }
            }
        }
        return selectedVariable;
    }

    /**
     * Creates a covariant return type constraint.
     *
     * @param descendant the descendant type constraint variable
     * @param descendentReturnType the return type of the descendant
     * @param ancestor the descendant ancestor constraint variable
     * @param ancestorReturnType the return type of the ancestor
     */
    public final void createCovariantTypeConstraint(ConstraintVariable2 descendant, ITypeBinding descendentReturnType, ConstraintVariable2 ancestor, ITypeBinding ancestorReturnType) {
        if (descendant == null || ancestor == null) {
            return;
        }
        descendant = constraintVariables.addOrGet(descendant);
        ancestor = constraintVariables.addOrGet(ancestor);

        if (descendentReturnType.getKey().equals(ancestorReturnType.getKey())) {
            createEqualityConstraint(descendant, ancestor);
        } else if (BindingUtil.isFirstSubTypeCompatible(descendentReturnType, ancestorReturnType)) {
            createSubtypeConstraint(descendant, ancestor);            
        } else if (BindingUtil.isFirstSubTypeCompatible(ancestorReturnType, descendentReturnType)) {
            createSubtypeConstraint(ancestor, descendant);            
        } else {
            // one is a type var or a parameterized type with type vars
            //System.out.println("WARNING: COVARIANT TYPES ARE IN NO RELATION: " + descendentReturnType.getKey() + ", " + ancestorReturnType.getKey());
            createEqualityConstraint(descendant, ancestor);
        }
    }

    /**
     * Creates a declaring type variable that cannot be changed or returns the existing one
     * matching the parameters.
     * 
     * <p>
     * A declaring type variable stands for a type where something has been
     * declared.
     * </p>
     * 
     * @param type the type binding
     * @param description the description for the possible reason why the refactoring cannot be executed
     * @return the created type variable
     */
    public final ConstraintVariable2 makeImmutableDeclaringTypeVariable(ITypeBinding type, String description, String bindingKey, ImmutabilityReason immutabilityReason) {
        if (type.isArray()) {
            type = type.getElementType();
        }
        type = type.getTypeDeclaration();
        return constraintVariables.addOrGet(new ImmutableTypeVariableWithDescr(createTType(type), description, bindingKey, immutabilityReason));
    }

    public final ConstraintVariable2 makeImmutableDeclaringTypeVariable(ITypeBinding type, String description, String bindingKey, EnumSet<ImmutabilityReason> immutabilityReason) {
        if (type.isArray()) {
            type = type.getElementType();
        }
        type = type.getTypeDeclaration();
        return constraintVariables.addOrGet(new ImmutableTypeVariableWithDescr(createTType(type), description, bindingKey, immutabilityReason));
    }

    /**
     * Creates a declaring type variable based on a method invocation or returns the existing one.
     * matching the parameters.
     * <p>
     * The returned declaring type variable stands for an (undefined) type where a method has been
     * declared.
     * </p>
     * 
     * @param method the method that has been invoked
     * @param type the type binding
     * @return the created declaring type variable
     */
    public ConstraintVariable2 makeDeclaringTypeVariable(IMethodBinding method, ITypeBinding type) {
        if (type.isArray()) {
            type = type.getElementType();
        }
        type = type.getTypeDeclaration();
        return constraintVariables.addOrGet(new DeclaringTypeVariable(method, createTType(type)));
    }

    /**
     * Creates an equality constraint.
     * 
     * @param left the left typeconstraint variable
     * @param right the right typeconstraint variable
     */
    public final void createEqualityConstraint(ConstraintVariable2 left, ConstraintVariable2 right) {
        createEqualityConstraint(left, right, true);
    }

    private void createEqualityConstraint(ConstraintVariable2 left, ConstraintVariable2 right, boolean createNewElementVars) {
        if (left != null && right != null && shouldBind(left.getType()) && shouldBind(right.getType())) {
            left = constraintVariables.addOrGet(left);
            right = constraintVariables.addOrGet(right);
            // LogUtil.debugOut(this, "equality constraint: " + left + ", right: "
            // + right);

            if (left instanceof TypeVariable2) {
                setSimpleName(left, right);
                setMethodName(left, right);
                TAA3EnhancedConstraintDataModel.setTypeVariableRange(right, ((TypeVariable2)left).getRange().getSourceRange());
            } else if (right instanceof TypeVariable2) {
                setSimpleName(right, left);
                setMethodName(right, left);
                TAA3EnhancedConstraintDataModel.setTypeVariableRange(left, ((TypeVariable2)right).getRange().getSourceRange());
            }

            final ITypeConstraint2 constraint = new EqualityTypeConstraint(left, right);
            final ITypeConstraint2 constraint2 = new EqualityTypeConstraint(right, left); 
            // TODO: hannes: why does equal or hashCode fail?
            if (!allConstraints.contains(constraint) && !allConstraints.contains(constraint2)) {
                allConstraints.add(constraint);
                setVariableUsage(left, constraint);
                setVariableUsage(right, constraint);

                createElementEqualsConstraints(left, right, false, createNewElementVars);
            }
        }
    }

    /** checks if a variable with this type should be used in constraints */
    private static boolean shouldBind(TType type) {
        if (type == null) {
            return true;
        }
        if (type.isWildcardType()) {
            type = ((WildcardType)type).getBound();
            if (type == null) {
                return false;
            }
        }
        if (type.isArrayType()) {
            type = ((ArrayType)type).getElementType();
        }
        return !type.isTypeVariable();
    }

    /**
     * Creates an exception variable or returns the existing one matching the parameter.
     * 
     * @param name the name of the thrown exception
     * @return the created exception variable
     */
    public final ConstraintVariable2 makeExceptionVariable(final Name name) {
        final ITypeBinding binding = name.resolveTypeBinding();
        if (isConstrainedType(binding)) {
            return constraintVariables.addOrGet(new TypeVariable2(createTType(binding), new CompilationUnitRange(RefactoringASTParser.getCompilationUnit(name), name)));
        }
        return null;
    }

    /**
     * Creates an independent type variable.
     * <p>
     * An independant type variable stands for an arbitrary type.
     * </p>
     * 
     * @param type the type binding
     * @return the created independant type variable
     */
    public final ConstraintVariable2 makeIndependentTypeVariable(ITypeBinding type) {
        if (type.isArray()) {
            type = type.getElementType();
        }
        if (isConstrainedType(type)) {
            return constraintVariables.addOrGet(new IndependentTypeVariable2(createTType(type)));
        }
        return null;
    }

    /**
     * Creates a subtype constraint.
     * 
     * @param descendant the descendant type constraint variable
     * @param ancestor the ancestor type constraint variable
     */
    public final void createSubtypeConstraint(ConstraintVariable2 descendant, ConstraintVariable2 ancestor) {
        createSubtypeConstraint(descendant, ancestor, true);
    }

    /**
     * Creates a subtype constraint.
     * 
     * @param descendant the descendant type constraint variable
     * @param ancestor the ancestor type constraint variable
     * @param createNewElementVars if new element variables should be generated for the first two parameters if they
     * are not existent yet
     */
    public void createSubtypeConstraint(ConstraintVariable2 descendant, ConstraintVariable2 ancestor, boolean createNewElementVars) {
        if (descendant == null || ancestor == null) {
            return;
        }
        if (!shouldBind(descendant.getType()) || !shouldBind(ancestor.getType())) {
            return;
        }
        
        descendant = constraintVariables.addOrGet(descendant);
        ancestor = constraintVariables.addOrGet(ancestor);

        final ITypeConstraint2 constraint = new SubTypeConstraint2(descendant, ancestor);
        //LogUtil.debugOut(this, "creating subtype constraint " + descendant + ", " + ancestor);
        if (!allConstraints.contains(constraint)) {
            allConstraints.add(constraint);
            setVariableUsage(descendant, constraint);
            setVariableUsage(ancestor, constraint);

            if (isElementFollowVarType(descendant) && isElementFollowVarType(ancestor)) {
                createElementEqualsConstraints(descendant, ancestor, true, createNewElementVars);
            }
        }
    }

    /**
     * Creates a new TType for the corresponding binding.
     * 
     * @param binding the type binding
     * @return the corresponding TType
     */
    public final TType createTType(final ITypeBinding binding) {
        final String key = binding.getKey();
        final TType cached = tTypeCache.get(key);
        if (cached != null) {
            return cached;
        }
        final TType type = typeEnvironment.create(binding);
        tTypeCache.put(key, type);
        return type;
    }

    /**
     * Creates a type variable or returns the existing one matching the parameters.
     * <p>
     * Type variables contain a source reference and represent the places where
     * the source code must be changed.
     * </p>
     * 
     * @param type the type binding
     * @param range the compilation unit range
     * @return the created type variable
     */
    public final TypeVariable2 makeTypeVariable(ITypeBinding type, final CompilationUnitRange range) {
        if (type.isArray()) {
            type = type.getElementType();
        }
        if (isConstrainedType(type)) {
            TType tType = createTType(type);
            TypeVariable2 ret = (TypeVariable2)constraintVariables.addOrGet(new TypeVariable2(tType, range));
            if (!hasElementVariables(ret)) {
                makeElementVariables(ret, tType, true);
            }
            return ret;
        }
        return null;
    }

    /**
     * Creates a constraint variable representing a local variable or a field or 
     * returns the existing one matching the binding.
     * 
     * @param variableBinding the binding
     * @return the created type constraint variable
     */
    public VariableVariable2 makeVariableVariable(IVariableBinding variableBinding) {
        if (variableBinding == null) {
            return null;
        }
        TType type = getBoxedType(variableBinding.getType(), /* no boxing */null);
        if (type == null) {
            return null;
        }
        
        VariableVariable2 cv = new VariableVariable2(type, variableBinding);
        VariableVariable2 storedCv = (VariableVariable2)constraintVariables.addOrGet(cv);

        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, type, true);
        }
        setSimpleName(storedCv, variableBinding.getName());
        IJavaElement variableElement = variableBinding.getJavaElement();
        if (variableElement != null) {
            TAA3EnhancedConstraintDataModel.setJavaElementBindingKey(storedCv, variableElement.getHandleIdentifier());
        }
        if (isTypeSelection() && selectedElement.isTypeSelection(variableBinding)) {
            completeTypeSelection.add(storedCv);            
        }
        return storedCv;
    }

    /**
     * Creates equality constraints between the corresponding element variables of 
     * the given constraint variables.
     * 
     * @param cv the (subtype) constraint variable
     * @param ancestorCv the (supertype) constraint variable
     * @param subtype if the relation between the variables is a subtype relation
     * @param createNew if new corresponding element variables should be created
     */
    public void createElementEqualsConstraints(ConstraintVariable2 cv, ConstraintVariable2 ancestorCv, boolean subtype, boolean createNew) {
        if (cv == null || ancestorCv == null) {
            return;
        }

        Set<String> handledKeys = new HashSet<String>();
        // LogUtil.debugOut2(this, "creating element equals for " + cv + ", " +
        // initializerCv);
        singleCreateElementEquals(cv, ancestorCv, handledKeys, subtype ? Subtype.SUBTYPE_FIRST : Subtype.SUBTYPE_NONE, createNew);
        singleCreateElementEquals(ancestorCv, cv, handledKeys, subtype ? Subtype.SUBTYPE_SECOND : Subtype.SUBTYPE_NONE, createNew);
    }

    private void singleCreateElementEquals(final ConstraintVariable2 baseCv, ConstraintVariable2 cv, Set<String> handledKeys, final Subtype subtype, final boolean createNew) {
        if (cv instanceof ElementVariable && cv.getType() == null) {
            final Set<ConstraintVariable2> variables = new HashSet<ConstraintVariable2>();
            try {
                // the connected constraints are visitied as long as element variables without types are encountered.
                ConstraintTraversal.visitAllConnectedConstraints(baseCv, new ConstraintTraversal.Visitor() {
                    public boolean visit(ConstraintVariable2 baseConstraintVariable, ConstraintVariable2 otherConstraintVariable2, ITypeConstraint2 constraint) throws JavaModelException {
                        if (variables.add(otherConstraintVariable2)) {
                            singleCreateElementEquals0(baseCv, otherConstraintVariable2, subtype, createNew);
                        }
                        return otherConstraintVariable2 instanceof ElementVariable && otherConstraintVariable2.getType() == null;
                    }

                });
            } catch (JavaModelException e) {
                e.printStackTrace();
                refactoringStatus.addFatalError(e.toString());                
            }
        }
        singleCreateElementEquals0(baseCv, cv, subtype, createNew);
    }

    private void singleCreateElementEquals0(ConstraintVariable2 baseCv, ConstraintVariable2 cv, Subtype subtype, boolean createNew) {
        Map elements = getElementVariables(baseCv);

        Map<ElementVariable, List<String>> varToKeys = new HashMap<ElementVariable, List<String>>();
        for (Iterator it = elements.entrySet().iterator(); it.hasNext();) {
            Map.Entry entry = (Map.Entry) it.next();
            ElementVariable elementVariable = (ElementVariable) entry.getValue();
            List<String> keys = varToKeys.get(elementVariable);
            if (keys == null) {
                keys = new ArrayList<String>();
                varToKeys.put(elementVariable, keys);
            }
            keys.add((String) entry.getKey());
        }

        for (Map.Entry<ElementVariable, List<String>> entry : varToKeys.entrySet()) {
            ElementVariable otherElementVariable = null;
            for (String currentKey : entry.getValue()) {
                otherElementVariable = getElementVariable(cv, currentKey);
                if (otherElementVariable != null) {
                    break;
                }
            }
            ElementVariable thisElementVariable = entry.getKey();
            TType cvType = cv.getType();
            if (cvType != null && cvType.isWildcardType()) {
                cvType = ((WildcardType)cvType).getBound();
            }
            if (createNew && otherElementVariable == null && (cvType != null && (cvType.isParameterizedType() || cvType.isGenericType()))) {
                //LogUtil.debugOut(this, "ATTENTION: CREATING NEW ELEMENT VARIABLE FOR " + thisElementVariable);
                otherElementVariable = makeElementVariable(cv, thisElementVariable);
            }
            if (otherElementVariable != null) {
                if (subtype == Subtype.SUBTYPE_NONE) {
                    createEqualityConstraint(thisElementVariable, otherElementVariable, createNew);
                } else {
                    ElementVariable descendent;
                    ElementVariable ancestor;
                    if (subtype == Subtype.SUBTYPE_FIRST) {
                        descendent = thisElementVariable;
                        ancestor = otherElementVariable;
                    } else {
                        descendent = otherElementVariable;
                        ancestor = thisElementVariable;
                    }
                    TType ancestorType = ancestor.getType();
                    if (ancestorType == null) {
                        createEqualityConstraint(thisElementVariable, otherElementVariable, createNew);
                    } else {
                        if (ancestorType.isWildcardType()) {
                            if (ancestorType.isSuperWildcardType()) {
                                createSubtypeConstraint(ancestor, descendent, createNew);
                            } else {
                                createSubtypeConstraint(descendent, ancestor, createNew);
                            }
                        } else {
                            createEqualityConstraint(thisElementVariable, otherElementVariable, createNew);
                        }
                    }

                }
            }
        }
    }

    private ElementVariable makeElementVariable(ConstraintVariable2 expressionCv, ElementVariable otherElementVariable) {
        if (expressionCv == null) {
            return null;
        }

        // LogUtil.debugOut2(this, "MAKE ELEMENT VAR for " + expressionCv);
        ElementVariable storedElementVariable = getElementVariable(expressionCv, otherElementVariable.getTypeVariableKey());
        if (storedElementVariable != null) {
            return storedElementVariable;
        }

        ElementVariable newCv = new ElementVariable(expressionCv, otherElementVariable);
        ElementVariable cv = (ElementVariable) constraintVariables.addOrGet(newCv);
        setElementVariable(expressionCv, cv, otherElementVariable.getTypeVariableKey());
        if (cv == newCv && isTypeSelection()) {
            ConstraintVariable2 base = cv.getBaseConstraintVariable();
            if (base instanceof ITypeConstraintVariable &&
                    selectedElement.isTypeSelection(cv.getType(), ((ITypeConstraintVariable)base).getCompilationUnit())) 
            {
                completeTypeSelection.add(cv);
            }

        }
        return cv;
    }

    private boolean isTypeSelection() {
        return selectedElement != null && selectedElement.isTypeSelection();
    }

    /**
     * Create equality constraints between generic type variables of
     * expressionCv and referenceCv. For example, the generic interface
     * <code>java.lang.Iterable&lt;E&gt;</code> defines a method
     * <code>Iterator&lt;E&gt; iterator()</code>. Given
     * <ul>
     * <li>an expressionCv of a subtype of <code>Iterable</code>,</li>
     * <li>a referenceCv of a subtype of <code>Iterator</code>, and</li>
     * <li>a reference binding of the Iterable#iterator()'s return type (the
     * parameterized type <code>Iterator&lt;E&gt;</code>),</li>
     * </ul>
     * this method creates an equality constraint between the type variable E in
     * expressionCV and the type variable E in referenceCV.
     * 
     * @param expressionCv
     *            the type constraint variable of an expression
     * @param methodTypeVariables
     * @param referenceCv
     *            the type constraint variable of a type reference
     * @param reference
     *            the declared type reference
     */
    public void createTypeVariablesEqualityConstraints(ConstraintVariable2 expressionCv, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, ConstraintVariable2 referenceCv, TType reference,
            TType currentType) {
        createTypeVariablesConstraints(expressionCv, methodTypeVariables, referenceCv, reference, currentType, ConstraintType.EQUALITY, true);
    }

    private void createTypeVariablesConstraints(ConstraintVariable2 expressionCv, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, ConstraintVariable2 referenceCv, TType reference,
            TType currentType, ConstraintType constraintType, boolean createNewSubVars) {
        reference = stripArray(reference);
        currentType = stripArray(currentType);
        
        if (reference.isParameterizedType() && currentType != null) {
            TType[] referenceTypeArguments = null;
            if (reference.isParameterizedType()) {
                referenceTypeArguments = ((ParameterizedType) reference).getTypeArguments();
            }
            TType[] referenceTypeParameters = ((GenericType) reference.getTypeDeclaration()).getTypeParameters();
            for (int i = 0; i < referenceTypeParameters.length; i++) {
                TypeVariable referenceTypeParameter = (TypeVariable) referenceTypeParameters[i];
                TType referenceTypeArgument;
                if (referenceTypeArguments == null) {
                    referenceTypeArgument = referenceTypeParameter.getErasure();
                } else {
                    referenceTypeArgument = referenceTypeArguments[i];
                }
                TType currentTypeArg = currentType.isParameterizedType() ? ((ParameterizedType) currentType).getTypeArguments()[i] : null;
                
                referenceTypeArgument = stripArray(referenceTypeArgument);
                currentTypeArg = stripArray(currentTypeArg);
                
                if (referenceTypeArgument.isTypeVariable()) {
                    ConstraintVariable2 referenceTypeArgumentCv = getElementTypeCv(referenceTypeArgument, currentTypeArg, expressionCv, methodTypeVariables, createNewSubVars);
                    ElementVariable referenceTypeParametersCv = makeElementVariable(referenceCv, referenceTypeParameter, referenceTypeArgument, true, createNewSubVars,
                            true);
                    // LogUtil.debugOut2(this, "TYPE VAR CONSTRAINT: " +
                    // referenceTypeArgumentCv + " =?= " +
                    // referenceTypeParametersCv + ", expr is " + expressionCv +
                    // ", arg is " + referenceTypeArgument + ", type: " +
                    // constraintType);
                    createConstraint(referenceTypeArgumentCv, referenceTypeParametersCv, constraintType, createNewSubVars);
                } else if (referenceTypeArgument.isWildcardType()) {
                    WildcardType wildcardType = (WildcardType) referenceTypeArgument;
                    if (wildcardType.isUnboundWildcardType()) {
                        ElementVariable referenceTypeParametersCv = makeElementVariable(referenceCv, referenceTypeParameter, referenceTypeArgument, true,
                                createNewSubVars, true);
                        ConstraintVariable2 referenceTypeArgumentCv = makeImmutableTypeVariable(wildcardType, "unbound wildcard", null, ImmutabilityReason.OTHER);
                        createEqualityConstraint(referenceTypeArgumentCv, referenceTypeParametersCv, createNewSubVars);
                    } else {                        
                        TType bound = stripArray(wildcardType.getBound());
                        
                        ElementVariable referenceTypeParametersCv = makeElementVariable(referenceCv, referenceTypeParameter, bound, true, createNewSubVars,
                                true);
                        // LogUtil.debugOut2(this, "referenceTypeParametersCv: " +
                        // referenceTypeParametersCv + ", type: " +
                        // referenceTypeParametersCv.getType());

                        if (bound.isParameterizedType()) {
                            createTypeVariablesConstraints(expressionCv, methodTypeVariables, referenceTypeParametersCv, referenceTypeArgument, bound, ConstraintType.EQUALITY, createNewSubVars);
                        } else if (bound.isTypeVariable()) {
                            try {
                                ConstraintVariable2 referenceTypeArgumentCv = getElementTypeCv(bound, currentTypeArg != null && currentTypeArg.isWildcardType() ? ((WildcardType) currentTypeArg).getBound() : currentTypeArg,
                                        expressionCv, methodTypeVariables, createNewSubVars);
                                // LogUtil.debugOut2(this, "TYPE VAR WILDCARD
                                // CONSTRAINT: " + referenceTypeArgumentCv + " =?= "
                                // + referenceTypeParametersCv + ", expr is " +
                                // expressionCv + ", arg is " +
                                // referenceTypeArgument + ", type: " +
                                // constraintType);
                                createConstraint(referenceTypeArgumentCv, referenceTypeParametersCv, ConstraintType.EQUALITY, createNewSubVars);
                            } catch (NullPointerException e) {
                                e.printStackTrace();
                            }
                        }
                    }

                } else if (referenceTypeArgument.isParameterizedType()) {
                    ElementVariable referenceTypeParametersCv = makeElementVariable(referenceCv, referenceTypeParameter, referenceTypeArgument, true, createNewSubVars,
                            true);
                    // LogUtil.debugOut2(this, "MAKE ELEMENT: " +
                    // referenceTypeParametersCv + " for " + referenceCv + ",
                    // expr is " + expressionCv);
                    createTypeVariablesConstraints(expressionCv, methodTypeVariables, referenceTypeParametersCv, referenceTypeArgument, currentTypeArg, ConstraintType.EQUALITY, createNewSubVars);
                }
            }
        }
    }

    private static TType stripArray(TType ttype) {
        if (ttype instanceof ArrayType) {
            ttype = ((ArrayType)ttype).getElementType();
        }
        return ttype;
    }

    private void createConstraint(ConstraintVariable2 referenceTypeArgumentCv, ElementVariable referenceTypeParametersCv, ConstraintType constraintType, boolean createNewSubVars) {
        switch (constraintType) {
        case EQUALITY:
            createEqualityConstraint(referenceTypeArgumentCv, referenceTypeParametersCv, createNewSubVars);
            break;
        case SUBTYPE:
            createSubtypeConstraint(referenceTypeArgumentCv, referenceTypeParametersCv, createNewSubVars);
            break;
        case SUPERTYPE:
            createSubtypeConstraint(referenceTypeParametersCv, referenceTypeArgumentCv, createNewSubVars);
            break;
        }
    }

    private ConstraintVariable2 getElementTypeCv(TType elementType, TType currentType, ConstraintVariable2 expressionCv, Map methodTypeVariables, boolean createNewSubVars) {
        if (elementType.isTypeVariable()) {
            ConstraintVariable2 elementTypeCv = (ConstraintVariable2) methodTypeVariables.get(elementType.getBindingKey());
            if (elementTypeCv != null) {
                return elementTypeCv;
            }
            if (expressionCv != null) {
                return makeElementVariable(expressionCv, (TypeVariable) elementType, currentType, true, createNewSubVars, true);
            }
        }
        return null;
    }

    /**
     * Retrieves a map between the type variable keys and the corresponding element 
     * variables of the given constraint variable.
     *  
     * @param constraintVariable the variable whose element variables should be returned 
     * @return the variable map. An empty map if there are no element variables.
     */
    @SuppressWarnings("unchecked")
    public static Map<String, ElementVariable> getElementVariables(ConstraintVariable2 constraintVariable) {
        if (constraintVariable == null) {
            return EMPTY_COLLECTION_ELEMENT_VARIABLES_MAP;            
        }
        Map<String, ElementVariable> elementVariables = (Map<String, ElementVariable>)constraintVariable.getData(DATA_ELEMENT_VARIABLES);
        if (elementVariables == null) {
            return EMPTY_COLLECTION_ELEMENT_VARIABLES_MAP;
        } else {
            return elementVariables;
        }
    }

    private BoundVariable makeBoundVariable(TypeVariable typeVariable, int index, TType bound, boolean mutable, boolean checkBounds) {
        BoundVariable cv;
        // TODO: IT4: support mutable bounds
        // if (mutable) {
        // cv = new BoundVariable(bound, typeVariable, index);
        // } else {
        cv = new ImmutableBoundVariable(bound, typeVariable, index);
        // }
        BoundVariable storedCv = (BoundVariable) constraintVariables.addOrGet(cv);
        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, storedCv, bound, false, true, true, false, checkBounds, new HashSet<TType>());
        }
        return storedCv;
    }

    /**
     * Creates a constraint variable representing a method parameter or returns the existing one
     * matching the parameters if no independent variable is requested.
     * <p>
     * For declarations or invocations with no type variables involved this is a single 
     * variable per parameter and used for matching invocations and declarations.
     * </p><p>
     * For invocations with type variables used in parameterized types, an independent variable
     * is created per invocation to connect the element variables of the argument and 
     * the receiver. 
     * </p>
     * @param methodBinding the binding
     * @param parameterIndex the index of the parameter
     * @param independent if true an IndependentParameterTypeVariable will be created, otherwise
     * the single ParameterTypeVariable2 for this parameter will be returned.
     * @return a ParameterTypeVariable2 or an IndependentParameterTypeVariable
     */
    public ConstraintVariable2 makeParameterTypeVariable(IMethodBinding methodBinding, int parameterIndex, boolean independent) {
        if (methodBinding == null) {
            return null;
        }
        TType type = getBoxedType(methodBinding.getParameterTypes()[parameterIndex], null);
        if (type == null) {
            return null;
        }
        ConstraintVariable2 cv;
        if (independent) {
            cv = new IndependentParameterTypeVariable(type, parameterIndex, methodBinding);
        } else {
            cv = new ParameterTypeVariable2(type, parameterIndex, methodBinding);
        }
        ConstraintVariable2 storedCv = constraintVariables.addOrGet(cv);

        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, type, true);
        }
        // TODO: hannes: check simple name for parameters
        setMethodName(storedCv, methodBinding.getName());

        return storedCv;
    }

    private static void setMethodName(ConstraintVariable2 cv, String name) {
        if (cv != null) {
            cv.setData(DATA_METHOD_NAME, name);
        }
    }

    private void makeElementVariables(ConstraintVariable2 expressionCv, TType type, boolean createVarsForTypeVars) {
        makeElementVariables(expressionCv, expressionCv, type, createVarsForTypeVars, true, true, true, true, new HashSet<TType>());
    }

    private void makeElementVariables(ConstraintVariable2 baseVar, ConstraintVariable2 expressionCv, TType type,
            boolean createVarsForTypeVars, boolean createNewSubVars, boolean mainRun, 
            boolean elementsMutable, boolean checkBounds,Set<TType> handledSupertypes) 
    {
        makeElementVariables(baseVar, expressionCv, type, type, createVarsForTypeVars, createNewSubVars, mainRun, 
                elementsMutable, checkBounds, handledSupertypes);
    }

    private void makeElementVariables(ConstraintVariable2 baseVar, ConstraintVariable2 expressionCv, TType type, TType declType,
            boolean createVarsForTypeVars, boolean createNewSubVars, boolean mainRun, 
            boolean elementsMutable, boolean checkBounds,Set<TType> handledSupertypes) 
    {
        if (declType instanceof WildcardType) {
            declType = ((WildcardType)declType).getBound();                
        }
        declType = stripArray(declType);

        if (type instanceof WildcardType) {
            type = ((WildcardType)type).getBound();                
        }
        type = stripArray(type);
        
        if (type == null) {
            return;
        }

        if (isAGenericType(type)) {
            GenericType genericType = (GenericType) type.getTypeDeclaration();
            TType[] typeParameters = genericType.getTypeParameters();
            TType[] typeArguments = null;
            TType[] declTypeArguments = null;
            if (declType instanceof ParameterizedType) {
                typeArguments = ((ParameterizedType) type).getTypeArguments();
                declTypeArguments = ((ParameterizedType)declType).getTypeArguments();
            }

            for (int i = 0; i < typeParameters.length; i++) {
                TypeVariable typeVariable = (TypeVariable) typeParameters[i];
                TType argument = typeArguments == null ? null : typeArguments[i];
                TType declArgument = declTypeArguments == null ? null : declTypeArguments[i];

                if (!createVarsForTypeVars && declArgument != null && declArgument.isTypeVariable()) {
                    if (baseVar != expressionCv) {
                        ElementVariable baseElementVar = getElementVariable(baseVar, declArgument.getBindingKey());
                        if (baseElementVar != null) {
                            setElementVariable(expressionCv, baseElementVar, typeVariable.getBindingKey());
                        }
                    }
                } else {
                    ElementVariable elementVariable = makeElementVariable(expressionCv, typeVariable, argument, createNewSubVars, createVarsForTypeVars, elementsMutable, checkBounds);
                    if (argument != null) {
                        makeElementVariables(baseVar, elementVariable, argument, declArgument, createVarsForTypeVars, createNewSubVars, false, elementsMutable, checkBounds, new HashSet<TType>(handledSupertypes));
                    }

                    if (baseVar != expressionCv && declArgument != null && declArgument.isTypeVariable()) {
                        ElementVariable baseElementVar = getElementVariable(baseVar, declArgument.getBindingKey());
                        if (baseElementVar != null) {
                            createEqualityConstraint(baseElementVar, elementVariable);
                        }
                    }
                }
            }
        }
        makeElementVariablesFromSupertypes(baseVar, expressionCv, type, createVarsForTypeVars, createNewSubVars, mainRun, elementsMutable, checkBounds, handledSupertypes);
    }

    private void makeElementVariablesFromSupertypes(ConstraintVariable2 baseVar, ConstraintVariable2 expressionCv, TType type, boolean createVarsForTypeVars, boolean createNewSubVars, boolean mainRun, boolean elementsMutable, boolean checkBounds, Set<TType> handledSupertypes) {
        TType typeDeclaration = type.getTypeDeclaration();
        TType declSuperclass = typeDeclaration.getSuperclass();
        TType superclass = type.getSuperclass();
        if (superclass != null) {
            makeSupertypeElementVariables(baseVar, expressionCv, superclass, declSuperclass, createVarsForTypeVars, createNewSubVars, mainRun, elementsMutable, checkBounds, handledSupertypes);
        }

        TType[] interfaces = type.getInterfaces();
        TType[] declInterfaces = typeDeclaration.getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            makeSupertypeElementVariables(baseVar, expressionCv, interfaces[i], declInterfaces[i], createVarsForTypeVars, createNewSubVars, mainRun, elementsMutable, checkBounds, handledSupertypes);
        }

    }

    private void makeSupertypeElementVariables(ConstraintVariable2 baseVar, ConstraintVariable2 expressionCv, TType supertype, TType declSupertype, boolean createVarsForTypeVars,
            boolean createNewSubVars, boolean mainRun, boolean elementsMutable, boolean checkBounds,
            Set<TType> handledSupertypes) {
        
        if (!handledSupertypes.add(supertype)) {
            return;
        }
        TType usedSuperType = supertype;
        if (createVarsForTypeVars || !mainRun) {
            usedSuperType = declSupertype;
            //supertype = declSupertype;
        }
        if (declSupertype.isParameterizedType() && supertype.isParameterizedType()) {
            TType[] typeArguments = ((ParameterizedType) declSupertype).getTypeArguments();
            TType[] paramTypeArguments = ((ParameterizedType)supertype).getTypeArguments();
            TypeVariable[] typeParameters = ((GenericType)usedSuperType.getTypeDeclaration()).getTypeParameters();
            for (int i = 0; i < typeParameters.length; i++) {
                TypeVariable typeParameter = typeParameters[i];
                TType referenceTypeArgument = typeArguments[i];
                TType referenceParamTypeArgument = paramTypeArguments[i];
                if (referenceTypeArgument.isTypeVariable()) {
                    if (createVarsForTypeVars || mainRun) {
                        ElementVariable referenceTypeArgumentCv = makeElementVariable(expressionCv, (TypeVariable) referenceTypeArgument, 
                                mainRun ? (TypeVariable)referenceTypeArgument : null, createVarsForTypeVars,
                                createNewSubVars, elementsMutable, checkBounds);
                        setElementVariable(expressionCv, referenceTypeArgumentCv, typeParameter.getBindingKey());
                    }
                } else {
                    ElementVariable elementVariable = makeElementVariable(expressionCv, typeParameter, referenceParamTypeArgument, createVarsForTypeVars, createNewSubVars,
                            elementsMutable, checkBounds);
                    if (referenceTypeArgument.isParameterizedType()) {
                        makeElementVariables(baseVar, elementVariable, referenceParamTypeArgument, referenceTypeArgument, createVarsForTypeVars, createNewSubVars, false, elementsMutable, checkBounds, new HashSet<TType>(handledSupertypes));
                    }
                }
            }
        }
        makeElementVariablesFromSupertypes(baseVar, expressionCv, usedSuperType, createVarsForTypeVars, createNewSubVars, false, elementsMutable, checkBounds, handledSupertypes);
    }

    /**
     * Creates a return type variable or returns the existing one 
     * matching the parameters if no independent variable is requested.
     * 
     * @param methodBinding the binding of the invoked method
     * @param independent if true, an independent variable will be created. Otherwise the
     * single return type variable for this method binding will be returned 
     * @return a ReturnTypeVariable2 or an IndependentReturnTypeVariable
     */
    public ConstraintVariable2 makeReturnTypeVariable(IMethodBinding methodBinding, boolean independent) {
        if (methodBinding == null) {
            return null;
        }

        TType declaredReturnType = getBoxedType(methodBinding.getMethodDeclaration().getReturnType(), null);

        if (declaredReturnType == null) {
            return null;
        }

        ConstraintVariable2 cv = new ReturnTypeVariable2(declaredReturnType.getErasure(), methodBinding.getMethodDeclaration());
        ConstraintVariable2 storedCv = constraintVariables.addOrGet(cv);

        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, declaredReturnType, true);
        }

        if (independent) {
            ConstraintVariable2 declaredCv = storedCv;

            TType returnType = getBoxedType(methodBinding.getReturnType(), null);
            cv = new IndependentReturnTypeVariable(returnType, methodBinding);
            storedCv = constraintVariables.addOrGet(cv);

            if (!hasElementVariables(storedCv)) {
                makeElementVariables(storedCv, returnType, true);
            }

            createEqualityConstraint(declaredCv, cv);
        }

        setMethodName(storedCv, methodBinding.getName());
        TAA3EnhancedConstraintDataModel.setJavaElementBindingKey(storedCv, methodBinding.getJavaElement().getHandleIdentifier());
        
        if (isTypeSelection() && selectedElement.isTypeSelection(methodBinding)) {
            completeTypeSelection.add(storedCv);            
        }        

        return storedCv;
    }

    public ConstraintVariable2 makeThisVariable(ITypeBinding binding) {
        return constraintVariables.addOrGet(new ThisVariable(createTType(binding)));
	}
    
	public ConstraintVariable2 makeStaticVariable(ITypeBinding binding) {
        return constraintVariables.addOrGet(new StaticVariable(createTType(binding)));
	}    
    
	public ConstraintVariable2 makeSuperVariable(ITypeBinding binding) {
        return constraintVariables.addOrGet(new SuperVariable(createTType(binding)));
	}    

    /**
     * Creates an independent variable for a field access to a generic class. 
     * 
     * @param variableBinding the binding of the field
     * @return an IndependentTypeVariable2
     */
    public ConstraintVariable2 makeIndependentFieldVariable(IVariableBinding variableBinding) {
        if (variableBinding == null) {
            return null;
        }

        TType declaredType = getBoxedType(variableBinding.getVariableDeclaration().getType(), null);

        if (declaredType == null) {
            return null;
        }

        ConstraintVariable2 cv = new VariableVariable2(declaredType.getErasure(), variableBinding.getVariableDeclaration());
        ConstraintVariable2 declaredCv = constraintVariables.addOrGet(cv);

        if (!hasElementVariables(declaredCv)) {
            makeElementVariables(declaredCv, declaredType, true);
        }

        TType type = getBoxedType(variableBinding.getType(), null);
        cv = new IndependentTypeVariable2(type);
        ConstraintVariable2 storedCv = constraintVariables.addOrGet(cv);

        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, type, true);
        }

        createEqualityConstraint(declaredCv, cv);

        return storedCv;
    }

    private static boolean isAGenericType(TType type) {
        return type.isGenericType() || type.isParameterizedType();
    }

    /**
     * Creates an immutable type variable that holds a description why it is immutable.
     * 
     * @param typeBinding the type
     * @param expression the expression the type is from
     * @param description the reason why the type is immutable
     * @return the constraint variable
     */
    public ImmutableTypeVariableWithDescr makeImmutableTypeVariable(ITypeBinding typeBinding, Expression expression, String description, String bindingKey, ImmutabilityReason immutabilityReason) {
        TType type = getBoxedType(typeBinding, expression);
        if (type == null) {
            return null;
        }
        return makeImmutableTypeVariable(type, description, bindingKey, immutabilityReason);
    }

    /**
     * Creates an immutable type variable that holds a description why it is immutable.
     * 
     * @param type the type 
     * @param description the reason why the type is immutable
     * @return the constraint variable
     */
    public ImmutableTypeVariableWithDescr makeImmutableTypeVariable(TType type, String description, String bindingKey, ImmutabilityReason immutabilityReason) {
        if (type.isTypeVariable() || (type.isWildcardType() && (((WildcardType) type).getBound() == null || ((WildcardType) type).getBound().isTypeVariable()))) {
            return null;
        }

        // LogUtil.debugOut2(this, "immutable type " + type);
        ImmutableTypeVariableWithDescr cv = new ImmutableTypeVariableWithDescr(type, description, bindingKey, immutabilityReason);
        ImmutableTypeVariableWithDescr storedCv = (ImmutableTypeVariableWithDescr) constraintVariables.addOrGet(cv);
        if (cv == storedCv) {
            makeElementVariables(storedCv, storedCv, type, false, true, true, true, true, new HashSet<TType>());
        }
        return storedCv;
    }
    
    /**
     * Makes the element variables of this variable immutable.
     * 
     * @param baseVar the parent variable
     * @param description the reason why the elements are immutable
     */
    public void makeElementsImmutable(ConstraintVariable2 baseVar, String description) {
        Set<ElementVariable> processed = new HashSet<ElementVariable>();
        for(ElementVariable elementVariable : getElementVariables(baseVar).values()) {
            if (processed.add(elementVariable)) {
                TType type = elementVariable.getType();
                if (type instanceof WildcardType) {
                    type = ((WildcardType)type).getBound();
                }
                if (type != null) {
                    type = stripArray(type);
                    if (!type.isTypeVariable()) {
                        ConstraintVariable2 immutableVar = makeImmutableTypeVariable(type, description, null, ImmutabilityReason.OTHER);
                        if (immutableVar != null) {
                            createSubtypeConstraint(elementVariable, immutableVar);
                        }
                    }
                    if (type.isParameterizedType()) {
                        makeElementsImmutable(elementVariable, description);
                    }
                }
            }
        }
    }

    

    /**
     * Creates an element variable or returns the existing one.
     * 
     * @param parentTypeCv the parent variable of this element
     * @param parameterType the type parameter
     * @param ttype the current argument type
     * @param createNewSubVars if true additional sub vars will be created for nested parameterized types
     * @return the element variable
     */
    public ElementVariable makeElementVariable(ConstraintVariable2 parentTypeCv, ITypeBinding parameterType, TType ttype, boolean createNewSubVars) {
        return makeElementVariable(parentTypeCv, (TypeVariable) createTType(parameterType), ttype, true, createNewSubVars, true);
    }

    private ElementVariable makeElementVariable(ConstraintVariable2 expressionCv, TypeVariable typeVariable, TType ttype, boolean createVarsForTypeVars, boolean createNewSubVars,
            boolean mutable) 
    {
        return makeElementVariable(expressionCv, typeVariable, ttype, createVarsForTypeVars, createNewSubVars, mutable, true);
    }

    private ElementVariable makeElementVariable(ConstraintVariable2 expressionCv, TypeVariable typeVariable, TType ttype, boolean createVarsForTypeVars, boolean createNewSubVars,
            boolean mutable, boolean checkBounds) 
    {
        if (expressionCv == null) {
            return null;
        }

        ElementVariable storedElementVariable = getElementVariable(expressionCv, typeVariable.getBindingKey());
        if (storedElementVariable != null) {
            return storedElementVariable;
        }

        if (ttype instanceof ArrayType) {
            ttype = ((ArrayType)ttype).getElementType();
        }
        
        ElementVariable newCv;
        if (mutable) {
            newCv = new ElementVariable(expressionCv, typeVariable, ttype);
        } else {
            newCv = new ImmutableElementVariable(expressionCv, typeVariable, ttype);
        }
        ElementVariable cv = (ElementVariable) constraintVariables.addOrGet(newCv);
        if (cv == newCv && isTypeSelection()) {
            ConstraintVariable2 base = cv.getBaseConstraintVariable();
            if (base instanceof ITypeConstraintVariable &&
                    selectedElement.isTypeSelection(cv.getType(), ((ITypeConstraintVariable)base).getCompilationUnit())) 
            {
                completeTypeSelection.add(cv);
            }

        }
        

        if (checkBounds) {
            TType[] bounds = typeVariable.getBounds();
            for (int i = 0; i < bounds.length; i++) {
                // TODO: IT4: support mutable bounds
                TType boundType = bounds[i];
                if (!boundType.isJavaLangObject()) {
                    if (!boundType.isTypeVariable()) {
                        BoundVariable boundVariable = makeBoundVariable(typeVariable, i, boundType, false, false);
                        createSubtypeConstraint(cv, boundVariable, true);
                    }
                    checkBoundDependencies(cv, boundType, expressionCv, null, createVarsForTypeVars, createNewSubVars, mutable, true);
                }
            }
        }
        setElementVariable(expressionCv, cv, typeVariable.getBindingKey());
        return cv;
    }

    private void checkBoundDependencies(ConstraintVariable2 cv, TType boundType, ConstraintVariable2 expressionCv, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, boolean createVarsForTypeVars, boolean createNewSubVars, boolean mutable, boolean topLevel) {
        checkBoundDependencies0(new HashSet<String>(), cv, boundType, expressionCv, methodTypeVariables, createVarsForTypeVars, createNewSubVars, mutable, topLevel);        
    }
    
    private void checkBoundDependencies0(Set<String> handledVars, ConstraintVariable2 cv, TType boundType, ConstraintVariable2 expressionCv, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, boolean createVarsForTypeVars, boolean createNewSubVars, boolean mutable, boolean topLevel) {
        if (cv == null) {
            return;
        }
        
        if (methodTypeVariables == null) {
            methodTypeVariables = Collections.emptyMap();
        }

        if (boundType.isParameterizedType()) {
            TType[] parameters = ((GenericType)boundType.getTypeDeclaration()).getTypeParameters();
            TType[] args = ((ParameterizedType)boundType).getTypeArguments();
            for (int i=0; i<args.length; i++) {
                ElementVariable paramVar = makeElementVariable(cv, (TypeVariable)parameters[i], null, 
                        createVarsForTypeVars, 
                        createNewSubVars, 
                        mutable,
                        false);
                String bindingKey = parameters[i].getBindingKey();
                setElementVariable(cv, paramVar, bindingKey);
                if (!handledVars.contains(bindingKey)) {
                    handledVars.add(bindingKey);
                    checkBoundDependencies0(handledVars, paramVar, args[i], expressionCv, methodTypeVariables, createVarsForTypeVars, createNewSubVars, mutable, false);
                }
            }
        } else if (boundType.isTypeVariable()) {
            ConstraintVariable2 boundVar = methodTypeVariables.get(boundType.getBindingKey());
            if (boundVar == null && expressionCv != null) {           
                boundVar = makeElementVariable(expressionCv, (TypeVariable)boundType, null, 
                        createVarsForTypeVars, 
                        createNewSubVars, 
                        mutable,
                        false);
            }
            if (topLevel) {
                createSubtypeConstraint(cv, boundVar);
            } else {
                createEqualityConstraint(cv, boundVar);                
            }
        }
        
    }

    /**
     * Creates a type variable or returns the existing one matching the parameter.
     * 
     * @param type the type from the ast
     * @param createChildrenForTypeVars if true, element variables will be also created
     * for arguments that are themselves type variables
     * @return the constraint variable
     */
    public TypeVariable2 makeTypeVariable(Type type, boolean createChildrenForTypeVars) {
        ICompilationUnit cu = RefactoringASTParser.getCompilationUnit(type);
        TType ttype = getBoxedType(type.resolveBinding(), /* no boxing */null);
        if (ttype == null) {
            return null;
        }

        CompilationUnitRange range = new CompilationUnitRange(cu, type);
        TypeVariable2 typeVariable = new TypeVariable2(ttype, range);
        TypeVariable2 storedCv = (TypeVariable2)constraintVariables.addOrGet(typeVariable);

        if (!hasElementVariables(storedCv)) {
            makeElementVariables(storedCv, ttype, createChildrenForTypeVars);
        }

        return storedCv;
    }

    /**
     * Check if element constraints should be created for the given variable type.
     */
    private boolean isElementFollowVarType(ConstraintVariable2 var) {
        return !(var instanceof DeclaringTypeVariable || 
                var instanceof ImmutableTypeVariableWithDescr);
    }

    @SuppressWarnings("unchecked")
    private void setElementVariable(ConstraintVariable2 typeConstraintVariable, final ElementVariable elementVariable, final String typeVariableKey) {
        Map<String, ElementVariable> keyToElementVar = (Map<String, ElementVariable>)typeConstraintVariable.getData(DATA_ELEMENT_VARIABLES);
        if (keyToElementVar == null) {
            keyToElementVar = new HashMap<String, ElementVariable>();
            typeConstraintVariable.setData(DATA_ELEMENT_VARIABLES, keyToElementVar);
        } else {
            ElementVariable existingElementVar = keyToElementVar.get(typeVariableKey);
            if (existingElementVar != null) {
                if (existingElementVar != elementVariable) {
                    // this can happen with bounds
                    createEqualityConstraint(existingElementVar, elementVariable);
                }
            }
        }
        keyToElementVar.put(typeVariableKey, elementVariable);
    }

    /**
     * Checks whether this variable has element variable children.
     * 
     * @param constraintVariable the variable to check
     * @return true if there are children
     */
    public boolean hasElementVariables(ConstraintVariable2 constraintVariable) {
        return (HashMap) constraintVariable.getData(DATA_ELEMENT_VARIABLES) != null;
    }

    private ElementVariable getElementVariable(ConstraintVariable2 constraintVariable, String typeVariableKey) {
        HashMap typeVarToElementVars = (HashMap) constraintVariable.getData(DATA_ELEMENT_VARIABLES);
        if (typeVarToElementVars == null) {
            return null;
        }
        return (ElementVariable) typeVarToElementVars.get(typeVariableKey);
    }

    private TType getBoxedType(ITypeBinding typeBinding, Expression expression) {
        if (typeBinding == null) {
            return null;
        }
        if (typeBinding.isArray()) {
            typeBinding = typeBinding.getElementType();
        }

        if (!typeBinding.isPrimitive()) {
            return createTType(typeBinding);
        }

        if (expression == null || !expression.resolveBoxing()) {
            return null;
        }

        String primitiveName = typeBinding.getName();
        ITypeBinding boxed = expression.getAST().resolveWellKnownType(getBoxedTypeName(primitiveName));
        return createTType(boxed);
    }

    private String getBoxedTypeName(String primitiveName) {
        if ("long".equals(primitiveName)) {
            return "java.lang.Long"; //$NON-NLS-1$
        } else if ("int".equals(primitiveName)) {
            return "java.lang.Integer"; //$NON-NLS-1$
        } else if ("short".equals(primitiveName)) {
            return "java.lang.Short"; //$NON-NLS-1$
        } else if ("char".equals(primitiveName)) {
            return "java.lang.Character"; //$NON-NLS-1$
        } else if ("byte".equals(primitiveName)) {
            return "java.lang.Byte"; //$NON-NLS-1$
        } else if ("boolean".equals(primitiveName)) {
            return "java.lang.Boolean"; //$NON-NLS-1$
        } else if ("float".equals(primitiveName)) {
            return "java.lang.Float"; //$NON-NLS-1$
        } else if ("double".equals(primitiveName)) {
            return "java.lang.Double"; //$NON-NLS-1$
        } else {
            return null;
        }
    }

    /**
     * Gets called when the creation of the model ends.
     */
    public final void endCreation() {
        // create equality constraints for the import selection.
        if (completeTypeSelection.size() > 0) {
            selectedVariable = completeTypeSelection.get(0);
            for (int i=1; i<completeTypeSelection.size(); i++) {
                createEqualityConstraint(selectedVariable, completeTypeSelection.get(i));
            }
        }
    }

    /**
     * Returns the constraint variables of this model.
     * 
     * @return the constraint variables 
     */
    public final Collection<ConstraintVariable2> getConstraintVariables() {
        return Collections.unmodifiableCollection(constraintVariables);
    }

    /**
     * Returns the type constraints of this model.
     * 
     * @return the type constraints
     */
    public final Collection<ITypeConstraint2> getTypeConstraints() {
        return Collections.unmodifiableCollection(allConstraints);
    }

    /**
     * Creates an independent type variable for use with generic methods. Bound dependecies are
     * handled by this method automatically.
     *  
     * @param argument the current argument type
     * @param methodTypeVariables the type variables of this method
     * @param receiverCv the variable of the receiver of this method
     * @param methodTypeParameter the type parameter
     * @return the new constraint variable
     */
    public GenericMethodTypeArgumentVariable makeGenericMethodTypeArgumentVariable(TType argument, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, ConstraintVariable2 receiverCv, ITypeBinding methodTypeParameter) {
        GenericMethodTypeArgumentVariable cv = new GenericMethodTypeArgumentVariable(argument);
        GenericMethodTypeArgumentVariable storedCv = (GenericMethodTypeArgumentVariable)constraintVariables.addOrGet(cv);
        if (cv == storedCv) {
            ITypeBinding[] bounds = methodTypeParameter.getTypeBounds();
            for (int i = 0; i < bounds.length; i++) {
                // TODO: IT4: support mutable bounds
                ITypeBinding boundType = bounds[i];
                if (!boundType.getQualifiedName().equals("java.lang.Object")) {
                    TType boundTType = createTType(boundType);
                    if (!boundType.isTypeVariable()) {
                        BoundVariable boundVariable = makeBoundVariable((TypeVariable)createTType(methodTypeParameter), i, boundTType, false, true);
                        createSubtypeConstraint(cv, boundVariable, true);
                    }
                    checkBoundDependencies(cv, boundTType, receiverCv, methodTypeVariables, true, true, false, true);
                }
            }
            
            
        }
        return storedCv;
    }

    /**
     * Returns all constraints that the specified constraint variable is used in.
     * 
     * @param variable the constraint variable
     * @return the usage of the constraint variable 
     */
    @SuppressWarnings("unchecked")
    public static Collection<ITypeConstraint2> getVariableUsage(final ConstraintVariable2 variable) {
        final Object data = variable.getData(DATA_USAGE);
        if (data == null) {
            return Collections.emptyList();
        } else if (data instanceof Collection) {
            return Collections.unmodifiableCollection((Collection<ITypeConstraint2>) data);
        } else {
            return Collections.singletonList((ITypeConstraint2) data);
        }
    }

    /**
     * Assigns the simple name of the second variable to the first one. Used for
     * the type name generator.
     * 
     * @param changeVar the variable to be changed
     * @param baseVariable the variable that holds the name
     */
    public static void setSimpleName(ConstraintVariable2 changeVar, ConstraintVariable2 baseVariable) {
        if (baseVariable != null && changeVar != null) {
            changeVar.setData(DATA_SIMPLE_NAME, baseVariable.getData(DATA_SIMPLE_NAME));
        }
    }
    
    /**
     * Assigns the method name of the second variable to the first one. Used for 
     * the type name generator.
     *
     * @param changeVar the variable to be changed
     * @param baseVariable the variable that holds the name
     */
    private static void setMethodName(ConstraintVariable2 changeVar, ConstraintVariable2 baseVariable) {
        if (baseVariable != null && changeVar != null) {
            changeVar.setData(DATA_METHOD_NAME, baseVariable.getData(DATA_METHOD_NAME));
        }
    }    

    /**
     * Assigns a simple name to this variable. Used for 
     * the type name generator.
     * 
     * @param var the variable
     * @param name the simple name
     */
    public static void setSimpleName(ConstraintVariable2 var, String name) {
        var.setData(DATA_SIMPLE_NAME, name);
    }

    /**
     * Retrieves the simple name for this variable.
     * 
     * @param variable the variable
     * @return the simple name or null if there is none assigned
     */
    public static String getSimpleName(final ConstraintVariable2 variable) {
        String val = (String) variable.getData(DATA_SIMPLE_NAME);
        if (val != null) {
            int arrayStart = val.indexOf('[');
            if (arrayStart > -1) {
                return val.substring(0, arrayStart);
            }
        }
        return val;
    }

    /**
     * Retrieves the method name for this variable.
     * 
     * @param variable the variable
     * @return the method name or null if there is none assigned
     */
    public static String getMethodName(final ConstraintVariable2 variable) {
        return (String) variable.getData(DATA_METHOD_NAME);
    }

    /**
     * Is the type represented by the specified binding a constrained type?
     * 
     * @param binding the binding to check, or <code>null</code>
     * @return <code>true</code> if it is constrained, <code>false</code>
     *         otherwise
     */
    private static boolean isConstrainedType(final ITypeBinding binding) {
        return binding != null && !binding.isSynthetic() && !binding.isPrimitive();
    }

    /**
     * Sets the usage of the specified constraint variable.
     * 
     * @param variable the constraint variable
     * @param constraint the type constraint
     */
    @SuppressWarnings("unchecked")
    private static void setVariableUsage(final ConstraintVariable2 variable, ITypeConstraint2 constraint) {
        final Object data = variable.getData(DATA_USAGE);
        if (data == null) {
            variable.setData(DATA_USAGE, constraint);
        } else if (data instanceof Collection) {
            ((Collection<ITypeConstraint2>)data).add(constraint);
        } else {
            final Collection<ITypeConstraint2> usage = new ArrayList<ITypeConstraint2>(2);
            usage.add((ITypeConstraint2) data);
            usage.add(constraint);
            variable.setData(DATA_USAGE, usage);
        }
    }

    /**
     * Retrieves the new type for this variable that was assigned during type calculation.
     * 
     * @param variable the variable
     * @return the new type or null if there is none assigned
     */
    public static TCType getNewType(ConstraintVariable2 variable) {
        if (variable == null) {
            return null;
        }
        
        return (TCType)variable.getData(InferTypeConstraintsModel.DATA_NEW_TYPE);
    }

    /**
     * Sets the new type for this variable during type calculation.
     * 
     * @param var the variable
     * @param newType the new type
     */
    public static void setNewType(ConstraintVariable2 var, TCType newType) {
        if (var != null) {       
            var.setData(InferTypeConstraintsModel.DATA_NEW_TYPE, newType);
        }
    }
    
    /**
     * Retrieves the access set for this variable that was assigned during access set calculation.
     * 
     * @param variable the variable
     * @return the access set or null if there is none assigned
     */
    public static AccessSet getAccessSet(ConstraintVariable2 variable) {
        if (variable == null) {
            return null;
        }
        
        return (AccessSet)variable.getData(InferTypeConstraintsModel.DATA_ACCESS_SET);
    }

    /**
     * Sets the access set for this variable during access set calculation.
     * 
     * @param var the variable
     * @param accessSet the access set
     */
    public static void setAccessSet(ConstraintVariable2 var, AccessSet accessSet) {
        if (var != null) {       
            var.setData(InferTypeConstraintsModel.DATA_ACCESS_SET, accessSet);
        }
    }

    /**
     * Retrieves the current refactoring status and uses a new one afterwards.
     * @return the current status
     */
    public RefactoringStatus getAndClearRefactoringStatus() {
        RefactoringStatus ret = refactoringStatus;
        refactoringStatus = new RefactoringStatus();
        return ret;
    }
    
    /**
     * Retrieves the selected element
     * @return the selectedElement
     */
    public SelectedElement getSelectedElement() {
        return selectedElement;
    }
    
    private enum Subtype {
        SUBTYPE_FIRST, SUBTYPE_SECOND, SUBTYPE_NONE;
    }

    private enum ConstraintType {
        EQUALITY, SUBTYPE, SUPERTYPE;
    }

}