/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.creation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.ParameterizedType;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.search.SearchMatch;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringSearchEngine2;
import org.eclipse.jdt.internal.corext.refactoring.SearchResultGroup;
import org.eclipse.jdt.internal.corext.refactoring.util.RefactoringASTParser;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;

/**
 * Helper class for the retrieval of affected compilation units.
 * Used exclusively by the {@link Collector}.
 * 
 * @author Hannes Kegel
 *
 */
class CollectorUtil {

    /**
     * Extracts the cus from the search result groups avoiding previously parsed units.
     * @param srGroups the search results 
     * @param processed the previously processed units
     * @return a set of compilation units
     */
    public static Set<ICompilationUnit> getUnitsToBeParsed(Collection<SearchResultGroup> srGroups, Set<ICompilationUnit> processed) {
        Set<ICompilationUnit> unitsToBeParsed = new HashSet<ICompilationUnit>(srGroups.size());
    
        for (SearchResultGroup group : srGroups) {
            SearchMatch[] matches = group.getSearchResults();
            for (SearchMatch element2 : matches) {
                Object element = element2.getElement();
                if (element instanceof IMember) {
                    ICompilationUnit current = ((IMember) element).getCompilationUnit();
                    if (current != null && !processed.contains(current)) {
                        unitsToBeParsed.add(current);                                    
                    }
                }
            }
        }
        return unitsToBeParsed;
    }

    /**
     * Maps cus to search result groups.
     * 
     * @param projectToGroups the search result groups
     * @return the map
     */
    public static Map<ICompilationUnit, SearchResultGroup> getCompUnitToGroups(final Map<IJavaProject, Collection<SearchResultGroup>> projectToGroups) {
        final Map<ICompilationUnit, SearchResultGroup> groups = new HashMap<ICompilationUnit, SearchResultGroup>();
        for (Collection<SearchResultGroup> srgs : projectToGroups.values()) {
            if (srgs != null) {
                for (SearchResultGroup group : srgs) {
                    SearchMatch[] matches = group.getSearchResults();
                    for (SearchMatch element2 : matches) {
                        Object element = element2.getElement();
                        if (element instanceof IMember) {
                            ICompilationUnit current = ((IMember) element).getCompilationUnit();
                            if (current != null) {
                                groups.put(current, group);
                            }
                        }
                    }
                }
            }
        }
        return groups;
    }
    
    /**
     * Retrieves a list of ITypes from the given hierarchy.
     * 
     * @param typeHierarchy the hierarchy whose ITypes should be returned
     * @return all types as ITypes from this hierarchy
     */
    public static Collection<IType> getITypesFromHierarchy(TypeHierarchy typeHierarchy) {
        List<IType> ret = new ArrayList<IType>();

        for (TCType tcType : typeHierarchy.getAllTypes()) {
            if (tcType instanceof ExistingTCType) {
                IType iType = ((ExistingTCType) tcType).getIType();
                ret.add(iType);
            }
        }

        return ret;
    }
    
    /**
     * Adds the current type and its subtype to the first argument if
     * previously searched types are used in parameterized types in the
     * extends or implements clause.
     * 
     * @param typesUsingOtherInParametricDeclaration the result set 
     * @param nodes the previously located nodes that should be checked
     * @throws JavaModelException
     */
    public static void getDeclarationReferencingCompilationUnits(Set<IType> typesUsingOtherInParametricDeclaration, ASTNode[] nodes) throws JavaModelException {
        for (ASTNode node : nodes) {
            IJavaProject project = RefactoringASTParser.getCompilationUnit(node).getJavaProject();
            if (project != null) {
                IType declaringType = getReferencingType(node);
                if (declaringType != null && typesUsingOtherInParametricDeclaration.add(declaringType)) {

                    ITypeHierarchy subHierarchy = Cache.getHierarchy(null, declaringType, false);
                    IType[] subTypes = subHierarchy.getAllSubtypes(declaringType);
                    typesUsingOtherInParametricDeclaration.addAll(Arrays.asList(subTypes));                        
                }
            }
        }
    }
    

    /**
     * Computes the compilation units of fields referencing the specified type
     * occurrences.
     * 
     * @param units
     *            the compilation unit map (element type:
     *            <code>&ltIJavaProject, Set&ltICompilationUnit&gt&gt</code>)
     * @param nodes
     *            the ast nodes representing the type occurrences
     * @throws JavaModelException
     *             if an error occurs
     */
    public static void getFieldReferencingCompilationUnits(final Map<IJavaProject, Set<ICompilationUnit>> units, final ASTNode[] nodes, RefactoringSearchEngine2 engine) throws JavaModelException {
        IJavaProject project = null;
        for (ASTNode node : nodes) {
            project = RefactoringASTParser.getCompilationUnit(node).getJavaProject();
            if (project != null) {
                final List<IField> fields = getReferencingFields(node, project);
                for (int offset = 0; offset < fields.size(); offset++) {
                    IField field = fields.get(offset);
                    
                    Set<ICompilationUnit> set = units.get(project);
                    if (set == null) {
                        set = new HashSet<ICompilationUnit>();
                        units.put(project, set);
                    }
                    final ICompilationUnit unit = field.getCompilationUnit();
                    if (unit != null) {
                        set.add(unit);
                    }

                    if (!Flags.isPrivate(field.getFlags())) {
                        findReferences(units, field, engine);
                    }
                }
            }
        }
    }

    
    /**
     * Computes the compilation units of methods referencing the specified type
     * occurrences.
     * 
     * @param units
     *            the compilation unit map (element type:
     *            <code>&ltIJavaProject, Set&ltICompilationUnit&gt&gt</code>)
     * @param nodes
     *            the ast nodes representing the type occurrences
     * @throws JavaModelException
     *             if an error occurs
     */
    public static void getMethodReferencingCompilationUnits(final Map<IJavaProject, Set<ICompilationUnit>> units, final ASTNode[] nodes, RefactoringSearchEngine2 engine) throws JavaModelException {
        for (ASTNode node : nodes) {
            IJavaProject project = RefactoringASTParser.getCompilationUnit(node).getJavaProject();
            if (project != null) {
                IMethod method = getReferencingMethod(node);
                if (method != null) {
                    //LogUtil.debugOut(this, "REFERENCING METHOD " + method.getKey());
                    Set<ICompilationUnit> set = units.get(project);
                    if (set == null) {
                        set = new HashSet<ICompilationUnit>();
                        units.put(project, set);
                    }
                    final ICompilationUnit unit = method.getCompilationUnit();
                    if (unit != null) {
                        set.add(unit);
                    }
                    findReferences(units, method, engine);
                }
            }
        }
    }

    
    
    /**
     * Returns the fields which reference the specified ast node.
     * 
     * @param node
     *            the ast node
     * @param project
     *            the java project
     * @return the referencing fields
     * @throws JavaModelException
     *             if an error occurs
     */
    private static List<IField> getReferencingFields(final ASTNode node, final IJavaProject project) throws JavaModelException {
        List<IField> result = Collections.emptyList();
        if (node instanceof Type) {
            final BodyDeclaration parent = (BodyDeclaration) ASTNodes.getParent(node, BodyDeclaration.class);
            if (parent instanceof FieldDeclaration) {
                final List fragments = ((FieldDeclaration) parent).fragments();
                result = new ArrayList<IField>(fragments.size());
                VariableDeclarationFragment fragment = null;
                for (final Iterator iterator = fragments.iterator(); iterator.hasNext();) {
                    fragment = (VariableDeclarationFragment) iterator.next();
                    final IField field = getCorrespondingField(fragment);
                    if (field != null) {
                        result.add(field);
                    }
                }
            }
        }
        return result;
    }
    
    private static IType getReferencingType(final ASTNode node) throws JavaModelException {
        if (node instanceof Type) {
            final ParameterizedType parameterizedType = (ParameterizedType) ASTNodes.getParent(node, ParameterizedType.class);
            if (parameterizedType != null) {
                final BodyDeclaration parent = (BodyDeclaration) ASTNodes.getParent(node, BodyDeclaration.class);
                if (parent instanceof AbstractTypeDeclaration) {
                    return (IType)((AbstractTypeDeclaration)parent).resolveBinding().getJavaElement();
                }
            }
        }
        return null;
    }
    

    /**
     * Returns the method which references the specified ast node.
     * 
     * @param node
     *            the ast node
     * @return the referencing method
     * @throws JavaModelException
     *             if an error occurs
     */
    private static IMethod getReferencingMethod(final ASTNode node) throws JavaModelException {
        if (node instanceof Type) {
            final BodyDeclaration parent = (BodyDeclaration) ASTNodes.getParent(node, BodyDeclaration.class);
            if (parent instanceof MethodDeclaration) {
                MethodDeclaration methodDeclaration = ((MethodDeclaration) parent);
                if (nodeContains(methodDeclaration.getReturnType2(), node)) { 
                    final IMethodBinding binding = methodDeclaration.resolveBinding();
                    if (binding != null) {
                        final ICompilationUnit unit = RefactoringASTParser.getCompilationUnit(node);
                        final IJavaElement element = unit.getElementAt(node.getStartPosition());
                        if (element instanceof IMethod) {
                            return (IMethod) element;
                        }
                    }
                }
            }
        }
        return null;
    }

    private static boolean nodeContains(Type possibleParent, ASTNode node) {
        ASTNode checkNode = node;
        do {
            if(checkNode.equals(possibleParent)) {
                return true;
            }
            checkNode = checkNode.getParent();
        } while (checkNode != null && !(checkNode instanceof BodyDeclaration));
        
        return false;
    }
    
    /**
     * Returns the field which corresponds to the specified variable declaration
     * fragment.
     * 
     * @param fragment
     *            the variable declaration fragment
     * @return the corresponding field
     * @throws JavaModelException
     *             if an error occurs
     */
    private static IField getCorrespondingField(final VariableDeclarationFragment fragment) throws JavaModelException {
        final IBinding binding = fragment.getName().resolveBinding();
        if (binding instanceof IVariableBinding) {
            final IVariableBinding variable = (IVariableBinding) binding;
            if (variable.isField()) {
                final ICompilationUnit unit = RefactoringASTParser.getCompilationUnit(fragment);
                final IJavaElement element = unit.getElementAt(fragment.getStartPosition());
                if (element instanceof IField) {
                    return (IField) element;
                }
            }
        }
        return null;
    }

    private static void findReferences(Map<IJavaProject, Set<ICompilationUnit>> units, IJavaElement javaElement, RefactoringSearchEngine2 engine) throws JavaModelException {
        Map<IJavaProject, Set<ICompilationUnit>> newUnits = Cache.findReferences(javaElement, engine);
        for (Map.Entry<IJavaProject, Set<ICompilationUnit>> entry : newUnits.entrySet()) {
            Set<ICompilationUnit> previousUnits = units.get(entry.getKey());
            if (previousUnits == null) {
                previousUnits = new HashSet<ICompilationUnit>();
                units.put(entry.getKey(), previousUnits);
            }
            previousUnits.addAll(entry.getValue());
        }
    }

}
