/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.creation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.TypeVariable2;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.util.CacheMap;

/**
 * This class handles the stuff needed for the usages of parameterized types
 * with simple types as parameters in the type hierarchy. For example
 * public class MyComparator implements Comparator<Test> {
 *     public int compare(Test o1, Test o2) { ... }
 * }
 * The class is used by the {@link InferTypeConstraintsCreator} exclusively.
 * 
 * @author Hannes Kegel
 *
 */
public class TypeDeclElementHandler {
    private HashMap<TType, List<TypeDeclElementInfo>> typeDeclElementMap = new HashMap<TType, List<TypeDeclElementInfo>>();    

    private List<MethodInfo> methodInfos = new ArrayList<MethodInfo>(); 
    
    private CacheMap<TType, List<TypeDeclElementInfo>> declInfoCache = new CacheMap<TType, List<TypeDeclElementInfo>>(500);
            
    /**
     * Receives the parameters from all method declarations encountered as well as from method calls 
     * on parametrized types. Saves all methods where the original method is 
     * parametrized for the finish step.
     * 
     * @param originalBinding the encountered method binding 
     * @param parameterBinding the parameter binding
     * @param origType the type of the original method declaration
     * @param methodDeclType the type of the current method declaration or invocation
     * @param declVar the constraint variable matching methodDeclType
     * @param model the model
     */
    public void registerParameter(IMethodBinding originalBinding, ITypeBinding parameterBinding, TType origType, TType methodDeclType, ConstraintVariable2 declVar, InferTypeConstraintsModel model) {
        ITypeBinding origDeclClass = originalBinding.getDeclaringClass();
        if (origDeclClass.isParameterizedType()) {
            addMethodInfo(origType, methodDeclType, declVar, parameterBinding, model);
        }
    }
    
    /**
     * Registers all type variables created for declaring the hierarchy (for example in the 
     * extends clause).
     * 
     * @param declaringType the type that contains the declaration as supertype
     * @param typeVariable the type variable of the supertype
     */
    public void registerTypeVariable(TType declaringType, TypeVariable2 typeVariable) {
        TypeDeclElementInfo typeDeclElementSet = new TypeDeclElementInfo(declaringType, typeVariable.getType(), typeVariable);
        List<TypeDeclElementInfo> list = typeDeclElementMap.get(declaringType);
        if (list == null) {
            list = new ArrayList<TypeDeclElementInfo>();
            typeDeclElementMap.put(declaringType, list);            
        }
        list.add(typeDeclElementSet);
    }    
    
    /**
     * Creates constraints between all relevant method declarations and calls
     * the super type parameterization.
     * 
     * @param model the model
     */
    public void finish(InferTypeConstraintsModel model) {
        for (MethodInfo methodInfo : methodInfos) {
            List<TypeDeclElementInfo> allDeclInfos = getAllDeclInfos(methodInfo.methodDeclType);
            
            for (TypeDeclElementInfo typeDeclElementInfo : allDeclInfos) {
                if (isSecondSubType(methodInfo.origType, typeDeclElementInfo.paramSuperType)) {
                    ElementVariable elementVariable = InferTypeConstraintsModel.getElementVariables(typeDeclElementInfo.typeVariable).get(methodInfo.key);
                    if (elementVariable != null) {
                        model.createEqualityConstraint(elementVariable, methodInfo.declVar);
                    }
                }
            }            
        }

        List<ConstraintInfo> constraintInfos = new ArrayList<ConstraintInfo>();
        for (ConstraintVariable2 cv : model.getConstraintVariables()) {
            if (cv instanceof TypeVariable2) {
                TType type = cv.getType().getTypeDeclaration();
                if (type != null) {
                    List<TypeDeclElementInfo> allDeclInfos = getAllDeclInfos(type);
                    for (TypeDeclElementInfo typeDeclElementInfo : allDeclInfos) {
                        //save ConstraintInfo to prevent ConcModEx
                        constraintInfos.add(new ConstraintInfo(cv, typeDeclElementInfo.typeVariable));
                    }            
                }
            }
        }
        
        for (ConstraintInfo constraintInfo : constraintInfos) {
            model.createElementEqualsConstraints(constraintInfo.cv1, constraintInfo.cv2, false, false);            
        }
        
        typeDeclElementMap = null;
        methodInfos = null;
        declInfoCache = null;
    }


    private void addMethodInfo(TType origType, TType methodDeclType, ConstraintVariable2 contraintVar, ITypeBinding currentBinding, InferTypeConstraintsModel model) {
        if (currentBinding.isWildcardType()) {
            currentBinding = currentBinding.getBound();
            if (currentBinding == null) {
                return;
            }
        }
        if (currentBinding.isArray()) {
            currentBinding = currentBinding.getElementType();
        }
        if (currentBinding.isTypeVariable()) {
            MethodInfo methodInfo = new MethodInfo(contraintVar, currentBinding.getKey(), origType, methodDeclType);
            methodInfos.add(methodInfo);
        } else if (currentBinding.isParameterizedType()) {
            ITypeBinding[] params = currentBinding.getTypeDeclaration().getTypeParameters();
            ITypeBinding[] args = currentBinding.getTypeArguments();
            for (int i=0; i < args.length; i++) {
                ElementVariable elementVariable = InferTypeConstraintsModel.getElementVariables(contraintVar).get(params[i].getKey());
                if (elementVariable != null) {
                    addMethodInfo(origType, methodDeclType, elementVariable, args[i], model);
                }
            }
        }
    }

    private List<TypeDeclElementInfo> getAllDeclInfos(TType type) {
        List<TypeDeclElementInfo> allDeclSets = declInfoCache.get(type);
        if (allDeclSets == null) {
            allDeclSets = new ArrayList<TypeDeclElementInfo>();
            addAllDeclInfos(allDeclSets, type);
            declInfoCache.put(type, allDeclSets);
        }
        return allDeclSets;
    }

    private void addAllDeclInfos(List<TypeDeclElementInfo> allDeclSets, TType type) {
        if (type == null) {
            return;
        }
        List<TypeDeclElementInfo> singleList = typeDeclElementMap.get(type);
        if (singleList != null) {
            allDeclSets.addAll(singleList);
        }
        
        addAllDeclInfos(allDeclSets, type.getSuperclass());
        
        for (TType interfaceType : type.getInterfaces()) {
            addAllDeclInfos(allDeclSets, interfaceType);            
        }
    }

    private static boolean isSecondSubType(TType possibleSuperType, TType type) {
        if (type != null) {
            if (possibleSuperType.equals(type)) {
                return true;
            }            
            if (isSecondSubType(possibleSuperType, type.getSuperclass())) {
                return true;
            }
            for (TType interfaceType : type.getInterfaces()) {
                if (isSecondSubType(possibleSuperType, interfaceType)) {
                    return true;
                }
            }
        }            
        return false;
    }
    
    private static class MethodInfo {
        ConstraintVariable2 declVar;
        String key;
        // maybe higher in the hierachy than TypeDeclElementInfo.superType
        TType origType; 
        // maybe lower in the hierachy than TypeDeclElementInfo.declaringType
        TType methodDeclType;

        public MethodInfo(ConstraintVariable2 declVar, String key, TType origType, TType methodDeclType) {
            this.declVar = declVar;
            this.key = key;
            this.origType = origType;
            this.methodDeclType = methodDeclType;
        }

        @Override
        public String toString() {
            return MethodInfo.class.getSimpleName() + ":\ndeclVar: " 
                    + declVar + "\nkey: " + key + "\norigType: " + origType 
                    + "\nmethodDeclType: " + methodDeclType
                    + "\n-------------------------------"; 
        }        
    }

    private static class ConstraintInfo {
        ConstraintVariable2 cv1;
        ConstraintVariable2 cv2;

        public ConstraintInfo(ConstraintVariable2 cv1, ConstraintVariable2 cv2) {
            this.cv1 = cv1;
            this.cv2 = cv2;
        }            
    }

    private static class TypeDeclElementInfo {
        TType declaringType;
        TType paramSuperType;
        TypeVariable2 typeVariable;
        
        public TypeDeclElementInfo(TType declaringType, TType superType, TypeVariable2 typeVariable) {
            this.declaringType = declaringType;
            this.paramSuperType = superType;
            this.typeVariable = typeVariable;
        }
        
        @Override
        public String toString() {
            return TypeDeclElementInfo.class.getSimpleName() + ":\ndeclaringType: " 
                        + declaringType + "\nsuperType: " + paramSuperType + "\ntypeVariable: " + typeVariable
                        + "\n-------------------------------"; 
        }        
        
    }
}
