/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.util;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.corext.refactoring.structure.constraints.ConditionalTypeConstraint;
import org.eclipse.jdt.internal.corext.refactoring.structure.constraints.CovariantTypeConstraint;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.intoJ.inferType3.internal.constraintModel.EqualityTypeConstraint;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;

/**
 * Helper class for conditional depth first traversal of the constraint graph.
 * Clients must implements the {@link Visitor}.
 *  
 * @author Hannes Kegel
 *
 */
public class ConstraintTraversal {
    /**
     * Visits all connected constraints considering subtype constraint direction 
     * with a depth-first approach.
     *  
     * @param baseCv the variable to start from 
     * @param constraintVisitor the visitor
     * @throws JavaModelException
     */
    public static void visitAllDirectedConstraints(ConstraintVariable2 baseCv, Visitor constraintVisitor) throws JavaModelException {
        visitConnectedConstraints(new HashSet<ITypeConstraint2>(), baseCv, constraintVisitor, true);
    }

    /**
     * Visits all connected constraints without considering subtype constraint direction 
     * with a depth-first approach.
     *  
     * @param baseCv the variable to start from 
     * @param constraintVisitor the visitor
     * @throws JavaModelException
     */
    public static void visitAllConnectedConstraints(ConstraintVariable2 baseCv, Visitor constraintVisitor) throws JavaModelException {
        visitConnectedConstraints(new HashSet<ITypeConstraint2>(), baseCv, constraintVisitor, false);
    }

    /**
     * Visits only the directly connected constraints considering subtype constraint direction.
     *  
     * @param currentProcessSet the constraints that are already processed. Will be filled by this method.
     * @param baseCv the variable to start from
     * @param constraintVisitor the visitor
     * @return false if the visitor returned false
     * @throws JavaModelException
     */
    public static boolean visitSingleStep(Set<ITypeConstraint2> currentProcessSet, ConstraintVariable2 baseCv, Visitor constraintVisitor) throws JavaModelException {

        Collection constraints= InferTypeConstraintsModel.getVariableUsage(baseCv);
        for (Iterator iter = constraints.iterator(); iter.hasNext();) {
            ITypeConstraint2 constraint = (ITypeConstraint2) iter.next();
            
            ConstraintVariable2 otherConstraintVariable2 = constraint.getLeft() == baseCv ? constraint.getRight() : constraint.getLeft();
            
            //LogUtil.debugOut(ConstraintTraversal.class, base.toString() + ": constraint: " + constraint + ", left: " + constraint.getLeft().getClass().getName() + ", right: " + constraint.getRight().getClass().getName());
            //LogUtil.debugOut(ConstraintTraversal.class, base.toString() + ": other: " + otherConstraintVariable2 + ", " + currentProcessSet.contains(otherConstraintVariable2));

            if (shouldFollowConstraint(baseCv, currentProcessSet, constraint, otherConstraintVariable2, true)) 
            {
                currentProcessSet.add(constraint);
                if (!constraintVisitor.visit(baseCv, otherConstraintVariable2, constraint)) {
                    return false;
                }
            }
        }   
        
        return true;
    }
    
    private static boolean shouldFollowConstraint(ConstraintVariable2 variable, Set<ITypeConstraint2> currentProcessSet, ITypeConstraint2 constraint, ConstraintVariable2 otherConstraintVariable2, boolean directed) {
        if (currentProcessSet.contains(constraint) || constraint instanceof ConditionalTypeConstraint
                || constraint instanceof CovariantTypeConstraint) 
        {
            return false;
        }
        if (directed) {
            return constraint.getLeft() == variable || constraint instanceof EqualityTypeConstraint;                     
        } else {
            return true;
        }
    }

    private static void visitConnectedConstraints(Set<ITypeConstraint2> currentProcessSet, ConstraintVariable2 base, Visitor constraintVisitor, boolean directed) throws JavaModelException {
        boolean done = false;
        while (!done) {
            Collection<ITypeConstraint2> constraints= InferTypeConstraintsModel.getVariableUsage(base);
            try {
                for (ITypeConstraint2 constraint : constraints) {
                    ConstraintVariable2 otherConstraintVariable2 = constraint.getLeft() == base ? constraint.getRight() : constraint.getLeft();
                    
                    //LogUtil.debugOut(ConstraintTraversal.class, base.toString() + ": constraint: " + constraint + ", left: " + constraint.getLeft().getClass().getName() + ", right: " + constraint.getRight().getClass().getName());
                    //LogUtil.debugOut(ConstraintTraversal.class, base.toString() + ": other: " + otherConstraintVariable2 + ", " + currentProcessSet.contains(otherConstraintVariable2));
        
                    if (shouldFollowConstraint(base, currentProcessSet, constraint, otherConstraintVariable2, directed)) 
                    {
                        currentProcessSet.add(constraint);
                        if (constraintVisitor.visit(base, otherConstraintVariable2, constraint)) {
                            visitConnectedConstraints(currentProcessSet, otherConstraintVariable2, constraintVisitor, directed);                    
                        }
                    }
                }
                done = true;
            } catch (ConcurrentModificationException e) {
                
            }
        }
    }
        
    /**
     * The visitor for constraint traversal.
     * 
     * @author Hannes Kegel
     *
     */
    public interface Visitor {
        /**
         * Called for visited constraints.
         * 
         * @param baseCv the variable we come from 
         * @param otherCv the variable we go to
         * @param constraint the constraint connecting the variables
         * @return true if you want to continue visiting connected constraints
         * @throws JavaModelException
         */
        public boolean visit(ConstraintVariable2 baseCv, ConstraintVariable2 otherCv, ITypeConstraint2 constraint) throws JavaModelException; 
    }
}
