/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.util;

import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IImportDeclaration;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.intoJ.inferType3.internal.util.ConversionUtil;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * Helper class to store all relevant information about the
 * declaration element(s) to be refactored.
 * 
 * @author Hannes Kegel
 *
 */
public class SelectedElement {

    private IBinding binding;
    private ITypeBinding declaringClass;
    
    private ITypeBinding declaredType;

    private ICompilationUnit compilationUnit;
    private String packageName;
    private int typeStartPos;
    
    private Scope scope;
    
    private List<String> typeVariableKeys;
    
    /**
     * Constructor for fields and return types
     * 
     * @param binding the binding 
     * @param typeStartPos the start position of the type (not the name)
     * @param typeVariableKeys if the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     */
    public SelectedElement(IBinding binding, int typeStartPos, List<String> typeVariableKeys) {
        this.binding = binding;
        this.typeStartPos = typeStartPos;
        this.typeVariableKeys = typeVariableKeys;

        // Basic stuff
        declaringClass = null;
        declaredType = null;
        
        scope = Scope.SINGLE_DECLARATION_ELEMENT;

        if (binding.getKind() == IBinding.METHOD) {
            declaringClass = ((IMethodBinding) binding).getDeclaringClass();
            declaredType = ((IMethodBinding) binding).getReturnType();
        }
        if (binding.getKind() == IBinding.VARIABLE) {
            declaringClass = ((IVariableBinding) binding).getDeclaringClass();
            declaredType = ((IVariableBinding) binding).getType();
        }
        if (binding.getKind() == IBinding.TYPE) {
            declaringClass = (ITypeBinding)binding;
            declaredType = (ITypeBinding)binding;
            scope = Scope.EVERYTHING;
        }
        if (typeVariableKeys != null) {
            declaredType = stripType(declaredType);
            for (String key : typeVariableKeys) {
                ITypeBinding[] typeVariables = declaredType.getTypeDeclaration().getTypeParameters();
                ITypeBinding[] typeArguments = declaredType.getTypeArguments();
                for (int i = 0; i < typeVariables.length; i++) {
                    if (typeVariables[i].getKey().equals(key)) {
                        declaredType = stripType(typeArguments[i]);
                        break;
                    }
                }
            }
        }        
        
        declaredType = stripType(declaredType);
        if (binding.getJavaElement() instanceof IMember) {
            compilationUnit = ((IMember)binding.getJavaElement()).getCompilationUnit();
        } else {
            compilationUnit = ((IMember)declaredType.getJavaElement()).getCompilationUnit();
        }
        // Note that local variables (and parameters) don't have a declaring
        // class.
    }

    /**
     * Constructor for local variables or the general case.
     * 
     * @param binding the binding
     * @param declaringClass the declaring class of the selection
     * @param typeStartPos the start position of the type (not of the identifier)
     * @param typeVariableKeys if the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     */
    public SelectedElement(IBinding binding, ITypeBinding declaringClass, int typeStartPos, List<String> typeVariableKeys) {
        this(binding, typeStartPos, typeVariableKeys);
        this.declaringClass = declaringClass;
    }
    
    /**
     * Constructor for the import selection within the editor.
     * 
     * @param binding the binding of the import type
     * @param coveringNode the covering ASTNode of the selection
     */
    public SelectedElement(ITypeBinding binding, ASTNode coveringNode) {
        this.binding = binding;
        this.declaredType = binding;
        
        CompilationUnit cu = (CompilationUnit)coveringNode.getRoot();
        
        compilationUnit = (ICompilationUnit)cu.getJavaElement();
        packageName = ((IPackageFragment)compilationUnit.getParent()).getElementName();
        
        List types = cu.types();
        if (types != null && types.size() > 0) {
            AbstractTypeDeclaration abstractTypeDeclaration =  (AbstractTypeDeclaration)types.get(0);
            declaringClass = abstractTypeDeclaration.resolveBinding();
        }        
    }

    /**
     * Constructor for the import selection in the outline view.
     * 
     * @param binding the binding of the import type
     * @param importDeclaration the selected import declaration
     * @param scope the selected scope
     */
    public SelectedElement(ITypeBinding binding, IImportDeclaration importDeclaration, Scope scope) throws JavaModelException, InternalCalculationException {
        this.binding = binding;
        this.declaredType = binding;
        this.scope = scope;

        this.compilationUnit = getCompilationUnit(importDeclaration);
        packageName = ((IPackageFragment)compilationUnit.getParent()).getElementName();
        
        IType[] types = compilationUnit.getTypes();
        if (types != null && types.length > 0) {
            this.declaringClass = ConversionUtil.getTypeBinding(types[0]);            
        }
    }
    
    
    /**
     * Checks if the selection is a complete type (via the import refac).
     * 
     * @return true if it is a complete type
     */
    public boolean isTypeSelection() {
        return binding instanceof ITypeBinding;
    }

    

    /**
     * Checks whether a complete type has been selected for refactoring via the import refac and
     * if the given binding is part of the selection.
     * 
     * @param methodBinding the binding to check
     * @return true if import refac and the binding is part of the selection
     */
    public boolean isTypeSelection(IMethodBinding methodBinding) {
        if (isTypeSelection() && methodBinding.getReturnType().getKey().equals(binding.getKey())) {
            return isInScope(getCompilationUnit(methodBinding.getDeclaringClass().getJavaElement()));
        }
        return false;
    }
    
    /**
     * Checks whether a complete type has been selected for refactoring via the import refac and
     * if the given binding is part of the selection.
     * 
     * @param variableBinding the binding to check
     * @return true if import refac and the binding is part of the selection
     */
    public boolean isTypeSelection(IVariableBinding variableBinding) {
        if (isTypeSelection() && variableBinding.getType().getKey().equals(binding.getKey())) {
            ITypeBinding varDecl = variableBinding.getDeclaringClass();
            if (varDecl == null) {
                varDecl = variableBinding.getDeclaringMethod().getDeclaringClass();
            }
            return isInScope(getCompilationUnit(varDecl.getJavaElement()));
        }
        return false;
    }
    

    /**
     * Check the given type is selected within the given cu by an import refactoring.
     * 
     * @param type the type to check
     * @param cu the cu where this type is used
     * @return true if this type is selected
     */
    public boolean isTypeSelection(TType type, ICompilationUnit cu) {
        if (isTypeSelection() && type != null && type.getBindingKey().equals(binding.getKey())) {
            return isInScope(cu);
        }
        return false;
    }

    
    
    /**
     * Returns the selected binding.
     * 
     * @return the binding
     */
    public IBinding getBinding() {
        return binding;
    }

    /**
     * Returns the declaring class. Important for local variables.
     * 
     * @return the declaring class
     */
    public ITypeBinding getDeclaringClass() {
        return declaringClass;
    }

    /**
     * Returns the declared type of the selection.
     * 
     * @return the declared type
     */
    public ITypeBinding getDeclaredType() {
        return declaredType;
    }
    
    /**
     * Returns the fully qualified name of the selection.
     * 
     * @return String
     */
    public String getFullyQualifiedNameOfType() {

        if (binding instanceof IVariableBinding) {
            return ((IVariableBinding) binding).getType().getQualifiedName();
        }
        if (binding instanceof IMethodBinding) {
            return ((IMethodBinding) binding).getReturnType().getQualifiedName();
        }
        return "";
    }

    /**
     * Returns the CU of the selection.
     * 
     * @return the compilation unit
     */
    public ICompilationUnit getCompilationUnit() {
        return compilationUnit;
    }

    /**
     * Sets the CU of the selection.
     * 
     * @param compilationUnit the cu to be set
     */
    public void setCompilationUnit(ICompilationUnit compilationUnit) {
        this.compilationUnit = compilationUnit;
    }

    /**
     * Retrieves the starting position of the type declaration of the selection.
     * @return the starting position of the type (not the name)
     */
    public int getTypeStartPos() {
        return typeStartPos;
    }
    
    /**
     * If the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     * 
     * @return the binding keys or null
     */
    public List<String> getTypeVariableKeys() {
        return typeVariableKeys;
    }    
    
    /**
     * Sets the scope for the import refactoring.
     * 
     * @param scope the scope
     */
    public void setScope(Scope scope) {
        this.scope = scope;
    }
    
    /**
     * Sets the name of the package in which this element's compilation unit is
     * declared. Needed for refactorings with scope {@link Scope#PACKAGE} or
     * {@link Scope#PACKAGE_INCLUDING_SUBPACKAGES}.
     *  
     * @param packageName the package name
     */
    protected void setPackageName(String packageName) {
    	this.packageName = packageName;
    }
    
    
    private boolean isInScope(ICompilationUnit cu) {
        if (cu != null) {
            String otherPackageName;
            switch (scope) {
	            case EVERYTHING:
	            	return true;
                case PROJECT:
                    return compilationUnit.getJavaProject().equals(cu.getJavaProject());
                case COMPILATION_UNIT:
                    return compilationUnit.equals(cu);
                case PACKAGE:
                    otherPackageName = ((IPackageFragment)cu.getParent()).getElementName();
                    return packageName.equals(otherPackageName);
                case PACKAGE_INCLUDING_SUBPACKAGES:
                    otherPackageName = ((IPackageFragment)cu.getParent()).getElementName();
                    return packageName.equals(otherPackageName) || otherPackageName.startsWith(packageName + ".");
            }
        }
        return false;
    }
   
    
    private static ICompilationUnit getCompilationUnit(IJavaElement javaElement) {
        while (javaElement != null) {
            if (javaElement instanceof ICompilationUnit) {
                return (ICompilationUnit)javaElement;
            }
            javaElement = javaElement.getParent();
        }
        return null;
    }

    private ITypeBinding stripType(ITypeBinding type) {
        if (type != null && type.isWildcardType()) {
            type = type.getBound(); 
        }
        if (type != null && type.isArray()) {
            type = type.getElementType(); 
        }
        return type;
    }
        
}
