/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.variables;

import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeVariable;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;

/**
 * An element variable represent a single type argument of a parameterized type.
 * 
 * @author Hannes Kegel
 *
 */
public class ElementVariable extends ConstraintVariable2 {

    private final ConstraintVariable2 fParentCv;
    private final String fTypeVariableKey;

    /**
     * Constructs a new element variable from a TypeVariable
     * 
     * @param parentCv the variable this is an element of
     * @param typeVariable the type variable of the parent
     * @param ttype the current type of the element
     */
    public ElementVariable(ConstraintVariable2 parentCv, TypeVariable typeVariable, TType ttype) {
        super(ttype);
        fParentCv= parentCv;
        //LogUtil.debugOut2(this, "Elem[" + fParentCv.toString() + "] has type " + ttype);
        fTypeVariableKey= typeVariable.getBindingKey();
    }

    /**
     * Constructs a new element variable by taking another one as a template.
     *  
     * @param parentCv the variable this is an element of
     * @param otherElementVariable the template
     */
    public ElementVariable(ConstraintVariable2 parentCv, ElementVariable otherElementVariable) {
        super(otherElementVariable.getType() == null || otherElementVariable.getType().isTypeVariable() ? null : otherElementVariable.getType());
        fParentCv= parentCv;
        fTypeVariableKey= otherElementVariable.getTypeVariableKey();
    }

    /*
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return fParentCv.hashCode() ^ fTypeVariableKey.hashCode();
    }

    /*
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object other) {
        if (this == other) {
            return true;
        }
        if (!(other instanceof ElementVariable)) {
            return false;
        }
        
        ElementVariable other2= (ElementVariable) other;
        return fParentCv == other2.fParentCv
                && fTypeVariableKey.equals(other2.fTypeVariableKey);
    }
    
    /**
     * Returns the binding key of the type variable.
     * 
     * @return the binding key
     */
    public String getTypeVariableKey() {
        return fTypeVariableKey;
    }

    /**
     * Returns the constraint variable this is an element of. For nested constraints
     * this can be also an ElementVariable.
     * 
     * @return the parent
     */
    public ConstraintVariable2 getParentConstraintVariable() {
        return fParentCv;
    }
    
    /**
     * Returns the base parameterized constraint variable this is an element of. 
     * This will be a TypeVariable2.
     * 
     * @return the base variable
     */
    public ConstraintVariable2 getBaseConstraintVariable() {
        ConstraintVariable2 base = this;
        do {
            base = ((ElementVariable)base).getParentConstraintVariable();
        } while (base instanceof ElementVariable);
        return base;
    }
    
        
    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2#toString()
     */
    @Override
    public String toString() {
        return "Elem[" + fParentCv.toString() + ", " + fTypeVariableKey + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
    }
}
