/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.manipulation;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.BindingKey;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.ui.CodeGeneration;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.NewTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ParameterizedTypeUtil;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * Generates the code snippets for the creator.
 * 
 * @author Philip Mayer
 * @author Hannes Kegel
 */
public class CodeGenerator {

	/**
	 * Returns the code comment for the given class.
	 * 
	 * @param fqName full qualified name
	 * @param workingCopy the compilation unit working on
	 * @return the type comment
	 * @throws CoreException
	 */
	public static String getTypeComment(String fqName, ICompilationUnit workingCopy) throws CoreException {

		return CodeGeneration.getTypeComment(workingCopy, fqName, getLineDelimitter());
	}

	/**
     * Creates the beginning of a new type.
     * 
	 * @param newTCType the type to be created
	 * @param importRewrite the import rewriter used to get simple names 
	 * @return StringBuffer
	 * @throws JavaModelException
	 * @throws InternalCalculationException
	 */
	public static StringBuffer getTypeBeginning(NewTCType newTCType, ImportRewrite importRewrite) throws JavaModelException, InternalCalculationException {
        StringBuffer sBuf = new StringBuffer();

		String typeName = newTCType.getSimpleName();
		String type = newTCType.isInterface() ? "public interface " : "public abstract class ";
		sBuf.append(type).append(typeName).append(newTCType.getErasureTypeParameters(importRewrite));
        if (newTCType.getSuperType() != null) {
            sBuf.append(" extends ");
            sBuf.append(newTCType.getSuperType().getHierarchyTypeName(newTCType, importRewrite));
        }
        int ifCount = newTCType.getInterfaces().size();
        if (ifCount > 0) {
            sBuf.append(newTCType.isInterface() ? " extends " : " implements ");
            int ifPos = 0;
            for (TCType interf : newTCType.getInterfaces()) {
                sBuf.append(interf.getHierarchyTypeName(newTCType, importRewrite));
                if (ifPos < ifCount - 1) {
                    sBuf.append(", ");
                }
                ifPos++;
            }
        }
        sBuf.append(" {");
		sBuf.append(getLineDelimitter()).append(getLineDelimitter());
		return sBuf;
	}

	/**
	 * Returns the line delimiter on the current system.
	 * 
	 * @return String
	 */
	public static String getLineDelimitter() {

		return System.getProperty("line.separator", "\n");
	}

	/**
     * Creates a StringBuffer with all required method declarations.
     *
	 * @param newTCType the type to be created
	 * @param importRewrite the import rewriter used to get simple names
	 * @return StringBuffer with all method declarations
	 * @throws JavaModelException
	 * @throws InternalCalculationException
	 */
	public static StringBuffer getMethodDeclarations(NewTCType newTCType, ImportRewrite importRewrite) throws JavaModelException, InternalCalculationException {
        BindingSet methodSet = newTCType.getDeclaredMethods();

		StringBuffer declarations = new StringBuffer();
		if (methodSet.isEmpty()) {
			return declarations;
        }
        
        Set<String> addedMethods = new HashSet<String>();
        
        for (BindingWrapper wrapper : methodSet) {
            IMethodBinding binding = findParametrizedBinding(wrapper.getMethodBinding(), newTCType.getBaseTypeBinding());
            
            String sig = new BindingKey(binding.getKey()).toSignature();
            if (addedMethods.add(binding.getName() + sig)) {
                IMethod originalMethod = wrapper.getOriginalMethod();
                declarations.append(GenerationUtil.getJavadoc(originalMethod));
                
                StringBuffer modifiers = new StringBuffer();
    			if (!newTCType.isInterface()) {
                    modifiers = GenerationUtil.getModifiers(binding, true);
                }
    
    			String returnTypeName = importRewrite.addImport(binding.getReturnType());
    			String methodName = binding.getName();
    			StringBuffer parameterList = GenerationUtil.getParametersList(binding, originalMethod, importRewrite);
    			StringBuffer exceptionList = GenerationUtil.getExceptionsList(binding, importRewrite);
    
    			declarations.append("\t").append(modifiers);
                
                StringBuffer methodTypeParameterList = GenerationUtil.getTypeParameterList(binding, importRewrite);
                if (methodTypeParameterList.length() != 0) {
                    declarations.append("<").append(methodTypeParameterList).append("> ");
                }
    
                declarations.append(returnTypeName).append(" ").append(methodName);
    			declarations.append("(").append(parameterList).append(")");
    			if (exceptionList.length() != 0) {
    				declarations.append(" throws ").append(exceptionList);
                }
    			declarations.append(";").append(getLineDelimitter());
            }
		}

		declarations.append(getLineDelimitter());
		return declarations;
	}

	public static IMethodBinding findParametrizedBinding(IMethodBinding wantedBinding, ITypeBinding subTypeBinding) throws JavaModelException, InternalCalculationException {
        ITypeBinding declaringClass = wantedBinding.getDeclaringClass().getTypeDeclaration();
        if (declaringClass.isGenericType()) {
            ITypeBinding parametrizedType = ParameterizedTypeUtil.getParametrizedTypeBinding(declaringClass.getKey(), subTypeBinding);
            if (parametrizedType != null) {
                for (IMethodBinding methodBinding : parametrizedType.getDeclaredMethods()) {
                    if (methodBinding.getMethodDeclaration().getKey().equals(wantedBinding.getMethodDeclaration().getKey())) {
                        return methodBinding;
                    }
                }
            }
        }
        return wantedBinding;
    }

}