/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.manipulation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeParameter;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.ArrayType;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.ParameterizedType;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.WildcardType;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;
import org.eclipse.jdt.internal.corext.dom.NodeFinder;
import org.eclipse.jdt.internal.corext.refactoring.Checks;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringCoreMessages;
import org.eclipse.jdt.internal.corext.refactoring.changes.CreateCompilationUnitChange;
import org.eclipse.jdt.internal.corext.refactoring.structure.ASTNodeSearchUtil;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.CompilationUnitRange;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraintVariable;
import org.eclipse.jdt.internal.corext.refactoring.util.ResourceUtil;
import org.eclipse.jdt.internal.corext.refactoring.util.TextEditBasedChangeManager;
import org.eclipse.ltk.core.refactoring.CategorizedTextEditGroup;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.CompositeChange;
import org.eclipse.ltk.core.refactoring.GroupCategory;
import org.eclipse.ltk.core.refactoring.GroupCategorySet;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.core.refactoring.TextEditBasedChange;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.NewTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * This class handles all source manipulations to be done by Infer Type. The new types
 * must have been calculated before calling calculateChanges.
 * 
 * @author Hannes Kegel
 *
 */
public class Manipulation {

    private static final String ID_INFER_TYPE = "org.intoJ.inferType3.internal";

    /** The infer type group category set */
    private static final GroupCategorySet SET_INFER_TYPE = new GroupCategorySet(new GroupCategory(ID_INFER_TYPE, "Infer Type", "Infer Type"));

    private SelectedElement selectedElement;
    private TypeHierarchy typeHierarchy;
    private InferTypeConstraintsModel model;

    private List<Change> newCuChanges = new ArrayList<Change>();
    private TextEditBasedChangeManager manager = new TextEditBasedChangeManager();
    
    // TODO: IT4: replace with serial processing, this is memory intensive
    private Map<ICompilationUnit, CompilationUnitRewrite> unitToRewrite = new HashMap<ICompilationUnit, CompilationUnitRewrite>();

    private int numberOfChanges = 0;
        
    /**
     * The single constructor.
     * 
     * @param model the Infer Type model
     * @param typeHierarchy the hierarchy of the selected element
     * @param selectedElement the selection
     */
    public Manipulation(InferTypeConstraintsModel model, TypeHierarchy typeHierarchy, SelectedElement selectedElement) {
        this.model = model;
        this.typeHierarchy = typeHierarchy;
        this.selectedElement = selectedElement;
    }

    /**
     * Creates all necessary changes based on the type calculation.
     * 
     * @param monitor the progress monitor to display progress
     * @param status the refactoring status
     * @param redeclareDEs if true the new type will be used for the selected DE. 
     * @param validationContext the validation context (a Shell), may be null
     * @throws JavaModelException
     * @throws CoreException
     * @throws InternalCalculationException 
     */
    public final void calculateChanges(final IProgressMonitor monitor, final RefactoringStatus status, boolean redeclareDEs, Object validationContext) throws JavaModelException, CoreException, InternalCalculationException {        
        Assert.isNotNull(status);
        Assert.isNotNull(monitor);
        
        try {
            monitor.beginTask("", 85);
            monitor.setTaskName(InferTypeRefactoring.MSG_CREATING_CHANGES);

            final IJavaProject project = selectedElement.getBinding().getJavaElement().getJavaProject();

            if (redeclareDEs) {
                rewriteDeclarationElements(model, status, new SubProgressMonitor(monitor, 65));
            }

            if (!status.hasFatalError()) {
                introduceTypes(status);
            }
            monitor.worked(10);
            if (!status.hasFatalError()) {
                rewriteHierarchy(project, status);
            }
            monitor.worked(10);

            if (!status.hasFatalError()) {
                for (CompilationUnitRewrite rewrite : unitToRewrite.values()) {
                    try {
                        manager.manage(rewrite.getCu(), rewrite.createChange());
                    } catch (Exception e) {
                        status.addError("Could not manage all rewrites");
                    }
                }
            }
            
            if (!status.hasFatalError()) {
                final RefactoringStatus validation = Checks.validateModifiesFiles(ResourceUtil.getFiles(manager.getAllCompilationUnits()), validationContext);
                if (!validation.isOK()) {
                    status.merge(validation);
                }
            }
            

        } finally {
            monitor.done();
        }
    }
    
    /**
     * Returns all changes as a single composite change object.
     * 
     * @return all changes or null if there are none
     * @throws JavaModelException
     */
    public Change getAllChanges() throws JavaModelException {
        CompositeChange compositeChange = new CompositeChange("Infer Type");

        final TextEditBasedChange[] changes = manager.getAllChanges();
        if (changes != null && changes.length != 0) {
            compositeChange.addAll(changes);
        }
        
        if (newCuChanges.size() > 0) {
            compositeChange.addAll(newCuChanges.toArray(new Change[newCuChanges.size()]));
        }

        if (compositeChange.getChildren().length > 0) {
            Cache.changesMade();
            return compositeChange;
        } 
        
        return compositeChange;
    }
    
    private void rewriteDeclarationElements(final InferTypeConstraintsModel model, final RefactoringStatus status, final IProgressMonitor monitor)
            throws CoreException {
        try {
            monitor.beginTask("", 300);
            monitor.setTaskName(InferTypeRefactoring.MSG_CREATING_CHANGES);

            final IProgressMonitor subMonitor = new SubProgressMonitor(monitor, 320);
            try {

                for (ConstraintVariable2 baseVariable : model.getConstraintVariables()) {
                    final TCType newType = InferTypeConstraintsModel.getNewType(baseVariable);
                    
                    if (newType != null) {
                        if (baseVariable instanceof ElementVariable) {
                            //LogUtil.debugOut2(this, "ELEMENT VAR WITH TYPE " + newType);
                            Stack<ElementVariable> stack = new Stack<ElementVariable>();
                            
                            ConstraintVariable2 parentCv = baseVariable;                        
                            while (parentCv instanceof ElementVariable) {
                                stack.push((ElementVariable)parentCv);
                                parentCv = ((ElementVariable)parentCv).getParentConstraintVariable();
                            }
                            if (parentCv instanceof ITypeConstraintVariable) {
                                CompilationUnitRange range = ((ITypeConstraintVariable)parentCv).getRange();
                                final ICompilationUnit compilationUnit = range.getCompilationUnit();
                                CompilationUnitRewrite rewrite = getCompilationUnitRewrite(compilationUnit);
        
                                ASTNode result = NodeFinder.perform(rewrite.getRoot(), ((ITypeConstraintVariable)parentCv).getRange().getSourceRange());
                                if (result != null && result instanceof ParameterizedType) {
                                    while (!stack.isEmpty() && result != null && result instanceof ParameterizedType) {
                                        ParameterizedType parameterizedType = (ParameterizedType)result;
                                        List arguments = parameterizedType.typeArguments();
                                        ElementVariable elementVariable = stack.pop();
                                        result = getParameterNode(parameterizedType, arguments, elementVariable);
                                    }
                                    
                                    if (result != null && stack.isEmpty()) {
                                        if (result instanceof WildcardType) {
                                            result = ((WildcardType)result).getBound();
                                        }
                                        if (result instanceof ArrayType) {
                                            result = ((ArrayType)result).getElementType();
                                        }
    
                                        if (result != null) {
                                            //LogUtil.debugOut2(this, "HAS NODE: " + result + ", " + result.getClass());
                                            numberOfChanges++;
                                            rewrite.getImportRemover().registerRemovedNode(result);
                                            rewrite.getASTRewrite().replace(result, createCorrespondingNode(rewrite, newType, 
                                                    baseVariable.getType()),
                                                    rewrite.createCategorizedGroupDescription("Update type occurences", SET_INFER_TYPE));
                                        }
                                    }
                                } else {
                                    // this is a raw reference and doesn't need changes
                                }
                            }
                                
                                
                        } else if (baseVariable instanceof ITypeConstraintVariable) {                        
                            CompilationUnitRange range = ((ITypeConstraintVariable) baseVariable).getRange();
                            final ICompilationUnit compilationUnit = range.getCompilationUnit();
                            CompilationUnitRewrite rewrite = getCompilationUnitRewrite(compilationUnit);
    
                            ASTNode result = NodeFinder.perform(rewrite.getRoot(), ((ITypeConstraintVariable) baseVariable).getRange().getSourceRange());
                            
                            if (result instanceof WildcardType) {
                                result = ((WildcardType)result).getBound();
                            }
                            if (result instanceof ArrayType) {
                                result = ((ArrayType)result).getElementType();
                            }                            

                            if (result != null) {
                                numberOfChanges++;
                                rewrite.getImportRemover().registerRemovedNode(result);
                                rewrite.getASTRewrite().replace(result, createCorrespondingNode(rewrite, newType, 
                                        baseVariable.getType()),
                                        rewrite.createCategorizedGroupDescription("Update type occurences", SET_INFER_TYPE));
                            }
                        }
                    }
                }

            } finally {
                subMonitor.done();
            }
        } finally {
            monitor.done();
        }
    }

    private ASTNode getParameterNode(ParameterizedType parameterizedType, List arguments, ElementVariable elementVariable) {
        ITypeBinding[] typeParameters = parameterizedType.resolveBinding().getTypeDeclaration().getTypeParameters();
        for (int i = 0; i < typeParameters.length; i++) {
            if (typeParameters[i].getKey().equals(elementVariable.getTypeVariableKey())) {
                ASTNode node = (ASTNode)arguments.get(i);
                if (node instanceof WildcardType) {
                    node = ((WildcardType)node).getBound();
                }
                if (node instanceof ArrayType) {
                    node = ((ArrayType)node).getElementType();
                }
                return node;
            }
        }
        return null;
    }

    private CompilationUnitRewrite getCompilationUnitRewrite(ICompilationUnit compilationUnit) {
        CompilationUnitRewrite rewrite = unitToRewrite.get(compilationUnit);
        if (rewrite == null) {
            rewrite = new CompilationUnitRewrite(compilationUnit);
            unitToRewrite.put(compilationUnit, rewrite);
        }
        return rewrite;
    }

    private static ASTNode createCorrespondingNode(final CompilationUnitRewrite rewrite, final TCType newType, TType oldType) {
        return rewrite.getImportRewrite().addImportFromSignature(newType.getSignature(oldType), rewrite.getAST());
    }

    private void introduceTypes(RefactoringStatus status) throws CoreException, InternalCalculationException {
        List<TypeCreator> typeCreators = new ArrayList<TypeCreator>();
        
        for (NewTCType newTCType : typeHierarchy.getNewTypes()) {
            TypeCreator typeCreator = new TypeCreator(newTCType, newTCType.getBaseTypeBinding().getJavaElement().getJavaProject());
            typeCreator.createTheType();
            typeCreators.add(typeCreator);
        }

        final IJavaProject project = selectedElement.getBinding().getJavaElement().getJavaProject();
        if (BindingUpdater.updateMethodBindings(typeCreators, typeHierarchy, project, unitToRewrite)) {
            for (TypeCreator typeCreator : typeCreators) {
                typeCreator.createTheType();
            }
            
        }

        final IFile file = ResourceUtil.getFile(selectedElement.getCompilationUnit());
        for (TypeCreator typeCreator : typeCreators) {
            newCuChanges.add(new CreateCompilationUnitChange(typeCreator.getCompilationUnit(), typeCreator.getContent(), file.getCharset(false)));
        }
    }

    private void rewriteHierarchy(IJavaProject javaProject, RefactoringStatus status) throws CoreException {
        for (ExistingTCType modifiedType : typeHierarchy.getModifiedTypes()) {
            IType iType = modifiedType.getIType();

            final CompilationUnitRewrite sourceRewrite = getCompilationUnitRewrite(iType.getCompilationUnit());

            try {
                final AbstractTypeDeclaration declaration = ASTNodeSearchUtil.getAbstractTypeDeclarationNode(iType, sourceRewrite.getRoot());
                if (declaration != null) {
                    createTypeSignature(sourceRewrite, declaration, status, modifiedType);
                }
            } catch (Exception e) {
                System.err.println("Could not change hierarchy of " + iType.getKey() + ": " + e);
                status.addError("Could not change hierarchy of " + iType.getKey() + ": " + e);
                e.printStackTrace();
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void createTypeSignature(final CompilationUnitRewrite rewrite, final AbstractTypeDeclaration declaration, final RefactoringStatus status, ExistingTCType modifiedType)
            throws JavaModelException, InternalCalculationException {
        Assert.isNotNull(rewrite);
        Assert.isNotNull(declaration);
        Assert.isNotNull(status);
        Assert.isNotNull(modifiedType);

        IType subType = modifiedType.getIType();

        final AST ast = declaration.getAST();
        final ITypeParameter[] parameters = subType.getTypeParameters();

        final ASTRewrite rewriter = rewrite.getASTRewrite();

        ListRewrite listRewrite = null;        
        if (declaration instanceof TypeDeclaration) {
            listRewrite = rewriter.getListRewrite(declaration, TypeDeclaration.SUPER_INTERFACE_TYPES_PROPERTY);
        } else if (declaration instanceof EnumDeclaration) {
            listRewrite = rewriter.getListRewrite(declaration, EnumDeclaration.SUPER_INTERFACE_TYPES_PROPERTY);
        }
        
        
        if (listRewrite != null) { 
            CategorizedTextEditGroup descr = rewrite.createCategorizedGroupDescription(RefactoringCoreMessages.ExtractInterfaceProcessor_add_super_interface, SET_INFER_TYPE);
            for (TCType newType : modifiedType.getNewInterfaces()) {
                Type type = getNewSuperType(ast, parameters, newType, modifiedType, rewrite);
                listRewrite.insertLast(type, descr);
            }
            if (!modifiedType.getRemovedInterfaces().isEmpty()) {
                descr = rewrite.createCategorizedGroupDescription("Remove interface", SET_INFER_TYPE);
                List<Type> originalList = listRewrite.getOriginalList();
                for (Type node : originalList) {
                    if (modifiedType.getRemovedInterfaces().contains(node.resolveBinding().getKey())) {
                        rewrite.getImportRemover().registerRemovedNode(node);
                        listRewrite.remove(node, descr);
                    }
                }
            }
        }
        
        TCType superType = modifiedType.getModifiedSuperType();
        if (superType != null && !superType.isObject()) {
            Type type = getNewSuperType(ast, parameters, superType, modifiedType, rewrite);
            rewriter.set(declaration, TypeDeclaration.SUPERCLASS_TYPE_PROPERTY, type, rewrite.createCategorizedGroupDescription("Change super type", SET_INFER_TYPE));
        }

    }

    @SuppressWarnings("unchecked")
    private Type getNewSuperType(final AST ast, final ITypeParameter[] parameters, TCType newType, ExistingTCType modifiedType, CompilationUnitRewrite rewrite) throws JavaModelException, InternalCalculationException {
        String completeTypeSignature = newType.getHierarchyTypeSignature(modifiedType);
        if (completeTypeSignature == null) {
            Type type = ast.newSimpleType(ast.newSimpleName(rewrite.getImportRewrite().addImport(newType.getTypeName())));
            if (parameters.length > 0) {
                final ParameterizedType parameterized = ast.newParameterizedType(type);
                for (ITypeParameter element : parameters) {
                    parameterized.typeArguments().add(ast.newSimpleType(ast.newSimpleName(element.getElementName())));
                }
                type = parameterized;
            }
            return type;
        } else {
            return rewrite.getImportRewrite().addImportFromSignature(completeTypeSignature, rewrite.getAST());
        }
    }

    /**
     * Returns the number of changed type occurences used for statistics.
     * 
     * @return the number
     */
    public int getNumberOfChanges() {
        return numberOfChanges ;
    }
}
