/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.manipulation;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.core.formatter.CodeFormatter;
import org.eclipse.jdt.internal.corext.codemanipulation.StubUtility;
import org.eclipse.jdt.internal.corext.util.JavaModelUtil;
import org.eclipse.jdt.internal.formatter.DefaultCodeFormatter;
import org.eclipse.jdt.ui.CodeGeneration;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.text.edits.TextEdit;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.NewTCType;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * 
 * This class creates a new (inferred) type.
 * 
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */
public class TypeCreator {

	private final IJavaProject javaProject;
    private final NewTCType newTCType;

    private String content;
    private ICompilationUnit compilationUnit;
        
   	/**
     * The single constructor.
     * 
   	 * @param newTCType the type to be created
   	 * @param javaProject the java project of the selection
   	 */
   	public TypeCreator(NewTCType newTCType, IJavaProject javaProject) {
        this.newTCType = newTCType;
        this.javaProject = javaProject;
    }

    /**
     * Retrieves the content of the new CU.
     * 
     * @return String
     */
    public String getContent() {
        return content;
    }

    /**
     * Retrieves the new and filled CU.
     * 
     * @return ICompilationUnit
     */
    public ICompilationUnit getCompilationUnit() {
        return compilationUnit;
    }

    /**
	 * Does the job.
     * 
	 * @throws CoreException
	 */
	public void createTheType() throws CoreException {

		try {
            compilationUnit = GenerationUtil.createCompilationUnit(newTCType.getTypeName(), javaProject);
            ICompilationUnit cuWorkingCopy = compilationUnit.getWorkingCopy(null);

			// building the content
			try {
                ImportRewrite rewrite= StubUtility.createImportRewrite(cuWorkingCopy, false);

                StringBuffer contents = new StringBuffer();

				//getting the type comment
				String typeComment = CodeGenerator.getTypeComment(newTCType.getTypeName(), compilationUnit);

				//generating the type-beginning
				contents.append(CodeGenerator.getTypeBeginning(newTCType, rewrite));

				// getting method declarations
				contents.append(CodeGenerator.getMethodDeclarations(newTCType, rewrite));

				//add class-ending
				contents.append(CodeGenerator.getLineDelimitter()).append("}");

				//setting content
				content = CodeGeneration.getCompilationUnitContent(compilationUnit, typeComment, contents.toString(), CodeGenerator.getLineDelimitter());

                cuWorkingCopy.getBuffer().setContents(content);
                
                TextEdit edits= rewrite.rewriteImports(null);
                JavaModelUtil.applyEdit(cuWorkingCopy, edits, false, null);
                    
                content = cuWorkingCopy.getSource();

			} finally {
                cuWorkingCopy.discardWorkingCopy();
            }

			// Format code
			try {
                CodeFormatter formatter = new DefaultCodeFormatter();
                TextEdit edit = formatter.format(CodeFormatter.K_COMPILATION_UNIT, content, 0, content.length(), 0, CodeGenerator.getLineDelimitter());

                Document doc = new Document(content);
                if (edit != null) {
                    edit.apply(doc);
                }
                content = doc.get();
			} catch (BadLocationException e) {
                throw new InternalCalculationException(e.toString());
			}
		} catch (JavaModelException e) {
            throw new InternalCalculationException(e.toString());
		}
	}
}