/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.jdt.core.dom.IMethodBinding;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.util.BindingUtil;

/**
 * This class holds the wrappers for a set of instance methods.
 * <p>
 * The class is not derived from Set on purpose, because it has a different
 * behaviour (and partly different method signatures) for all operations where comparisons
 * are involved to meet the covariance rules of Java 5.
 * </p>
 * @author Hannes Kegel
 *
 */
public class BindingSet implements Iterable<BindingWrapper> {

    private Set<BindingWrapper> store = new HashSet<BindingWrapper>();

    /**
     * Adds a new binding wrapper to this set. If a method with an override
     * relationship is already in this set, the one with the most specific return
     * type will be in this set afterwards.
     * 
     * @param wrapper a binding wrapper
     * @return true if the wrapper was added
     */
    public boolean add(BindingWrapper wrapper) {
        BindingWrapper previous = findElement(wrapper, CovarianceType.FIND_SUPERTYPE);
        if (previous != null) {
            store.remove(previous);
            return store.add(wrapper);            
        } else {
            previous = findElement(wrapper, CovarianceType.RECEIVER_POSSIBLE_SUBTYPE);
            if (previous == null) {
                return store.add(wrapper);                                
            } else if (BindingUtil.isFirstSubTypeCompatible(wrapper.getMethodBinding().getDeclaringClass(), 
                            previous.getMethodBinding().getDeclaringClass())) 
            {
                store.remove(previous);
                return store.add(wrapper);                
            } else {
                return false;
            }
        }
    }

    /**
     * Clear the set.
     */
    public void clear() {
        store.clear();
    }

    /**
     * Checks if this set equals the argument using the given CovarianceType.
     * 
     * @param otherSet the set to be compared
     * @param covarianceType the possible covariance relationship between the methods in the sets
     * @return true if equal
     */
    public boolean equals(BindingSet otherSet, CovarianceType covarianceType) {
        if (otherSet.size() != size()) {
            return false;
        }
        
        for (BindingWrapper bindingWrapper : otherSet) {
            if (!contains(bindingWrapper, covarianceType)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks if this set is empty.
     * 
     * @return true if empty
     */
    public boolean isEmpty() {
        return size() == 0;
    }

    /**
     * Returns the number of methods in this set.
     * 
     * @return int
     */
    public int size() {
        return store.size();
    }

    /**
     * Returns an array of the {@link BindingWrapper} instances in this set.
     * @return an array containing BindingWrappers
     */
    public Object[] toArray() {
        return store.toArray();
    }

    /**
     * Remove all methods from the given set in this set. CovarianceType.NONE used for comparisons.
     * 
     * @param allSuperMethods the methods to be removed
     */
    public void removeAllSuperMethods(BindingSet allSuperMethods) {
        for (BindingWrapper bindingWrapper : allSuperMethods) {
            BindingWrapper storeElement = findElement(bindingWrapper, CovarianceType.NONE);
            if (storeElement != null) {
                store.remove(storeElement);
            }
        }
    }
    
    /**
     * Checks if this set contains an equal wrapper or a wrapper with an override
     * relationship as specified by covarianceType.
     *  
     * @param wrapper the wrapper to search for
     * @param covarianceType the type of covariance relationship allowed
     * @return true if contained
     */
    public boolean contains(BindingWrapper wrapper, CovarianceType covarianceType) {
        return findElement(wrapper, covarianceType) != null;
    }

    /**
     * Check if all given methods are contained in this set. Works like 
     * {@link #contains(BindingWrapper, CovarianceType)}.
     * 
     * @param allMethods the set containing the methods to check for
     * @param covarianceType the type of covariance relationship allowed
     * @return true if all are contained
     */
    public boolean containsAll(BindingSet allMethods, CovarianceType covarianceType) {
        for (BindingWrapper bindingWrapper : allMethods) {
            if (!contains(bindingWrapper, covarianceType)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns a collection of the {@link BindingWrapper} instances contained in this set.
     * @return an unmodifiable collection
     */
    public Collection<BindingWrapper> getCollection() {
        return Collections.unmodifiableCollection(store);
    }

    /**
     * Add all methods from the given set as specified in {@link #add(BindingWrapper)}.
     * @param members the set containing the methods.
     */
    public void addAll(BindingSet members) {
        if (members == this) {
            return;
        }
        for (BindingWrapper bindingWrapper : members) {
            add(bindingWrapper);
        }
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj) {
        throw new UnsupportedOperationException("only use equals(BindingSet, CovarianceType)");
    }

    /* (non-Javadoc)
     * @see java.lang.Iterable#iterator()
     */
    public Iterator<BindingWrapper> iterator() {
        return store.iterator();
    }    
    
    private BindingWrapper findElement(BindingWrapper otherWrapper, CovarianceType covarianceType) {
        IMethodBinding otherBinding = otherWrapper.getMethodBinding();

        for (BindingWrapper wrapper : store) {
            IMethodBinding binding = wrapper.getMethodBinding();
            if (binding.getKey().equals(otherBinding.getKey())) {
                return covarianceType == CovarianceType.FIND_SUPERTYPE ? null : wrapper;
            } else if (binding.isSubsignature(otherBinding)) {
                String otherRetKey = otherBinding.getReturnType().getKey();
                String thisRetKey = binding.getReturnType().getKey();
                if (otherRetKey.equals(thisRetKey)) {
                    return covarianceType == CovarianceType.FIND_SUPERTYPE ? null : wrapper;
                } else  {
                    switch (covarianceType) {
                        case RECEIVER_POSSIBLE_SUBTYPE:
                            if (binding.getReturnType().isSubTypeCompatible(otherBinding.getReturnType())) {
                                return wrapper;
                            }
                            break;
                        case ARGUMENT_POSSIBLE_SUBTYPE:
                        case FIND_SUPERTYPE:
                            if (otherBinding.getReturnType().isSubTypeCompatible(binding.getReturnType())) {
                                return wrapper;
                            }
                            break;
                        case NONE:
                    }
                    return null;
                }
            }
        }
        return null;
    }       
    
}
