/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.jdt.core.BindingKey;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.InferTypePlugin;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;
import org.intoJ.inferType3.internal.typeCalc.CovarianceType;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * The abstract class for the types in the type hierarchy of the selection.
 * <p>
 * Used for the calculation of the types to be used and to hold the IMethodBindings
 * declared in the hierarchy.
 * </p>
 * @author Hannes Kegel
 *
 */
public abstract class TCType {   

    /** The supertype of this type. null for interfaces */
    protected TCType superType;
    /** All declared interfaces of this type */
    protected Set<TCType> interfaces;
    /** 
     * The type binding this type is based on. For existing types
     * this is the binding of the type. For new types it is the binding of the
     * type the new type is derived from.
     */
    protected ITypeBinding baseTypeBinding;
    
    private TType argumentTType;

    /**
     * Returns the fully qualified name of this type.
     * @return the name
     */
    public abstract String getTypeName();

    /**
     * Returns all public instance methods available for this type.
     * @return a BindingSet with all methods including the methods declared in super types
     */    
    public abstract BindingSet getAllMethods();

    /**
     * Checks if this type is an interface.
     * @return true if this type is an interface
     */    
    public abstract boolean isInterface();
    
    /**
     * Checks if the type calculation could move the type up or down in the hierarchy. 
     * @return true if it can be moved
     */    
    public abstract boolean canBeMovedInHierarchy();
    
    /**
     * Checks if this type can be modified by the refactoring 
     * (that means if it is a new or a source type).
     * 
     * @return true if modifiable
     */
    public abstract boolean isModifiable();
    
    /**
     * Checks if this type has declared members besides public
     * instance methods.
     * 
     * @return true if there are any other declared members
     */
    public abstract boolean hasOtherDeclaredMembers();
    
    /**
     * The single constructor.
     * 
     * @param baseTypeBinding the base type binding
     */
    public TCType(ITypeBinding baseTypeBinding) {
        this.baseTypeBinding = baseTypeBinding;
    }
                
    /**
     * Sets the type that the type arguments are based on.
     * 
     * @param argumentTType a TType
     */
    public void setArgumentTType(TType argumentTType) {
        this.argumentTType = argumentTType;
    }
    
    /**
     * Returns the binding this type is based on.
     * 
     * @return ITypeBinding
     */
    public ITypeBinding getBaseTypeBinding() {
        return baseTypeBinding;
    }
    
    /**
     * Sets the binding this type is based on.
     * 
     * @param baseTypeBinding the base type binding 
     */
    public void setBaseTypeBinding(ITypeBinding baseTypeBinding) {
        this.baseTypeBinding = baseTypeBinding;
    }
    
    /**
     * Checks if this type is generic or parameterized.
     * 
     * @return true if generic or parameterized
     */
    public boolean isGenericType() {
        return baseTypeBinding.isGenericType() || baseTypeBinding.isParameterizedType();        
    }            
    
    /**
     * Retrieves the name without package of this type.
     * 
     * @return the simple name
     */
    public String getSimpleName() {
        int pos = getTypeName().lastIndexOf('.');
        return pos == -1 ? getTypeName() : getTypeName().substring(pos + 1);
    }
    
    /**
     * Checks if the methods of this type are contained in the given methods.
     * 
     * @param otherMethods the methods that should contain all methods of this type
     * @return true if otherMethods contain all methods of this type
     * @throws JavaModelException
     */
    public boolean isContainedIn(BindingSet otherMethods) throws JavaModelException {
        return otherMethods.containsAll(getAllMethods(), CovarianceType.NONE);
    }

    /**
     * Checks if this type fully contains the given other type. 
     * 
     * @param otherType the type to check
     * @return true if this type contains all public methods of otherType and
     * otherType has no additional declared members
     */
    public boolean containsAll(TCType otherType) {
        return !otherType.hasOtherDeclaredMembers() && getAllMethods().containsAll(otherType.getAllMethods(), CovarianceType.NONE);
    }

    /**
     * Returns the declared super interfaces of this type.
     * 
     * @return the super interfaces, empty set if there are none
     */
    public Set<TCType> getInterfaces() {
        return interfaces;
    }
    
    /**
     * Returns the super type of this type.
     * 
     * @return the super type or null if there is none
     */
    public TCType getSuperType() {
        return superType;
    }
    
    /**
     * Retrieves all supertypes.
     * 
     * @return all superclasses and all interfaces in the super hierarchy
     */
    public Set<TCType> getAllSupertypes() {
        Set<TCType> allSupertypes = new HashSet<TCType>();

        if (superType != null) {
            allSupertypes.add(superType);
            allSupertypes.addAll(superType.getAllSupertypes());
        }
        
        for (TCType currentIf : interfaces) {
            allSupertypes.add(currentIf);
            allSupertypes.addAll(currentIf.getAllSupertypes());            
        }

        return allSupertypes;
    }

    /**
     * Retrieves all declared supertypes.
     * 
     * @return the superclass and the declared interfaces
     */
    public Set<TCType> getDeclaredSupertypes() {
        Set<TCType> declaredSupertypes = new HashSet<TCType>();

        if (superType != null) {
            declaredSupertypes.add(superType);
        }
        declaredSupertypes.addAll(interfaces);

        return declaredSupertypes;
    }
    
    /**
     * Sets a new supertype for this type.
     * 
     * @param superType the new supertype
     * @param status the refactoring status used to record errors
     * @throws JavaModelException
     */
    public void setSuperType(TCType superType, RefactoringStatus status) throws JavaModelException {
        if (!isModifiable()) {
            status.addFatalError("tried to set supertype of non-modifiable type " + getTypeName() + " to " + superType.getTypeName());
        }
        if (this.superType == null || !this.superType.getAllSupertypes().contains(superType)) {
            this.superType = superType;
            if (superType instanceof ExistingTCType) {                
                Cache.clearHierarchyCache((IType)((ExistingTCType)superType).getBaseTypeBinding().getJavaElement());
            }
        }
    }
    
    /**
     * Adds a new interface to this type if necessary.
     *   
     * @param newInterface the interface to be added
     * @param status the refactoring status used to record errors
     * @return true if the interface was added
     * @throws JavaModelException
     */
    public boolean addInterface(TCType newInterface, RefactoringStatus status) throws JavaModelException {
        for (Iterator<TCType> iter = interfaces.iterator(); iter.hasNext();) {
            TCType interf = iter.next();
            if (interf.getAllSupertypes().contains(newInterface)) {
                return false;
            }
            if (newInterface.getAllSupertypes().contains(interf)) {
                removeInterfaceInt(interf);
                iter.remove();
            }
        }
        if (!isModifiable()) {
            status.addFatalError("tried to add new interface " + newInterface.getTypeName() + " to non-modifiable type " + getTypeName());
            return false;
        }
        if (interfaces.add(newInterface)) {
            if (newInterface instanceof ExistingTCType) {
                Cache.clearHierarchyCache((IType)((ExistingTCType)newInterface).getBaseTypeBinding().getJavaElement());
            }
            return true;            
        } else {
            return false;
        }
    }
    
    /**
     * Remove the given interface from the declared interfaces.
     * 
     * @param interf the interface to be removed
     */
    public void removeInterface(TCType interf) {
        removeInterfaceInt(interf);
        interfaces.remove(interf);
    }
    
    /**
     * Handle subclass stuff for interface removal.
     * 
     * @param interf the interface to be removed
     */
    protected void removeInterfaceInt(TCType interf) {
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return getTypeName();
    }
    
    /**
     * Checks if this type has accessible members besides the public
     * instance methods.
     * 
     * @return true if there are other accessible members
     */
    public boolean hasOtherAccessibleMembers() {
        Set<TCType> allSupertypes = getDeclaredSupertypes();
        for (TCType type : allSupertypes) {
            if (type.hasOtherAccessibleMembers()) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Checks if this is java.lang.Object.
     * 
     * @return true if it is the Object class 
     */
    public boolean isObject() {
        return false;
    }
    
    /**
     * Retrieves the name of this type for usage in a supertype declaration 
     * of the first parameter. Needed for generic types.
     * 
     * @param subTCType the subtype this type will be a declared supertype
     * @param importRewrite the import rewrite used to get simple names
     * @return the name for direct use in java source files 
     * @throws JavaModelException
     * @throws InternalCalculationException
     */
    public String getHierarchyTypeName(TCType subTCType, ImportRewrite importRewrite) throws JavaModelException, InternalCalculationException {
        ParameterizedTypeUtil.ImportRewriteProxy importRewriteProxy = new ParameterizedTypeUtil.ImportRewriteProxy(importRewrite);

        if (baseTypeBinding.isGenericType()) {
            ITypeBinding parametrizedTypeBinding = ParameterizedTypeUtil.getParametrizedTypeBinding(baseTypeBinding.getKey(), subTCType.baseTypeBinding);
            String typeName = importRewriteProxy.addImport(getTypeName()) + ParameterizedTypeUtil.getTypeArguments(parametrizedTypeBinding, importRewriteProxy);
            return typeName;
        } else {
            return importRewriteProxy.addImport(getTypeName());            
        }
    }

    /**
     * Retrieves the signature of this type for usage in a supertype 
     * declaration of the parameter. Needed for generic types.
     * 
     * @param subTCType the subtype this type will be a declared supertype
     * @return the signature with dots usable for ImportRewrite
     * @throws JavaModelException
     * @throws InternalCalculationException
     */
    public String getHierarchyTypeSignature(TCType subTCType) throws JavaModelException, InternalCalculationException {
        String nameSig = "L" + getTypeName();
        if (baseTypeBinding.isGenericType()) {
            ITypeBinding parametrizedTypeBinding = ParameterizedTypeUtil.getParametrizedTypeBinding(baseTypeBinding.getKey(), subTCType.baseTypeBinding);
            if (parametrizedTypeBinding != null && parametrizedTypeBinding.isParameterizedType()) {
                String baseSig = new BindingKey(parametrizedTypeBinding.getKey()).toSignature();
                return nameSig + baseSig.substring(baseSig.indexOf('<'));
            } else {
                return null;
            }
        } else {
            return nameSig + ";";
        }
    }

    /**
     * Retrieves the type signature for this type. Needed for generic types.
     * Wildcard usage is based on oldType.
     * 
     * @param oldType the type previously used to retain wildcard usage
     * @return the signature with dots usable for ImportRewrite
     */
    public String getSignature(TType oldType) {
        return "L" + getTypeName() + ParameterizedTypeUtil.getTypeArgumentSignature(argumentTType, oldType) + ";";
    }

    /**
     * Check if there are other members that should be hidden for minimal types.
     * 
     * @return true if there are other members
     */
    public boolean hasOtherMembersToBeHidden() {
        return hasOtherAccessibleMembers() || 
                (InferTypePlugin.getDefault().getPreferenceStore().getBoolean(InferTypePlugin.PREFNAME_CONSIDER_PRIVATE) 
                        && hasOtherDeclaredMembers());
    }
}
