/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.util;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;

/**
 * Utility class for {@link ITypeBinding}.
 * 
 * @author Hannes Kegel
 *
 */
public class BindingUtil {

    /**
     * Find the type with the given qualified name in the supertype hierarchy of the binding.
     * 
     * @param hierarchyType the binding whose supertype hierarchy is searched
     * @param fullyQualifiedTypeName the name of the type we are looking for
     * @return the binding of the type with the given name or null
     */
    public static ITypeBinding findTypeInHierarchy(ITypeBinding hierarchyType, String fullyQualifiedTypeName) {
        if (hierarchyType.isArray() || hierarchyType.isPrimitive()) {
            return null;
        }
        if (fullyQualifiedTypeName.equals(hierarchyType.getTypeDeclaration().getQualifiedName())) {
            return hierarchyType;
        }
        ITypeBinding superClass= hierarchyType.getSuperclass();
        if (superClass != null) {
            ITypeBinding res= findTypeInHierarchy(superClass, fullyQualifiedTypeName);
            if (res != null) {
                return res;
            }
        }
        ITypeBinding[] superInterfaces= hierarchyType.getInterfaces();
        for (ITypeBinding element : superInterfaces) {
            ITypeBinding res= findTypeInHierarchy(element, fullyQualifiedTypeName);
            if (res != null) {
                return res;
            }           
        }
        return null;
    }

    /**
     * Checks if the first type is subtype compatible with the second type. That means
     * that the first type could be assigned to the second. Extra method because bindings 
     * can be from different AST parses and then the internal method does not work correctly.
     * 
     * @param subType the possible subtype
     * @param superType the possible superType
     * @return true if the types are identical or subtype compatible
     */
    public static boolean isFirstSubTypeCompatible(ITypeBinding subType, ITypeBinding superType) {        
        if (subType == null || superType == null) {
            return false;
        }

        if (superType.getQualifiedName().equals("java.lang.Object")) {
            return true;
        }
        
        if (superType.isWildcardType()) {
            if (superType.getBound() == null) {
                return true;
            } else if (superType.isUpperbound()) {
                superType = superType.getBound();
            }
        }
        
        if (subType.getKey().equals(superType.getKey())) {
            return true;
        }
        if (subType.isParameterizedType() && superType.isParameterizedType()
                && subType.getErasure().getKey().equals(superType.getErasure().getKey())) {
            ITypeBinding[] subTypeArgs = subType.getTypeArguments();
            ITypeBinding[] superTypeArgs = superType.getTypeArguments();
            if (subTypeArgs.length == superTypeArgs.length) {
                for (int i=0; i<subTypeArgs.length; i++) {
                    if (!isFirstSubTypeCompatible(subTypeArgs[i], superTypeArgs[i])) {
                        return false;
                    }
                }
                return true;                
            }
        }
        
        if (isFirstSubTypeCompatible(subType.getSuperclass(), superType)) {
            return true;
        }
        
        for (ITypeBinding interf : subType.getInterfaces()) {
            if (isFirstSubTypeCompatible(interf, superType)) {
                return true;
            }            
        }
        
        return false;
    }
    
    
    public static boolean isFirstSubclassOrEqual(ITypeBinding subType, ITypeBinding possibleSupertype) {
        if (subType == null || possibleSupertype == null) {
            return false;
        }

        if (possibleSupertype.getQualifiedName().equals("java.lang.Object")) {
            return true;
        }

        if (subType.getTypeDeclaration().getKey().equals(possibleSupertype.getTypeDeclaration().getKey())) {
    		return true;
    	}
    	
		ITypeBinding superClass = subType.getSuperclass();
		if (superClass != null) {
			return isFirstSubclassOrEqual(superClass, possibleSupertype);
		}
		return false;
    }
    

}
