/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.util;

import java.util.Iterator;
import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AnnotationTypeDeclaration;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;

/**
 * This class provides methods to retrieve IBindings for IJavaElements.
 * 
 * @author Philip Mayer
 * @author Hannes Kegel
 */
public class ConversionUtil {

    /**
     * Retrieves an IBinding for the given IMember. There will be AST parses
     * so be careful when calling this method.
     * 
     * @param element the member we want a binding for
     * @return a {@link Result} object containing the binding, the start pos of the type and
     * the binding of java.lang.Object
     * @throws InternalCalculationException
     */
    public static Result findBinding(IMember element) throws InternalCalculationException {
        Result result = new Result();
        
        if ((element.getElementType() != IJavaElement.FIELD) && (element.getElementType() != IJavaElement.METHOD) && (element.getElementType() != IJavaElement.TYPE)) {
            throw new InternalCalculationException("Element type not supported: " + element.getElementType() + " in IJavaElementToBinding.getBinding()");
        }

        ICompilationUnit icu = element.getCompilationUnit();
        if (icu == null) {
            throw new InternalCalculationException("Type can not be parsed.");
        }

        ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setSource(icu);
        parser.setResolveBindings(true);
        CompilationUnit cu = (CompilationUnit) parser.createAST(null);
        
        result.objectBinding = cu.getAST().resolveWellKnownType("java.lang.Object");

        Visitor vis = new Visitor(element);
        cu.accept(vis);
        result.binding = vis.getBinding();
        if (result.binding != null) {
            result.typeStartPos = vis.getTypeStartPos();
            return result;
        }
        throw new InternalCalculationException("Please select a field or a method; use the editor for local variables.");
    }

    /**
     * This methods returns the corresponding ITypeBinding from a given IType.
     * 
     * @param type the type we want a binding for
     * @return the corresponding ITypeBinding
     * @throws InternalCalculationException
     *             is thrown if type could not be parsed (f.e. if type is a
     *             binary one)
     */
    public static ITypeBinding getTypeBinding(IType type) throws InternalCalculationException {

        return (ITypeBinding)findBinding(type).getBinding();
    }

    private static class Visitor extends ASTVisitor {

        private IBinding binding;
        private IMember element;
        private int typeStartPos;

        public Visitor(IMember element) {

            this.element = element;
        }

        public int getTypeStartPos() {
            return typeStartPos;
        }

        @Override
        public boolean visit(FieldDeclaration node) {

            if (node != null) {
	            List fragments = node.fragments();
	            Iterator i = fragments.iterator();
	            while (i.hasNext()) {
	                VariableDeclarationFragment currentFragment = (VariableDeclarationFragment) i.next();
	                IVariableBinding currentFieldBinding = currentFragment.resolveBinding();
	                if (isElementBinding(currentFieldBinding)) {
	                    binding = currentFieldBinding;
	                    typeStartPos = node.getType().getStartPosition();
	                }
	            }
            }	
            return super.visit(node);

        }

        @Override
        public boolean visit(MethodDeclaration node) {
            if (node != null) {
	            IMethodBinding currentMethodBinding = node.resolveBinding();
	            if (isElementBinding(currentMethodBinding)) {
	                binding = currentMethodBinding;
	                typeStartPos = node.getReturnType2().getStartPosition();
	            }
            }
            return super.visit(node);
        }
        
        

        @Override
        public void endVisit(EnumDeclaration node) {
            if (node != null) {
	            ITypeBinding currentTypeBinding = node.resolveBinding();
	            if (isElementBinding(currentTypeBinding)) {
	                binding = currentTypeBinding;
	            }
            }
            
        }

        @Override
        public void endVisit(AnnotationTypeDeclaration node) {
            if (node != null) {
	            ITypeBinding currentTypeBinding = node.resolveBinding();
	            if (isElementBinding(currentTypeBinding)) {
	                binding = currentTypeBinding;
	            }
            }
        }
        
        

        @Override
		public boolean visit(AnnotationTypeDeclaration node) {
            if (node != null) {
	            ITypeBinding currentTypeBinding = node.resolveBinding();
	            if (isElementBinding(currentTypeBinding)) {
	                binding = currentTypeBinding;
	            }
            }
            return super.visit(node);
		}

		@Override
		public boolean visit(AnonymousClassDeclaration node) {
            if (node != null) {
	            ITypeBinding currentTypeBinding = node.resolveBinding();
	            if (isElementBinding(currentTypeBinding)) {
	                binding = currentTypeBinding;
	            }
            }
            return super.visit(node);
		}

		@Override
        public boolean visit(TypeDeclaration node) {
            if (node != null) {
	            ITypeBinding currentTypeBinding = node.resolveBinding();
	            if (isElementBinding(currentTypeBinding)) {
	                binding = currentTypeBinding;
	            }
            }
            return super.visit(node);
        }

        /**
         * @return Returns the binding.
         */
        public IBinding getBinding() {
            return binding;
        }
        
        /**
         * Returns if the given binding is the binding of this Visitor's
         * element.
         * 
         * @param binding the binding to check
         * @return <code>false</code> if the binding is <code>null</code> or
         * not the corresponding binding of the element, otherwise
         * <code>true</code>
         */
        private boolean isElementBinding(IBinding binding) {
        	return binding != null && binding.getJavaElement().equals(element);
        }

    }
    
    /**
     * The result class for the above methods.
     * 
     * @author Hannes Kegel
     *
     */
    public static class Result {
        private IBinding binding;
        private int typeStartPos;        
        private ITypeBinding objectBinding;

        /**
         * Returns the requested binding
         * @return a {@link ITypeBinding}
         */
        public IBinding getBinding() {
            return binding;
        }
        
        /**
         * Returns the type (not the identifier) start position of the requested binding
         * @return int
         */
        public int getTypeStartPos() {
            return typeStartPos;
        }
        
        /**
         * Returns the type binding of java.lang.Object
         * @return the Object binding
         */
        public ITypeBinding getObjectBinding() {
            return objectBinding;
        }
        
        
    }

    
}
