/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.calculation.calculation.typeNames;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.intoJ.inferType.model.model.DeclarationElement;

/**
 * <p>This is the (one and only) InferType type name generator.</p>
 * 
 * <p>The generator generates names for new types for a certain Declaration Element. 
 * It is guaranteed that the given type name is a valid and unambiguous one. It looks like this:</p>
 * 
 * <pre>
 * SimpleName(declaredType) _ SimpleName(DE)    _of_SimpleName(method)  _ x  IT
 *
 *                                              | only                |   x: unambiguous  
 *                                              | for                 |      number
 *                                              | parameters          |
 * </pre>
 * 
 * <p>Note that the first element (the declared type) is ALWAYS the original type of the DE, not
 * an InferType2-induced type. For example, if the DE already has a type "SomeClass_field1_1IT",
 * only the first part ("SomeClass") will be re-used for the new type name.</p>
 * 
 * <p>Note also that each name is unambiguous for the complete project (regardless of the package).</p>
 * 
 * @version: $Date: 2007/09/09 19:40:25 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class DefaultTypeNameGenerator extends TypeNameGenerator {

	/**
	 * The constructor.
	 * 
	 * @param javaProject the current IJavaProject
	 */
	public DefaultTypeNameGenerator(IJavaProject javaProject) {

		/**
		 * Do not reset the type name generator, i.e. do not reset the "usedNames" list.
		 * The usedNames list is a static field, ensuring that no two types with the same
		 * names (regardless of the package!) are created during a COMPLETE run of Infer type.
		 * This ensures that no import name clashes happen, and also fixes the Windows problem
		 * of different cases in file names.
		 * 
		 */
		super(javaProject, false);
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.util.TypeNameGenerator#determineDefaultPackage(org.intoJ.inferType2.ast.model.DeclarationElement)
	 */
	protected String determinePackage(DeclarationElement de) {

		//ARRAY use package of element type
		ITypeBinding b = de.getDeclaredType();
		if (b.isArray())
			b = b.getElementType();

		return b.getPackage().getName();
	}

	/**
	 * Creates a new type name for this DE.
	 * 
	 * @param de DeclarationElement
	 * @return String
	 */
	public String getName(DeclarationElement de) {

		String postfix = "IT";
		String name;

		//ARRAY use element type
		ITypeBinding theType = de.getDeclaredType();
		if (theType.isArray())
			theType = theType.getElementType();

		// Add type prefix

		if (theType.getName().endsWith(postfix)) {
			// is IT type! get first 'element' before _ (is original type name)
			name = theType.getName().substring(0, theType.getName().indexOf('_'));
		} else {
			// is original type, get name
			name = transformName(theType.getName());
		}

		// Add simple name of de
		name += "_" + transformName(de.getSimpleName());

		// append suffix for parameters
		if (de.getKind() == DeclarationElement.PARAMETER)
			name += "_of_" + de.getMethodBinding().getName();

		// append count suffix and render the name unambigous. 
		int counter = 1;
		String packageName = determinePackage(de);
		name = name + "_" + counter + postfix;

		while (!isValid(packageName, name)) {

			name = name.substring(0, name.lastIndexOf("_"));
			name += "_" + counter + postfix;
			counter++;
		}

		usedNames.add(name);
		return packageName + "." + name;
	}

	/**
	 * This method transforms the declaration elements simple name to a valid
	 * type name by removing characters $, (, ), _, and ','.
	 * 
	 * The first character is uppercased.
	 *  
	 * @param simpleName the simple name of a declaration element
	 * @return a valid name for a type
	 */
	private String transformName(String simpleName) {

		StringBuffer name = new StringBuffer(simpleName.replaceAll("[$()_,]", ""));

		if (name.length() > 1)
			return Character.toUpperCase(name.charAt(0)) + name.substring(1);
		else
			return name.toString();

	}

}
