/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.model;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.intoJ.inferType.model.collection.DESet_Key;

/**
 * 
 * A composite declaration element represents a declaration elements and
 * overriding elements, for example, a field in a superclass and the same field
 * in a subclass.
 * 
 * Note that many attributes are the same between those elements, for example,
 * the declared type, signature etc.; however, some are different, for example,
 * the key of the binding.
 * 
 * @version: $Date: 2007/09/09 19:40:25 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class CompositeDeclarationElement extends DeclarationElement {

	/**
	 * A composite may only contain SingleDeclarationElements.
	 */
	private DESet_Key declarationElements = new DESet_Key();

	private String[] params = null;

	/**
	 * Adds a declaration element.
	 * 
	 * @param e DeclarationElement
	 */
	public void addDeclarationElement(DeclarationElement e) {

		declarationElements.addDeclaration(e);
	}

	/**
	 * Returns the fully qualified name of the type of all of the
	 * elements in this composite element.
	 * 
	 * @return String
	 */
	public String getFullyQualifiedNameOfType() {

		return getFirst().getFullyQualifiedNameOfType();
	}

	/**
	 * Returns a beaufified name of this composite element.
	 * 
	 * @return String
	 */
	public String getBeautifiedName() {

		StringBuffer buf = new StringBuffer("(");
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			buf.append(d.getBeautifiedName() + ",");
		}
		if (declarationElements.size() > 0)
			buf.deleteCharAt(buf.length() - 1);
		buf.append(")");
		return buf.toString();
	}

	/**
	 * This method is crucial to the success of the mission. It checks for
	 * variable and/or method equality.
	 * 
	 * DeclarationElements represent a lot of things: -> Invocations on
	 * variables and methods (Methods and Fields) -> Uses (where something is
	 * invoked on) (Methods, Fields, Local variables, Parameters) ->
	 * Declarations (a method or field declaration)
	 * 
	 * Each and every one of them shall be equal, for example, the invocation of
	 * a method and the declaration of it should be the same, because, for
	 * example, we need to check whether a type contains all of the methods
	 * invoked on a certain variable.
	 * 
	 * This is what the equals method is about.
	 * @param o Object
	 * @return boolean
	 */
	public boolean equals(Object o) {

		if (o instanceof CompositeDeclarationElement) {

			CompositeDeclarationElement other = (CompositeDeclarationElement) o;
			if (this.declarationElements.size() == other.declarationElements.size()) {
				//BUGFIX... I don't believe it ... compare it with the OTHER declaration elements, PLEASE. Thank you.
				return declarationElements.containsDeclarations(other.declarationElements);
			}
		}
		return false;
	}

	/**
	 * Method hashCode.
	 * @return int
	 */
	public int hashCode() {

		return toString().hashCode();
	}

	/**
	 * True if the given declaration element is equal according to signature to this one.
	 */
	public boolean equalsSignature(DeclarationElement d) {

		return getFirst().equalsSignature(d);
	}

	/**
	 * Returns the kind of this element.
	 * @see SingleDeclarationElement.getKind()
	 */
	public int getKind() {

		return getFirst().getKind();
	}

	/**
	 * Returns the simple name(s) of the elements in this composite.
	 * The simple name is NOT the same for every element in case of overridden parameters.
	 * 
	 * @return String
	 */
	public String getSimpleName() {

		// first create a set of all DIFFERENT simple names
		List simpleNames = new LinkedList();
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			String currentSimpleName = ((SingleDeclarationElement) i.next()).getSimpleName();
			if (!simpleNames.contains(currentSimpleName))
				simpleNames.add(currentSimpleName);
		}

		// if set has only one element then return this one
		if (simpleNames.size() == 1)
			return (String) simpleNames.iterator().next();

		// if simpleNames-set has more than 1 element create a String like this (name1,name2,...)
		// because we need a deterministic string creation (for the tests) first sort this set
		Collections.sort(simpleNames);
		StringBuffer buf = new StringBuffer("(");
		for (Iterator i = simpleNames.iterator(); i.hasNext();) {
			String simpleName = (String) i.next();
			buf.append(simpleName + ",");
		}
		if (declarationElements.size() > 0)
			buf.deleteCharAt(buf.length() - 1);
		buf.append(")");
		return buf.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#getDeclaredType()
	 */
	public ITypeBinding getDeclaredType() {

		return getFirst().getDeclaredType();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#getStartPosition()
	 */
	public int getStartPosition() {

		return 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#getModifiers()
	 */
	public int getModifiers() {

		DeclarationElement first = getFirst();
		return first.getModifiers();
	}

	/**
	 * Returns the first element from this composite.
	 * 
	 * @return SingleDeclarationElement
	 */
	public SingleDeclarationElement getFirst() {

		SingleDeclarationElement first = (SingleDeclarationElement) declarationElements.deIterator().next();
		return first;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#getBindings()
	 */
	public IBinding[] getBindings() {

		IBinding[] tmp = new IBinding[declarationElements.size()];
		int j = 0;
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			tmp[j] = d.getBinding();
			j++;
		}
		return tmp;
	}

	/**
	 * Returns the declaring classes of the elements in this composite
	 * 
	 * @return ITypeBinding[]
	 */
	public ITypeBinding[] getDeclaringClasses() {

		ITypeBinding[] tmp = new ITypeBinding[declarationElements.size()];
		int j = 0;
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			tmp[j] = d.getDeclaringClass();
			j++;
		}
		return tmp;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#toString()
	 */
	public String toString() {

		StringBuffer buf = new StringBuffer("(");
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			buf.append(d.toString() + ",");
		}
		if (declarationElements.size() > 0)
			buf.deleteCharAt(buf.length() - 1);
		buf.append(")");
		return buf.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#hasKey(java.lang.String)
	 */
	public boolean hasKey(String key) {

		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			if (d.getBinding().getKey().equals(key))
				return true;
		}
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.intoJ.model.DeclarationElement#containsDeclarationOnce(java.lang.String)
	 */
	public boolean containsDeclarationOnce(String string) {

		/*   for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
		 SingleDeclarationElement d = (SingleDeclarationElement) i.next();
		 if (d.getSimpleName().equals(string))
		 return true;
		 } */
		return getSimpleName().equals(string);
		//return false;
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getDeclaringClassesToString()
	 */
	public String getDeclaringClassesToString() {

		StringBuffer sBuf = new StringBuffer();
		ITypeBinding[] classes = getDeclaringClasses();
		sBuf.append("(");
		for (int i = 0; i < classes.length; i++) {
			sBuf.append(classes[i].getQualifiedName());
			if (i < classes.length - 1)
				sBuf.append("|");
		}
		sBuf.append(")");
		return sBuf.toString();
	}

	/**
	 * Returns the declarationElements
	 * 
	 * @return DESet_Key
	 */
	public DESet_Key getDeclarationElements() {

		return declarationElements;
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getMethodBinding()
	 */
	public IMethodBinding getMethodBinding() {

		DeclarationElement first = getFirst();
		return first.getMethodBinding();
	}

	public void setMethodParameterTypes(String[] params) {

		this.params = params;
	}

	public String[] getMethodParameterTypes() {

		if (params != null)
			return params;

		if (getFirst().getBinding() instanceof IMethodBinding) {
			IMethodBinding b = (IMethodBinding) getFirst().getBinding();
			ITypeBinding[] parameters = b.getParameterTypes();
			String[] returner = new String[parameters.length];
			for (int i = 0; i < returner.length; i++)
				returner[i] = parameters[i].getName();
			return returner;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#setNewDeclaredType(org.intoJ.inferType2.ast.model.TypeElement)
	 */
	public void setNewDeclaredType(TypeElement newDeclaredType) {

		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			d.setNewDeclaredType(newDeclaredType);
		}
	}

	/* (non-Javadoc)
	 * @see org.intoJ.inferType2.ast.model.DeclarationElement#getNewDeclaredType()
	 */
	public TypeElement getNewDeclaredType() {

		DeclarationElement first = getFirst();
		return first.getNewDeclaredType();
	}

	public String getCompleteInformation() {

		StringBuffer buf = new StringBuffer("COMPOSITE (");
		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			SingleDeclarationElement d = (SingleDeclarationElement) i.next();
			buf.append(d.getCompleteInformation() + ",");
		}
		if (declarationElements.size() > 0)
			buf.deleteCharAt(buf.length() - 1);
		buf.append(")");
		return buf.toString();
	}

}