/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import java.util.Iterator;
import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.intoJ.inferType.exception.InternalCalculationException;

/**
 * 
 * This class provides three functions for converting IJavaElements to IBindings.
 * 
 * * IType   -> ITypeBinding
 * * IMethod -> IMethodBinding
 * * IField  -> IVariableBinding
 * 
 * @version: $Date: 2007/09/09 19:40:25 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class Util_Conversion {

	/**
	 * Converts IJavaElements of type FIELD, METHOD, or TYPE into 
	 * their corresponding IBindings.
	 * 
	 * @param element
	 * @return IBinding
	 * @throws InternalCalculationException
	 */
	public static IBinding getBinding(IMember element) throws InternalCalculationException {

		if ((element.getElementType() != IJavaElement.FIELD) && (element.getElementType() != IJavaElement.METHOD)
				&& (element.getElementType() != IJavaElement.TYPE))
			throw new InternalCalculationException("Error", "Element type not supported: " + element.getElementType()
					+ " in IJavaElementToBinding.getBinding()");

		ICompilationUnit icu = element.getCompilationUnit();
		if (icu == null)
			throw new InternalCalculationException("InternalCalculationException", "Type can not be parsed.");

		ASTParser parser = ASTParser.newParser(AST.JLS2);
		parser.setSource(icu);
		parser.setResolveBindings(true);
		CompilationUnit cu = (CompilationUnit) parser.createAST(null);

		Visitor vis = new Visitor(element);
		cu.accept(vis);
		IBinding b = vis.getBinding();
		if (b != null)
			return b;

		throw new InternalCalculationException("Sorry", "Please select a field or a method; use the editor for local variables.");
	}

	/**
	 * This methods returns the corresponding ITypeBinding from a given IType.
	 * 
	 * @param type
	 * @return corresponding ITypeBinding
	 * @throws InternalCalculationException
	 *             is thrown if type could not be parsed (f.e. if type is a
	 *             binary one)
	 */
	public static ITypeBinding getTypeBinding(IType type) throws InternalCalculationException {

		return (ITypeBinding) getBinding(type);
	}

	private static class Visitor extends ASTVisitor {

		private IBinding binding;
		private IMember element;

		public Visitor(IMember element) {

			this.element = element;
		}

		public boolean visit(FieldDeclaration node) {

			if (node == null)
				return true;

			List fragments = node.fragments();
			Iterator i = fragments.iterator();
			while (i.hasNext()) {
				VariableDeclarationFragment currentFragment = (VariableDeclarationFragment) i.next();
				IVariableBinding currentFieldBinding = currentFragment.resolveBinding();
				if (currentFieldBinding.getJavaElement().equals(element))
					binding = currentFieldBinding;
			}

			return true;

		}

		public boolean visit(MethodDeclaration node) {

			if (node == null)
				return true;

			IMethodBinding currentMethodBinding = node.resolveBinding();
			if (currentMethodBinding.getJavaElement().equals(element))
				binding = currentMethodBinding;
			return true;
		}

		public boolean visit(TypeDeclaration node) {

			if (node == null)
				return true;

			ITypeBinding currentTypeBinding = node.resolveBinding();
			if (currentTypeBinding.getJavaElement().equals(element))
				binding = currentTypeBinding;
			return true;
		}

		/**
		 * @return Returns the binding.
		 */
		public IBinding getBinding() {

			return binding;
		}
	}
}
