/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.riwd.multirefac;

import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.print.attribute.standard.Severity;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaModelMarker;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.search.IJavaSearchScope;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.ltk.core.refactoring.RefactoringCore;
import org.eclipse.swt.widgets.Display;
import org.intoJ.inferType.calculation.calculation.typeNames.DefaultTypeNameGenerator;
import org.intoJ.inferType.exception.InternalCalculationException;
import org.intoJ.inferType.exception.UserAbortedException;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.model.model.DeclarationElement;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoj.riwd.ExternalRiwdRunnable;
import org.intoj.riwd.Manipulation;

/**
 * <p>
 * This class performs InferType2 Multi Refactoring on a given IJavaElement
 * (project, package, type), which means that all DEs of all the types inside
 * the scope are refactored according to InferType2, until no more changes are
 * possible.
 * </p>
 * 
 * 
 * <hr>
 * <h3>The algorithm:</h3>
 * 
 * <pre>
 *  currentCompilationUnit = first
 *  changedAll=false
 *  done=false
 *  do
 *   changed=false
 *   types = getTypesForCompilationUnit()
 *   for each type
 *     des = getDEsForType()
 *     for each de
 *       if (isExecutable())
 *         execute()
 *         changed=true
 *         changedAll=true
 *         goto continue
 *       endif
 *     endfor
 *   endfor
 *   
 *   :continue
 *   if (isLastCompilationUnit())
 *     if (!changedAll) done=true
 *     currentCompilationUnit = first
 *     changedAll=false
 *   else
 *     if (!changed)
 *       currentCompilationUnit = next compilation unit
 *   endif    
 *  
 *  while not done.
 * </pre>
 * 
 * <hr>
 * 
 * <p>
 * For more information about the ongoing process, please enable one or more of
 * the tracing options available in this plugin and the inferType plugin. If you
 * are only concerned with the ongoing process, enable only the multi refac
 * trace and, optionally, grep "STATUS" from the console output.
 * </p>
 * 
 * @see DefaultTypeNameGenerator
 * 
 * $Date: 2007/09/14 09:00:14 $Author: kegel $Revision: 1.7 $
 * @author Philip Mayer
 */
public class MultiRefactorer {

	protected IJavaProject javaProject;

	private String errorMessageIfNot;

	private IJavaSearchScope projectScope;

	private String errorMessage;

	private boolean error;

	int standardLen = 1000;

	int currentWorkedICUs = 0; // see below

	int currentWorkedLenAmount = 0; // see below

	private Map<ImmutabilityReason, Integer> immutabilityStats = new HashMap<ImmutabilityReason, Integer>();
	private int forwardingCount = 0;
	private int delegationCount = 0;
	private int errorCount = 0;
	private int totalTypes = 0;

	private final boolean excludeTestPackages;
	private final String excludePathFragment;

	public MultiRefactorer(boolean excludeTestPackages, String excludePathFragment) {
		this.excludeTestPackages = excludeTestPackages;
		this.excludePathFragment = excludePathFragment;
	}

	/**
	 * Main method for Infer Type Multi Refac.
	 * 
	 * Valid javaElements are IJavaProjects, IPackageFragments and
	 * ICompilationUnits.
	 * 
	 * @param theJavaElement
	 *            selected java element
	 * @param monitor
	 *            the monitor
	 * @throws InterruptedException
	 */
	public void run(IJavaElement javaElement, IProgressMonitor monitor)
			throws InterruptedException {

		javaProject = javaElement.getJavaProject();

		try {
			monitor.beginTask("Starting up...", standardLen);

			error = false;

			int currentICUIndex = 0; // Start at ICU 0
			int changesMade = 0;

			long startTime = System.currentTimeMillis();

			boolean clearCache = true;
			// get SORTED list of compilation units.
			List<ICompilationUnit> allICUs = grabCompilationUnits(javaElement);

			LogWriter logWriter = new LogWriter("c:\\riwd_" + javaProject.getElementName() + ".csv");

			for (ICompilationUnit currentICU : allICUs) {
				String path = currentICU.getPath().makeAbsolute().toString();
				if (excludeTestPackages && path.toLowerCase().contains(excludePathFragment.toLowerCase())) {
					currentICUIndex++;
					continue;
				}

				ILog
						.log(1,
								"---------------------------------------------------------------------");
				ILog.log(1,
						"RIWD Multirefac STATUS UPDATE - Changes made so far: "
								+ attachZeroes(changesMade, 4)
								+ ", Compilation Unit: "
								+ attachZeroes(currentICUIndex + 1, 3) + "/"
								+ attachZeroes(allICUs.size(), 3) + " ("
								+ currentICU.getElementName() + ")");
				ILog
						.log(1,
								"---------------------------------------------------------------------");

				monitor.setTaskName("RIWD Multi Changes: "
						+ attachZeroes(changesMade, 4) + "; Files: "
						+ attachZeroes(currentICUIndex + 1, 3) + "/"
						+ attachZeroes(allICUs.size(), 3));

				if (currentICUIndex != currentWorkedICUs) {
					int absolute = (standardLen * (currentICUIndex + 1))
							/ allICUs.size();
					int relative = absolute - currentWorkedLenAmount;
					if (relative > 0) { // if <0, wait (effect induced by adding
						// comp. units)
						currentWorkedLenAmount = absolute;
						currentWorkedICUs = currentICUIndex;
						monitor.worked(relative);
					}
				}

				// get a list of ITypes in a DEFINED ORDER
				IType[] types = currentICU.getAllTypes();
				totalTypes += types.length;


				for (int i = 0; i < types.length; i++) {

					IType currentType = types[i];
					final int currentIno = i + 1;
					ILog.log(2, "IType " + currentIno + " of " + types.length
							+ ": " + currentType.getElementName());

					try {
						ExternalRiwdRunnable riwdRunnable = doRiwdRunnable(
								new SubProgressMonitorWoTaskName(monitor, 0),
								currentType, clearCache, logWriter);
						if (riwdRunnable != null) {
							boolean compileError = false;
							if (riwdRunnable.wasApplicable()) {
								compileError = !compile(currentType.getFullyQualifiedName());
								if (!compileError) {
									changesMade++;
								}
							}
							if (!compileError) {
								logWriter.log(currentType, riwdRunnable);
								types = currentICU.getAllTypes();
							}							
						}
						clearCache = false;
					} catch (UserAbortedException e) {
						throw new UserAbortedException("RIWD",
								"RIWD was canceled.");
					} catch (Exception e) {
						ILog.logError("Exception while running RIWD.", e);
						e.printStackTrace();
						error = true;
						return; // PANIC
					}

				}

				currentICUIndex++;
			}

			long timeNow = System.currentTimeMillis();

			ILog
					.log(1,
							"---------------------------------------------------------------------");
			if (error) {
				errorMessage = "Multirefac has come to a halt due to an error. Sorry.";

			} else {
				errorMessage = "Multirefac has come to its final end. Thank you for using IT technology.";
			}
			ILog.log(1, errorMessage);
			ILog.log(1, "Done. Changes made: " + changesMade + ".");
			ILog.log(1, "Time taken: " + readableTime(timeNow - startTime));
			ILog
					.log(1,
							"---------------------------------------------------------------------");

			monitor.done();
			logWriter.close();

		} catch (IOException e) {
			ILog.logError("There was a problem writing the log.", e);
			throw new InterruptedException(
					"There was a problem writing the log.");
		} catch (JavaModelException e) {
			ILog.logError("There was a problem accessing the Java Model.", e);
			throw new InterruptedException(
					"There was a problem accessing the Java Model.");
		} catch (CoreException e) {
			ILog.logError(("Core Exception in MultiRefac"), e);
			throw new InternalCalculationException("Core Exception",
					"Problem in MultiRefac: " + e.getMessage());
		}
	}

	private boolean compile(String currentTypeInfo) throws CoreException {
		if (!cleanAndRebuild()) {
			errorCount++;
			Activator.getDefault().getLog()
					.log(
							new Status(IStatus.ERROR, Activator.getDefault()
									.getBundle().getSymbolicName(), IStatus.OK,
									"Error while refactoring "
											+ currentTypeInfo, null));
			final CoreException[] exHolder = new CoreException[1];
			Display.getDefault().syncExec(new Runnable() {
				public void run() {
					try {
						RefactoringCore.getUndoManager().performUndo(null,
								new NullProgressMonitor());
					} catch (CoreException e) {
						exHolder[0] = e;
					}
				}
			});
			if (exHolder[0] != null || !cleanAndRebuild()) {
				Activator.getDefault().getLog().log(
						new Status(IStatus.ERROR, Activator.getDefault()
								.getBundle().getSymbolicName(), IStatus.OK,
								"Could not undo " + currentTypeInfo, null));
				error = true;
				throw new CoreException(new Status(Status.ERROR,
						Activator.PLUGIN_ID, "could not undo"));
			}
			return false;
		}
		return true;
	}

	private void addImmutabilityReason(ImmutabilityReason reason) {
		Integer count = (Integer) immutabilityStats.get(reason);
		if (count == null) {
			count = Integer.valueOf(0);
		}
		immutabilityStats.put(reason, Integer.valueOf(count.intValue() + 1));
	}

	private void addImmutabilityReasons(EnumSet<ImmutabilityReason> reasons) {
		for (ImmutabilityReason immutabilityReason : reasons) {
			addImmutabilityReason(immutabilityReason);
		}
	}

	private boolean cleanAndRebuild() {

		// CLEAN AND REBUILD:

		try {
			javaProject.getProject().build(
					IncrementalProjectBuilder.CLEAN_BUILD,
					new NullProgressMonitor());
			javaProject.getProject().build(
					IncrementalProjectBuilder.FULL_BUILD,
					new NullProgressMonitor());
			javaProject.close();
			javaProject.open(new NullProgressMonitor());
		} catch (CoreException e) {
			ILog.log(1, "Sorry. CoreException while rebuilding.");
			return false;
		}

		if (hasRealErrors(javaProject.getProject())) {
			ILog
					.log(1,
							"Sorry. Aborted due to compilation errors in project. ");
			return false;
		}

		return true;
	}

	/**
	 * 
	 * Returns true if there was a change.
	 * 
	 * @param monitor
	 * @param des
	 * @param currentDE
	 * @param currentType
	 * @param clearCache
	 * @param logWriter
	 * @return
	 * @throws InterruptedException
	 */
	private ExternalRiwdRunnable doRiwdRunnable(final IProgressMonitor monitor,
			IType currentType, boolean clearCache, LogWriter logWriter)
			throws InterruptedException {
		ExternalRiwdRunnable riwdRunnable = null;

		ILog.log(3, "=============> GOTO Riwd <===============");
		try {
			try {
				riwdRunnable = new ExternalRiwdRunnable(currentType,
						clearCache, false);

				riwdRunnable.run(monitor);
				if (riwdRunnable.wasApplicable()) {
					if (riwdRunnable.getForwardingStatus().isPossible()) {
						forwardingCount++;
					} else {
						delegationCount++;
					}
				} else {
					addImmutabilityReasons(riwdRunnable.getCommonStatus()
							.getReasons());
					addImmutabilityReasons(riwdRunnable.getForwardingStatus()
							.getReasons());
					addImmutabilityReasons(riwdRunnable.getDelegationStatus()
							.getReasons());
				}
			} catch (OperationCanceledException e1) {
				throw new UserAbortedException("cancled", "");
			} catch (InvocationTargetException e) {
				Throwable target = e.getTargetException();
				if (target instanceof OperationCanceledException) {
					throw new UserAbortedException("cancled", "");
				}
				target.printStackTrace();
				ILog.logError("Error refactoring "
						+ currentType.getFullyQualifiedName(), target);
			} catch (Exception e) {
				e.printStackTrace();
				ILog.logError("Error refactoring "
						+ currentType.getFullyQualifiedName(), e);
			}
		} finally {
			ILog.log(3, "=============> COMEFROM Riwd <===============");
			ILog.log(3, "Done multi-refaccing type "
					+ currentType.getFullyQualifiedName());
		}

		return riwdRunnable;
	}

	public boolean hasRealErrors(IResource resource) {

		try {
			if (resource != null && resource.exists()) {
				IMarker[] marker = resource.findMarkers(
						IJavaModelMarker.JAVA_MODEL_PROBLEM_MARKER, false,
						IResource.DEPTH_INFINITE);
				for (int i = 0; i < marker.length; i++) {
					if (((Integer) marker[i].getAttribute(IMarker.SEVERITY))
							.intValue() == IMarker.SEVERITY_ERROR)
						return true;
				}
			}
		} catch (CoreException e) {
			// assume there are no problems
		}
		return false;
	}

	/**
	 * Returns all ICompilationUnits of the given IJavaElement.
	 * 
	 * @param javaElement
	 *            java element
	 * @return List of ICompilationUnit
	 * @throws InterruptedException
	 */
	private List grabCompilationUnits(IJavaElement javaElement)
			throws JavaModelException {

		List scopeElements = new ArrayList();
		ICompilationUnit[] units = null;

		if (javaElement instanceof ICompilationUnit)
			scopeElements.add(javaElement);

		else if (javaElement instanceof IPackageFragment) {
			units = ((IPackageFragment) javaElement).getCompilationUnits();
			for (int j = 0; j < units.length; j++)
				scopeElements.add(units[j]);
		}

		else if (javaElement instanceof IJavaProject) {
			IPackageFragment[] packages = ((IJavaProject) javaElement)
					.getPackageFragments();
			for (int i = 0; i < packages.length; i++) {
				if (packages[i].getKind() != IPackageFragmentRoot.K_BINARY) {
					units = packages[i].getCompilationUnits();
					for (int j = 0; j < units.length; j++)
						scopeElements.add(units[j]);
				}
			}
		}

		Collections.sort(scopeElements, new CompilationUnitComparator());

		return scopeElements;
	}

	/**
	 * This comparator compares ICompilationUnits.
	 * 
	 * @author Andreas Mei&szlig;ner
	 */
	private class CompilationUnitComparator implements Comparator {

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object arg0, Object arg1) {

			if ((arg0 == null) || (arg1 == null))
				return 0;
			if (!(arg0 instanceof ICompilationUnit))
				return 0;
			if (!(arg1 instanceof ICompilationUnit))
				return 0;

			ICompilationUnit icu0 = (ICompilationUnit) arg0;
			ICompilationUnit icu1 = (ICompilationUnit) arg1;

			return icu0.getElementName().compareTo(icu1.getElementName());
		}

	}

	/**
	 * This comparator compares DeclarationElements.
	 * 
	 * @author Philip Mayer
	 */
	private class DeclarationElementComparator implements Comparator {

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object arg0, Object arg1) {

			if ((arg0 == null) || (arg1 == null))
				return 0;
			if (!(arg0 instanceof DeclarationElement))
				return 0;
			if (!(arg1 instanceof DeclarationElement))
				return 0;

			DeclarationElement de0 = (DeclarationElement) arg0;
			DeclarationElement de1 = (DeclarationElement) arg1;

			return de0.getSimpleName().compareTo(de1.getSimpleName()); // simple
			// name
			// does
			// not
			// change
			// in a
			// refactoring!
		}

	}

	public static String readableTime(long time) {

		long allSeconds = time / 1000;

		long allMinutes = allSeconds / 60;
		long allHours = allMinutes / 60;

		long minutes = allMinutes % 60;
		long seconds = allSeconds % 60;

		return attachZero(allHours) + ":" + attachZero(minutes) + ":"
				+ attachZero(seconds);
	}

	/**
	 * @param number
	 * @return
	 */
	private static String attachZero(long number) {

		return (number < 10) ? "0" + number : String.valueOf(number);
	}

	private static String attachZeroes(long number, int digits) {

		int add = digits - String.valueOf(number).length();
		if (add <= 0)
			return String.valueOf(number);
		StringBuffer b = new StringBuffer();
		for (int i = 0; i < add; i++)
			b.append("0");
		b.append(number);
		return b.toString();
	}

	public boolean isError() {

		return error;
	}

	public String getErrorMessage() {
		StringBuilder message = new StringBuilder(errorMessage).append("\n");
		message.append("Types: ").append(totalTypes).append(", Forwarding: ")
				.append(forwardingCount);
		message.append(", Delegation: ").append(delegationCount).append(
				", Errors: ").append(errorCount);

		for (Map.Entry<ImmutabilityReason, Integer> entry : immutabilityStats
				.entrySet()) {
			message.append("\n").append(entry.getKey()).append(": ").append(
					entry.getValue());
		}

		return message.toString();
	}

	private static class SubProgressMonitorWoTaskName extends
			SubProgressMonitor {
		public SubProgressMonitorWoTaskName(IProgressMonitor monitor, int ticks) {
			super(monitor, ticks, SUPPRESS_SUBTASK_LABEL);
		}

		public void setTaskName(String name) {
			// ignore
		}
	}

	private static class LogWriter {
		private PrintWriter pw;

		public LogWriter(String fileName) throws IOException {
			pw = new PrintWriter(new FileOutputStream(fileName), true);

			pw.print("name;forwarding;delegation;methodsBefore;methodsAfter;staticBefore;staticAfter;delegating;reverse;super;newInterfaces;deadMethods;"
							+ "newlyDeclaredBefore;newlyDeclaredStaticBefore;onlyInheritedBefore;onlyInheritedStaticBefore;newSuperclass");

			for (ImmutabilityReason reason : ImmutabilityReason.values()) {
				pw.print(";" + reason);
			}
			pw.println();
		}

		public void log(IType currentType, ExternalRiwdRunnable riwdRunnable) {
			Manipulation.Stats stats = riwdRunnable.getStats();
			if (stats == null) {
				return;
			}

			EnumSet<ImmutabilityReason> commonReasons = riwdRunnable
					.getCommonStatus().getReasons();
			if (commonReasons.contains(ImmutabilityReason.NO_CLASS)
					|| commonReasons
							.contains(ImmutabilityReason.OBJECT_SUPERTYPE)
					|| commonReasons
							.contains(ImmutabilityReason.NO_TOP_LEVEL_TYPE)) {
				return;
			}

			pw.print(currentType.getFullyQualifiedName() + ";");
			pw.print(getNumber(riwdRunnable.wasApplicable()
					&& riwdRunnable.getForwardingStatus().isPossible())
					+ ";");
			pw.print(getNumber(riwdRunnable.wasApplicable()
					&& riwdRunnable.getDelegationStatus().isPossible())
					+ ";");

			if (riwdRunnable.wasApplicable()) {
				pw.print(stats.getNonPrivateMethodsBefore() + ";");
				pw.print(stats.getNonPrivateMethodsAfter() + ";");
				pw.print(stats.getNonPrivateStaticMethodsBefore() + ";");
				pw.print(stats.getNonPrivateStaticMethodsAfter() + ";");
				pw.print(stats.getDelegatingMethods() + ";");
				pw.print(stats.getReverseDelegatingMethods() + ";");
				pw.print(stats.getSuperDelegatingMethods() + ";");
				pw.print(stats.getNewInterfaces() + ";");
				pw.print(stats.getDeadMethods() + ";");

				pw.print(stats.getNewlyDeclaredNonPrivateMethodsBefore() + ";");
				pw.print(stats.getNewlyDeclaredNonPrivateStaticMethodsBefore() + ";");
				pw.print(stats.getOnlyInheritedMethodsBefore() + ";");
				pw.print(stats.getOnlyInheritedStaticMethodsBefore() + ";");

				pw.print(getNumber(stats.isNewSupertype()));
			} else {
				pw.print(stats.getNonPrivateMethodsBefore() + ";");
				pw.print(stats.getNonPrivateMethodsBefore() + ";");
				pw.print(stats.getNonPrivateStaticMethodsBefore() + ";");
				pw.print(stats.getNonPrivateStaticMethodsBefore() + ";");
				pw.print("0;");
				pw.print("0;");
				pw.print("0;");
				pw.print("0;");
				pw.print("0;");

				pw.print(stats.getNewlyDeclaredNonPrivateMethodsBefore() + ";");
				pw.print(stats.getNewlyDeclaredNonPrivateStaticMethodsBefore() + ";");
				pw.print(stats.getOnlyInheritedMethodsBefore() + ";");
				pw.print(stats.getOnlyInheritedStaticMethodsBefore() + ";");

				pw.print("");
			}
			
			//TODO: remove temp
			EnumSet<ImmutabilityReason> allReasons = EnumSet.copyOf(commonReasons);
			allReasons.addAll(riwdRunnable.getForwardingStatus().getReasons());
			allReasons.addAll(riwdRunnable.getDelegationStatus().getReasons());
			if (allReasons.contains(ImmutabilityReason.INSTANCE_OF)) {
				allReasons.add(ImmutabilityReason.ASSIGNMENT_TO_SUPERCLASS);
			}
			if (allReasons.contains(ImmutabilityReason.CAST)) {
				allReasons.add(ImmutabilityReason.ASSIGNMENT_TO_SUPERCLASS);
			}

			for (ImmutabilityReason reason : ImmutabilityReason.values()) {
				if (reason == ImmutabilityReason.INSTANCE_OF || reason == ImmutabilityReason.CAST) {
					continue; 
				}
				pw.print(";" + getNumber(allReasons.contains(reason)));
			}
			pw.println();
			pw.flush();
		}

		private String getNumber(boolean b) {
			if (b)
				return "1";
			else
				return "";
		}

		public void close() {
			pw.close();
		}

	}
}
