/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoj.riwd.actions;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ltk.ui.refactoring.RefactoringWizardOpenOperation;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.util.RefactoringSaveHelper;
import org.intoj.riwd.RiwdRefactoring;
import org.intoj.riwd.ui.RiwdWizard;

/**
 * Abstract superclass of all actions.
 *
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */

public class AbstractRiwdAction {
    
    protected ISelection selection;    
	
    /**
     * Starts Infer Type by running a RefactoringWizardOpenOperation.
     * 
     * @param declarationElement the selected element
     * @throws InterruptedException
     */
    protected final void runWithSelectedElement(SelectedElement declarationElement) throws InterruptedException {
        if (checkPrerequesites(declarationElement)) {        
	        RiwdRefactoring refactoring= new RiwdRefactoring(declarationElement);
	        RiwdWizard refactoringWizard= new RiwdWizard(refactoring);
	        
	        RefactoringSaveHelper saveHelper = new RefactoringSaveHelper();
	        if (saveHelper.saveEditors(getShell())) {
	            RefactoringWizardOpenOperation op= new RefactoringWizardOpenOperation(refactoringWizard);                
	            op.run(getShell(), "Infer Type");
	        }
        }
    }
    
    /**
     * Sets the selection.
     * 
     * @param action IAction
     * @param selection ISelection
     */
    public void selectionChanged(IAction action, ISelection selection) {
        this.selection = selection;
    }
        

    /**
     * Checks whether it is possible to work with selected declaration element.
     * 
     * @param de the declaration element
     * @return true if it is possible to work with the selected declaration
     *         element
     */
    private boolean checkPrerequesites(SelectedElement de) {

        //ARRAY use element type
        ITypeBinding typeToCheck = de.getDeclaredType();
        if (typeToCheck.isArray()) {
            typeToCheck = typeToCheck.getElementType();
        }

        if (!typeToCheck.isFromSource()) {
            showError("Non-project type", "The selected type is not part of this project.");
            return false;
        }
        if (typeToCheck.isEnum()) {
            showError("Enum", "Replace Inheritance With Delegation is not defined for enums.");
            return false;
        }
        if (typeToCheck.isAnnotation()) {
            showError("Annotation", "Replace Inheritance With Delegation is not defined for annotation types.");
            return false;
        }
        if (typeToCheck.isInterface()) {
            showError("Interface", "Replace Inheritance With Delegation is not defined for interfaces.");
            return false;
        }
        return true;
    }

	/**
     * Shows an error message.
     * 
     * @param title the title of the dialog
     * @param msg the error message
     */
    protected final void showError(String title, String msg) {
        MessageDialog.openError(getShell(), title, msg);
    }

    /**
     * Returns the current shell
     * 
     * @return shell
     */
    protected Shell getShell() {

        IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
        if (window == null) {
            return null;
        }
        return window.getShell();
    }   
}
